/* $Id: kink_conf.c,v 1.17 2005/01/19 00:53:47 kamada Exp $ */
/*
 * Copyright (C) 2004 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "config.h"

#include <sys/types.h>
#include <sys/socket.h>

#include <netinet/in.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "racoon.h"
#include "utils.h"
#include "sockmisc.h"
#include "bbkk.h"
#include "kink_conf.h"
#include "rct_ipsecdoi.h"	/* XXX tmp */
#include "pfkey.h"		/* XXX tmp */
#include "isakmp.h"		/* XXX tmp */
#include "proposal.h"		/* XXX tmp */
#include "ipsec_doi.h"		/* XXX tmp */


extern char *config_file;

extern struct rcf_default *rcf_default_head;
extern struct rcf_interface *rcf_interface_head;
extern struct rcf_remote *rcf_remote_head;
extern struct rcf_selector *rcf_selector_head;


static struct saproto *conv_ipsec2saproto(const struct rcf_ipsec *ipsec);
static struct saproto *conv_sa2saproto(const struct rcf_sa *sa, int encmode);
static int rcf_alg_keylen(const struct rc_alglist *alg);


char *
get_default_principal(void)
{
	char *dst;
	vchar_t *src;

	if (rcf_default_head == NULL ||
	    rcf_default_head->remote == NULL ||
	    rcf_default_head->remote->kink == NULL ||
	    rcf_default_head->remote->kink->my_principal == NULL) {
		kinkd_log(KLLV_ERR,
		    "%s: specify my principal in "
		    "default remote kink my_principal\n", config_file);
		return NULL;
	}

	src = rcf_default_head->remote->kink->my_principal;
	if ((dst = (char *)malloc(src->l + 1)) == NULL) {
		kinkd_log(KLLV_FATAL, "out of memory\n");
		return NULL;
	}
	memcpy(dst, src->v, src->l);
	dst[src->l] = '\0';
	return dst;
}

int
get_nonce_size(struct rcf_remote *rm)
{
	if (rm != NULL &&
	    rm->kink != NULL &&
	    rm->kink->nonce_size != 0)
		return rm->kink->nonce_size;
	if (rcf_default_head != NULL &&
	    rcf_default_head->remote != NULL &&
	    rcf_default_head->remote->kink != NULL &&
	    rcf_default_head->remote->kink->nonce_size != 0)
		return rcf_default_head->remote->kink->nonce_size;
	return DEFAULT_NONCE_SIZE;
}

int
match_saddr_with_config(struct sockaddr *sa, const char *ifname)
{
	char port[16];
	struct rc_addrlist *al;
	const char *macro;

	/* if no config, use all addresses */
	if (rcf_interface_head == NULL ||
	    rcf_interface_head->kink == NULL) {
		setport(sa, KINK_DEFAULT_PORT);
		return 1;
	}

#define STRCMP2(target, prefix, ifname)					\
	((strncmp(target, prefix, sizeof(prefix) - 1) == 0 &&		\
	  strcmp(target + sizeof(prefix) - 1, ifname) == 0) ? 0 : 1)
#define SA_IS_IN6_LINKLOCAL(sa)					\
	IN6_IS_ADDR_LINKLOCAL(&((struct sockaddr_in6 *)(sa))->sin6_addr)
#define SA_IS_IN6_SITELOCAL(sa)					\
	IN6_IS_ADDR_SITELOCAL(&((struct sockaddr_in6 *)(sa))->sin6_addr)

	/* check addr in config */
	for (al = rcf_interface_head->kink; al != NULL; al = al->next) {
		switch (al->type) {
		case RCT_ADDR_INET:
			if (rcs_cmpsa_wop(al->a.ipaddr, sa) == 0)
				goto match;
			break;
		case RCT_ADDR_MACRO:
			macro = vmem2str(al->a.vstr);
			if (strcmp(macro, "MY_IP") == 0)
				goto match;
			if (STRCMP2(macro, "MY_IP%", ifname) == 0)
				goto match;
			if (strcmp(macro, "MY_IPV6") == 0 &&
			    sa->sa_family == AF_INET6)
				goto match;
			if (STRCMP2(macro, "MY_IPV6%", ifname) == 0 &&
			    sa->sa_family == AF_INET6)
				goto match;
			if (strcmp(macro, "MY_IPV6_GLOBAL") == 0 &&
			    sa->sa_family == AF_INET6 &&
			    !SA_IS_IN6_LINKLOCAL(sa) &&
			    !SA_IS_IN6_SITELOCAL(sa))
				goto match;
			if (STRCMP2(macro, "MY_IPV6_GLOBAL%", ifname) == 0 &&
			    sa->sa_family == AF_INET6 &&
			    !SA_IS_IN6_LINKLOCAL(sa) &&
			    !SA_IS_IN6_SITELOCAL(sa))
				goto match;
			if (strcmp(macro, "MY_IPV6_LINKLOCAL") == 0 &&
			    sa->sa_family == AF_INET6 &&
			    SA_IS_IN6_LINKLOCAL(sa))
				goto match;
			if (STRCMP2(macro, "MY_IPV6_LINKLOCAL%", ifname) == 0 &&
			    sa->sa_family == AF_INET6 &&
			    SA_IS_IN6_LINKLOCAL(sa))
				goto match;
			if (strcmp(macro, "MY_IPV4") == 0 &&
			    sa->sa_family == AF_INET)
				goto match;
			if (STRCMP2(macro, "MY_IPV4%", ifname) == 0 &&
			    sa->sa_family == AF_INET)
				goto match;
			break;
		default:
			break;
		}
	}
	return 0;		/* unmatch */

match:
	/*
	 * XXX libracoon is tooooooooooo kind and returns its own
	 * default port...    but "remote kink"s port is only a
	 * source port, and default destination port is managed by kinkd
	 * as KINK_DEFAULT_PORT.
	 * we may be able to use RC_PORT_KINK, but embedding it in
	 * 2 binaries (libracoon and kinkd) is the source of troubles...
	 * so... what should we do?
	 */
	if (al->port == 0)	/* this never match */
		setport(sa, KINK_DEFAULT_PORT);
	else {
		(void)snprintf(port, sizeof(port), "%u", al->port);
		setport(sa, port);
	}
	return 1;		/* match */
}

/* ----------------------------------------------------------------
 * XXX temporary functions to convert rcf_policy to saprop
 * policy <-> SA payload conversion and policy matching should
 * be reworked.
 * There functions assume automatic key management so
 * rc_alglist->key, etc are not converted.
 * ---------------------------------------------------------------- */

struct saprop *
conv_policy2saprop(const struct rcf_policy *policy)
{
	const struct rcf_ipsec *ipsec;
	struct saprop *pp, **pp_ptr, *tmp;
	int propno;

	propno = 0;
	pp = NULL;
	pp_ptr = &pp;

	if (policy == NULL) {
		kinkd_log(KLLV_ERR, "INTERNAL ERROR: policy == NULL\n");
		return NULL;
	}

	for (ipsec = policy->ips; ipsec != NULL; ipsec = ipsec->next) {
		if ((tmp = newsaprop()) == NULL) {
			kinkd_log(KLLV_FATAL, "out of memory");
			goto fail;
		}
		tmp->prop_no = ++propno;
		tmp->lifetime = ipsec->ipsec_sa_lifetime_time;
		tmp->lifebyte = ipsec->ipsec_sa_lifetime_byte;
		/* pfs_group */
		/* claim */
		tmp->head = conv_ipsec2saproto(ipsec);
		*pp_ptr = tmp;
		pp_ptr = &tmp->next;

		if (tmp->head == NULL)
			goto fail;
	}
	return pp;

fail:
	flushsaprop(pp);
	return NULL;
}

static struct saproto *
conv_ipsec2saproto(const struct rcf_ipsec *ipsec)
{
	struct saproto *pr, **pr_ptr, *tmp;

	pr = NULL;
	pr_ptr = &pr;

	if (ipsec->sa_ah != NULL) {
		tmp = conv_sa2saproto(ipsec->sa_ah,
		    rcf2ipsecdoi_mode(ipsec->ipsec_mode));
		if (tmp == NULL)
			goto fail;
		*pr_ptr = tmp;
		pr_ptr = &tmp->next;
	}
	if (ipsec->sa_esp != NULL) {
		tmp = conv_sa2saproto(ipsec->sa_esp,
		    rcf2ipsecdoi_mode(ipsec->ipsec_mode));
		if (tmp == NULL)
			goto fail;
		*pr_ptr = tmp;
		pr_ptr = &tmp->next;
	}
	if (ipsec->sa_ipcomp != NULL) {
		tmp = conv_sa2saproto(ipsec->sa_ipcomp,
		    rcf2ipsecdoi_mode(ipsec->ipsec_mode));
		if (tmp == NULL)
			goto fail;
		*pr_ptr = tmp;
		pr_ptr = &tmp->next;
	}
	return pr;

fail:
	flushsaproto(pr);
	return NULL;
}

static struct saproto *
conv_sa2saproto(const struct rcf_sa *sa, int encmode)
{
	const struct rc_alglist *aalg, *ealg;
	struct saproto *pr;
	struct satrns *tr, **tr_ptr;
	int trnsno;

	if ((pr = newsaproto()) == NULL)
		return NULL;
	trnsno = 0;

	switch (sa->sa_protocol) {
	case RCT_SATYPE_ESP:
		pr->proto_id = IPSECDOI_PROTO_IPSEC_ESP;
		pr->spisize = 4;
		pr->encmode = encmode;

		tr_ptr = &pr->head;
		for (ealg = sa->enc_alg; ealg != NULL; ealg = ealg->next) {
			for (aalg = sa->auth_alg; aalg != NULL; aalg = aalg->next) {
				if ((tr = newsatrns()) == NULL) {
					kinkd_log(KLLV_FATAL, "out of memory");
					goto fail;
				}
				tr->trns_no = ++trnsno;
				tr->trns_id = rcf2ipsecdoi_ealg(ealg->algtype);
				tr->encklen = rcf_alg_keylen(ealg);
				tr->authtype = rcf2ipsecdoi_aattr(aalg->algtype);

				*tr_ptr = tr;
				tr_ptr = &tr->next;
			}
		}
		/* ignore comp_alg */
		break;
	case RCT_SATYPE_AH:
		pr->proto_id = IPSECDOI_PROTO_IPSEC_AH;
		pr->spisize = 4;
		pr->encmode = encmode;

		tr_ptr = &pr->head;
		for (aalg = sa->auth_alg; aalg != NULL; aalg = aalg->next) {
			if ((tr = newsatrns()) == NULL) {
				kinkd_log(KLLV_FATAL, "out of memory");
				goto fail;
			}
			tr->trns_no = ++trnsno;
			tr->trns_id = rcf2ipsecdoi_aalg(aalg->algtype);
			tr->encklen = 0;
			tr->authtype = rcf2ipsecdoi_aattr(aalg->algtype);

			*tr_ptr = tr;
			tr_ptr = &tr->next;
		}
		/* ignore enc_alg and comp_alg */
		break;
	case RCT_SATYPE_IPCOMP:
		kinkd_log(KLLV_ERR, "XXX implement me\n");
		goto fail;
		/* ignore enc_alg and auth_alg */
		break;
	default:
		/* this rcf_sa is from config so bundle type must not appears */
		kinkd_log(KLLV_ERR,
		    "unexpected satype (%s)\n", rct2str(sa->sa_protocol));
		goto fail;
	}
	return pr;

fail:
	flushsaproto(pr);
	return NULL;
}

static int
rcf_alg_keylen(const struct rc_alglist *alg)
{
	if (alg->keylen != 0)
		return alg->keylen;
	switch (alg->algtype) {
	case RCT_ALG_AES128_CBC:
		return 128;
	case RCT_ALG_AES192_CBC:
		return 192;
	case RCT_ALG_AES256_CBC:
		return 256;
	default:
		return 0;
	}
}


struct rcf_remote *
get_remote(const vchar_t *rm_index)
{
	struct rcf_remote *rm;

	for (rm = rcf_remote_head; rm != NULL; rm = rm->next)
		if (vmemcmp(rm_index, rm->rm_index) == 0)
			return rm;
	return NULL;
}

struct rcf_remote *
get_remote_by_principal(bbkk_context con, const char *principal)
{
	struct rcf_remote *rm;
	const char *rm_princ;

	for (rm = rcf_remote_head; rm != NULL; rm = rm->next) {
		if (rm->kink == NULL ||
		    rm->kink->peers_principal == NULL)
			continue;
		rm_princ = vmem2str(rm->kink->peers_principal);
		if (bbkk_cmp_principal(con, principal, rm_princ) == 0)
			return rm;
	}
	return NULL;
}

struct rcf_selector *
get_selector(int slid)
{
	struct rcf_selector *sl;

	for (sl = rcf_selector_head; sl != NULL; sl = sl->next)
		if (sl->sl_index == slid)
			return sl;
	return NULL;
}

struct rcf_selector *
get_selector_by_fqdn(const char *str, int dir)
{
	struct rcf_selector *sl, *candidate;
	struct rc_addrlist *al;
	char *tptr;
	size_t tlen;

	candidate = NULL;
	for (sl = rcf_selector_head; sl != NULL; sl = sl->next) {
		if (sl->direction != dir)
			continue;
		if (sl->direction == RCT_DIR_OUTBOUND)
			al = sl->dst;
		else
			al = sl->src;
		for (; al != NULL; al = al->next) {
			if (al->type != RCT_ADDR_FQDN)
				continue;
			tptr = al->a.vstr->v;
			tlen = al->a.vstr->l;
			/* FQDNs in config may have trailing dots */
			if (tlen > 0 && tptr[tlen - 1] == '.')
				tlen--;
			if (tlen != strlen(str) || memcmp(tptr, str, tlen) != 0)
				continue;
			break;
		}
		if (al == NULL)
			continue;		/* not match */

		if (candidate == NULL ||
		    sl->order < candidate->order)
			candidate = sl;
	}
	return candidate;
}

struct rcf_selector *
get_selector_by_sa(const struct sockaddr *my_sa,
    const struct sockaddr *peers_sa)
{
	struct rcf_selector *sl, *candidate;
	struct rc_addrlist *mal, *pal;

	candidate = NULL;
	for (sl = rcf_selector_head; sl != NULL; sl = sl->next) {
		if (sl->pl != NULL &&
		    sl->pl->ips != NULL &&
		    sl->pl->ips->ipsec_mode == RCT_IPSM_TUNNEL) {
			mal = sl->pl->my_sa_addr;
			pal = sl->pl->peers_sa_addr;
		} else {
			if (sl->direction == RCT_DIR_OUTBOUND) {
				mal = sl->src;
				pal = sl->dst;
			} else {
				mal = sl->dst;
				pal = sl->dst;
			}
		}

		if (my_sa != NULL) {
			for (; mal != NULL; mal = mal->next) {
				if (mal->type != RCT_ADDR_INET)
					continue;
				if (rcs_cmpsa_wop(my_sa, mal->a.ipaddr) != 0)
					continue;
				break;
			}
			if (mal == NULL)
				continue;		/* not match */
		}
		if (peers_sa != NULL) {
			for (; pal != NULL; pal = pal->next) {
				if (pal->type != RCT_ADDR_INET)
					continue;
				if (rcs_cmpsa_wop(peers_sa, pal->a.ipaddr) != 0)
					continue;
				break;
			}
			if (pal == NULL)
				continue;		/* not match */
		}

		if (candidate == NULL ||
		    sl->order < candidate->order)
			candidate = sl;
	}
	return candidate;
}
