/* $Id: kink_fmt.c,v 1.52 2004/12/24 06:08:51 kamada Exp $ */
/*
 * Copyright (C) 2003, 2004 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/types.h>

#include <netinet/in.h>			/* for htonl(), etc */

#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../lib/vmbuf.h"
#include "utils.h"
#include "bbkk.h"
#include "kink.h"
#include "handle.h"
#include "kink_fmt.h"
#include "kink_crypto.h"		/* XXX not from this file */


extern int f_debug_crypt;
extern int f_debug_payload;


/*
 * legend
 *   |  boundary between payloads
 *          (KINK payloads with +, inner payloads without +)
 *   :  boundary between payload header and body
 *   N  (Inner)NextPayload field
 *
 * usual
 *    KINK header
 *    +------+-----------+------------+------------------
 *    |      |   :       |   :        |
 *    |      |N  :       |N  :        |
 *    +------+-----------+------------+--------------
 *    ^                   ^           ^
 *    buf             nptype_loc      p
 *
 * when preparing the contents of KINK_ENCRYPT
 *
 *                      KINK_ENCRYPT header
 *    KINK header        |  KINK_ENCRYPT body
 *    |                  |   |InnerNextPayload field
 *    v                  v   v|
 *    +------+-----------+----|------------------------------
 *    |KINK  |   :       |   :v|   :      |
 *    |header|   :       |N  :N|N  :      |
 *    +------+-----------+-------------------------------
 *    ^                  ^   ^  ^         ^
 *    buf           encrypt  |  |         p
 *                  encrypt->b  nptype_loc
 */
struct payload_setter {
	vchar_t *buf;
	char *p;			/* next payload begins here */
	uint16_t *length_loc;		/* length field of the base header */
	uint8_t *nptype_loc;		/* nptype field of the last payload */
	void *cksum_ptr;

	struct kink_pl_encrypt *encrypt;
};


static int kink_decode_check_generic(vchar_t *packet);
static int kink_decode_payload_list(struct kink_handle *kh,
    struct kink_payload *p, int nptype, size_t remlen, int in_encrypt);

static vchar_t *make_kink_ap_req(struct kink_handle *kh);
static vchar_t *make_kink_ap_rep(struct kink_handle *kh);
static vchar_t *make_kink_isakmp(struct kink_handle *kh);
static vchar_t *make_kink_error(struct kink_handle *kh);
static vchar_t *make_krb_error(struct kink_handle *kh, int32_t bbkkret);

static int setpl_kink_header(struct payload_setter *ps,
    size_t size, struct kink_handle *kh, int type, int need_cksum);
static void setpl_kink_payload(struct payload_setter *ps,
    vchar_t *data, int nptype);
static void setpl_begin_encrypt_here(struct payload_setter *ps);
static vchar_t *setpl_finalize(struct payload_setter *ps,
    struct kink_handle *kh);
static void setpl_cancel(struct payload_setter *ps);
static int setpl_do_encrypt(struct payload_setter *ps,
    struct kink_handle *kh);
static vchar_t *get_kink_payload(void *ptr);



vchar_t *
kink_encode_create(struct kink_handle *kh)
{
	struct payload_setter ps;
	vchar_t *ap_req, *isakmp;
	size_t size;
	int need_cksum;

	ap_req = isakmp = NULL;
	need_cksum = 1;

	/*
	 * make payloads
	 */
	if ((ap_req = make_kink_ap_req(kh)) == NULL)
		goto fail;
	if ((isakmp = make_kink_isakmp(kh)) == NULL)
		goto fail;

	/*
	 * set header & payloads
	 */
	size = sizeof(struct kink_payload) + ap_req->l +
	    sizeof(struct kink_payload) + isakmp->l;

	if (setpl_kink_header(&ps, size, kh, KINK_MSGTYPE_CREATE,
	    need_cksum) != 0)
		goto fail;
	setpl_kink_payload(&ps, ap_req, KINK_NPTYPE_AP_REQ);
	setpl_begin_encrypt_here(&ps);
	setpl_kink_payload(&ps, isakmp, KINK_NPTYPE_ISAKMP);
	if (setpl_do_encrypt(&ps, kh) != 0) {
		setpl_cancel(&ps);
		goto fail;
	}

	vfree(ap_req);
	vfree(isakmp);

	return setpl_finalize(&ps, kh);

fail:
	vfree(ap_req);		/* vfree() accepts NULL */
	vfree(isakmp);
	return NULL;
}

vchar_t *
kink_encode_delete(struct kink_handle *kh)
{
	struct payload_setter ps;
	vchar_t *ap_req, *isakmp;
	size_t size;
	int need_cksum;

	ap_req = isakmp = NULL;
	need_cksum = 1;

	/*
	 * make payloads
	 */
	if ((ap_req = make_kink_ap_req(kh)) == NULL)
		goto fail;
	if ((isakmp = make_kink_isakmp(kh)) == NULL)
		goto fail;

	/*
	 * set header & payloads
	 */
	size = sizeof(struct kink_payload) + ap_req->l +
	    sizeof(struct kink_payload) + isakmp->l;

	if (setpl_kink_header(&ps, size, kh, KINK_MSGTYPE_DELETE,
	    need_cksum) != 0)
		goto fail;
	setpl_kink_payload(&ps, ap_req, KINK_NPTYPE_AP_REQ);
	setpl_begin_encrypt_here(&ps);
	setpl_kink_payload(&ps, isakmp, KINK_NPTYPE_ISAKMP);
	if (setpl_do_encrypt(&ps, kh) != 0) {
		setpl_cancel(&ps);
		goto fail;
	}

	vfree(ap_req);
	vfree(isakmp);

	return setpl_finalize(&ps, kh);

fail:
	vfree(ap_req);		/* vfree() accepts NULL */
	vfree(isakmp);
	return NULL;
}

vchar_t *
kink_encode_ack(struct kink_handle *kh)
{
	struct payload_setter ps;
	vchar_t *ap_req;
	size_t size;
	int need_cksum;

	ap_req = NULL;
	need_cksum = 1;

	/*
	 * make payloads
	 */
	if ((ap_req = make_kink_ap_req(kh)) == NULL)
		goto fail;

	/*
	 * set header & payloads
	 */
	size = sizeof(struct kink_payload) + ap_req->l;

	if (setpl_kink_header(&ps, size, kh, KINK_MSGTYPE_ACK,
	    need_cksum) != 0)
		goto fail;
	setpl_kink_payload(&ps, ap_req, KINK_NPTYPE_AP_REQ);

	vfree(ap_req);

	return setpl_finalize(&ps, kh);

fail:
	vfree(ap_req);
	return NULL;
}

vchar_t *
kink_encode_status(struct kink_handle *kh)
{
	struct payload_setter ps;
	vchar_t *ap_req, *isakmp;
	size_t size;
	int need_cksum;

	ap_req = isakmp = NULL;
	need_cksum = 1;

	/*
	 * make payloads
	 */
	if ((ap_req = make_kink_ap_req(kh)) == NULL)
		goto fail;
	if (kh->isakmp != NULL &&
	    (isakmp = make_kink_isakmp(kh)) == NULL)
		goto fail;

	/*
	 * set header & payloads
	 */
	size = sizeof(struct kink_payload) + ap_req->l;
	if (isakmp != NULL)
		size += sizeof(struct kink_payload) + isakmp->l;

	if (setpl_kink_header(&ps, size, kh, KINK_MSGTYPE_STATUS,
	    need_cksum) != 0)
		goto fail;
	setpl_kink_payload(&ps, ap_req, KINK_NPTYPE_AP_REQ);
	if (isakmp != NULL) {
		setpl_begin_encrypt_here(&ps);
		setpl_kink_payload(&ps, isakmp, KINK_NPTYPE_ISAKMP);
		if (setpl_do_encrypt(&ps, kh) != 0) {
			setpl_cancel(&ps);
			goto fail;
		}
	}

	vfree(ap_req);
	vfree(isakmp);

	return setpl_finalize(&ps, kh);

fail:
	vfree(ap_req);          /* vfree() accepts NULL */
	vfree(isakmp);
	return NULL;
}

vchar_t *
kink_encode_reply(struct kink_handle *kh)
{
	struct payload_setter ps;
	size_t size;
	vchar_t *ap_rep, *isakmp;
	int need_cksum;

	ap_rep = isakmp = NULL;
	need_cksum = 1;

	/*
	 * make payloads
	 */
	if ((ap_rep = make_kink_ap_rep(kh)) == NULL)
		goto fail;
	/* XXX check in_isakmp here, or in make_kink_isakmp()? */
	if (kh->in_isakmp != NULL &&
	    (isakmp = make_kink_isakmp(kh)) == NULL)
		goto fail;

	/*
	 * set header & payloads
	 */
	size = sizeof(struct kink_payload) + ap_rep->l;
	if (isakmp != NULL)
		size += sizeof(struct kink_payload) + isakmp->l;

	if (setpl_kink_header(&ps, size, kh, KINK_MSGTYPE_REPLY,
	    need_cksum) != 0)
		goto fail;
	setpl_kink_payload(&ps, ap_rep, KINK_NPTYPE_AP_REP);
	if (isakmp != NULL) {
		setpl_begin_encrypt_here(&ps);
		setpl_kink_payload(&ps, isakmp, KINK_NPTYPE_ISAKMP);
		if (setpl_do_encrypt(&ps, kh) != 0) {
			setpl_cancel(&ps);
			goto fail;
		}
	}

	vfree(ap_rep);
	vfree(isakmp);

	return setpl_finalize(&ps, kh);

fail:
	vfree(ap_rep);		/* vfree() accepts NULL */
	vfree(isakmp);
	return NULL;
}



vchar_t *
kink_encode_reply_kink_error(struct kink_handle *kh)
{
	struct payload_setter ps;
	size_t size;
	vchar_t *ap_rep, *error;
	int need_cksum;

	if (kh->error_code == 0) {
		kinkd_log(KLLV_FATAL,
		    "INTERNAL ERROR: making KINK_ERROR while no error\n");
		abort();
	}

	ap_rep = error = NULL;
	need_cksum = kh->krb_ap_rep != NULL;

	/*
	 * make payloads
	 */
	if (kh->krb_ap_rep != NULL &&
	    (ap_rep = make_kink_ap_rep(kh)) == NULL)
		goto fail;
	if ((error = make_kink_error(kh)) == NULL)
		goto fail;

	/*
	 * set header & payloads
	 */
	size = sizeof(struct kink_payload) + error->l;
	if (ap_rep != NULL)
		size += sizeof(struct kink_payload) + ap_rep->l;

	if (setpl_kink_header(&ps, size, kh, KINK_MSGTYPE_REPLY,
	    need_cksum) != 0)
		goto fail;
	if (ap_rep != NULL) {
		setpl_kink_payload(&ps, ap_rep, KINK_NPTYPE_AP_REP);
		setpl_begin_encrypt_here(&ps);
	}
	setpl_kink_payload(&ps, error, KINK_NPTYPE_ERROR);
	if (ap_rep != NULL && setpl_do_encrypt(&ps, kh) != 0) {
		setpl_cancel(&ps);
		goto fail;
	}

	vfree(ap_rep);
	vfree(error);

	return setpl_finalize(&ps, kh);

fail:
	vfree(ap_rep);		/* vfree() accepts NULL */
	vfree(error);
	return NULL;
}

vchar_t *
kink_encode_reply_krb_error(struct kink_handle *kh, int32_t bbkkret)
{
	struct payload_setter ps;
	size_t size;
	vchar_t *ap_rep, *krb_error;
	int need_cksum;

	ap_rep = krb_error = NULL;
	need_cksum = kh->krb_ap_rep != NULL;
	kinkd_log(KLLV_INFO, "need_cksum: %d\n", need_cksum);

	/*
	 * make payloads
	 */
	/* this condition should be (kh->krb_ap_rep != NULL)? */
	if (kh->auth_context != NULL &&
	    (ap_rep = make_kink_ap_rep(kh)) == NULL)
		goto fail;
	if ((krb_error = make_krb_error(kh, bbkkret)) == NULL)
		goto fail;

	/*
	 * set header & payloads
	 */
	size = sizeof(struct kink_payload) + krb_error->l;
	if (ap_rep != NULL)
		size += sizeof(struct kink_payload) + ap_rep->l;

	if (setpl_kink_header(&ps, size, kh, KINK_MSGTYPE_REPLY,
	    need_cksum) != 0)
		goto fail;
	if (ap_rep != NULL) {
		setpl_kink_payload(&ps, ap_rep, KINK_NPTYPE_AP_REP);
		setpl_begin_encrypt_here(&ps);
	}
	setpl_kink_payload(&ps, krb_error, KINK_NPTYPE_KRB_ERROR);
	if (ap_rep != NULL && setpl_do_encrypt(&ps, kh) != 0) {
		setpl_cancel(&ps);
		goto fail;
	}

	vfree(ap_rep);
	vfree(krb_error);

	return setpl_finalize(&ps, kh);

fail:
	vfree(ap_rep);		/* vfree() accepts NULL */
	vfree(krb_error);
	return NULL;
}



int
kink_decode_generic(struct kink_handle *kh, vchar_t *packet)
{
	struct kink_header *kheader;
	struct kink_payload *p;
	size_t remlen, cksum_len_pad;
	int nptype;

	/*
	 * If program reached here, packet->l must not be less than
	 * sizeof(*kheader).
	 * (kink_decode_get_msgtype() should have been called.)
	 */
	kheader = (struct kink_header *)packet->v;

	/* XID should be retrieved first, to REPLY with KINK_ERROR */
	kh->xid = ntohl(kheader->xid);
	kh->flags = kheader->flags;

	kh->error_code = kink_decode_check_generic(packet);
	if (kh->error_code != 0)
		return 1;

	cksum_len_pad = ALIGN(kheader->cksum_len, 4);
	remlen = packet->l - sizeof(*kheader) - cksum_len_pad;
	p = (struct kink_payload *)
	    (packet->v + sizeof(*kheader) + cksum_len_pad);
	nptype = kheader->next_payload;

	return kink_decode_payload_list(kh, p, nptype, remlen, 0);
}

int
kink_decode_kink_encrypt(struct kink_handle *kh)
{
	struct kink_pl_encrypt_b *p;
	void *dec_ptr;
	size_t dec_len;
	int32_t bbkkret;
	int ret;

	if (kh->encrypt == NULL)
		return 0;
	if (kh->auth_context == NULL) {
		kinkd_log(KLLV_ERR, "KINK_ENCRYPT without auth_context\n");
		return 1;
	}

	if (f_debug_payload)
		kinkd_log(KLLV_DEBUG, "decoding KINK_ENCRYPT\n");

	/* decode */
	bbkkret = bbkk_decrypt(kh->g->context, kh->auth_context,
	    kh->encrypt->v, kh->encrypt->l,
	    &dec_ptr, &dec_len);
	if (bbkkret != 0) {
		kinkd_log(KLLV_ERR, "failed to decrypt KINK_ENCRYPT: %s\n",
		    bbkk_get_err_text(kh->g->context, bbkkret));
		return 1;
	}
	if (f_debug_crypt) {
		kinkd_log(KLLV_DEBUG, "KINK_ENCRYPT after decryption\n");
		kinkd_log_dump(KLLV_DEBUG, dec_ptr, dec_len);
	}

	p = (struct kink_pl_encrypt_b *)dec_ptr;
	ret = kink_decode_payload_list(kh,
	    (struct kink_payload *)((char *)p + sizeof(*p)),
	    p->in_nptype,
	    dec_len - sizeof(*p), 1);
	free(dec_ptr);
	return ret;
}



/*
 * return KINK_ERROR error code, or 0
 */
static int
kink_decode_check_generic(vchar_t *packet)
{
	struct kink_header *kheader;

	kheader = (struct kink_header *)packet->v;

	if (packet->l < sizeof(*kheader)) {
		kinkd_log(KLLV_INFO,
		    "received packet too short (KINK_PROTOERR)\n");
		return KINK_ERR_PROTOERR;
	}
	if (packet->l < ntohs(kheader->length)) {
		kinkd_log(KLLV_INFO,
		    "packet len < len in header (KINK_PROTOERR)\n");
		return KINK_ERR_PROTOERR;
	}
	if (packet->l > ntohs(kheader->length))
		kinkd_log(KLLV_INFO,
		    "longer packet length than one in the header\n");
	if (packet->l < sizeof(*kheader) + ALIGN(kheader->cksum_len, 4)) {
		kinkd_log(KLLV_INFO,
		    "shorter packet than checksum length (KINK_PROTOERR)\n");
		return KINK_ERR_PROTOERR;
	}

	if ((kheader->ver >> 4) != KINK_MAJOR_VERSION) {
		kinkd_log(KLLV_INFO,
		    "Invalid Major Version (KINK_INVMAJ)\n");
		return KINK_ERR_INVMAJ;
	}
	if ((kheader->ver & 0x0f) != KINK_MINOR_VERSION) {
		kinkd_log(KLLV_INFO,
		    "Invalid Minor Version (KINK_INVMIN)\n");
		return KINK_ERR_INVMIN;
	}
	if (ntohl(kheader->doi) != 1) {		/* XXX IPsec DOI */
		kinkd_log(KLLV_INFO,
		    "Invalid DOI (KINK_INVDOI)\n");
		return KINK_ERR_INVDOI;
	}
	if ((kheader->flags & ~KINK_FLAG_ACKREQ) != 0 ||
	    kheader->reserved != 0)
		kinkd_log(KLLV_INFO,
		    "Unknown flag with %s\n", kink_msgtype2str(kheader->type));

	return 0;
}

static int
kink_decode_payload_list(struct kink_handle *kh,
    struct kink_payload *p, int nptype, size_t remlen, int in_encrypt)
{
	int flag_kink_encrypt;
	size_t len, padding;

	flag_kink_encrypt = 0;
	padding = 0;
	while (nptype != KINK_NPTYPE_DONE) {
		if (flag_kink_encrypt)
			kinkd_log(KLLV_INFO, "Data after KINK_ENCRYPT\n");

		if (remlen < padding)
			goto fail_trunc;
		remlen -= padding;
		if (remlen < sizeof(*p) || remlen < ntohs(p->length))
			goto fail_trunc;
		len = ntohs(p->length);

		switch (nptype) {
		case KINK_NPTYPE_AP_REQ:
			if (f_debug_payload)
				kinkd_log(KLLV_DEBUG, "  KINK_AP_REQ\n");
			if (kh->ap_req == NULL)
				kh->ap_req = get_kink_payload(p);
			else
				kinkd_log(KLLV_INFO,
				    "duplicate KINK_AP_REQ; ignored\n");
			break;
		case KINK_NPTYPE_AP_REP:
			if (f_debug_payload)
				kinkd_log(KLLV_DEBUG, "  KINK_AP_REP\n");
			if (kh->ap_rep == NULL)
				kh->ap_rep = get_kink_payload(p);
			else
				kinkd_log(KLLV_INFO,
				    "duplicate KINK_AP_REP; ignored\n");
			break;
		case KINK_NPTYPE_KRB_ERROR:
			if (f_debug_payload)
				kinkd_log(KLLV_DEBUG, "  KINK_KRB_ERROR\n");
			if (kh->krb_error == NULL) {
				kh->krb_error = get_kink_payload(p);
				if (in_encrypt)
					kh->encrypted.krb_error = 1;
			} else
				kinkd_log(KLLV_INFO,
				    "duplicate KINK_KRB_ERROR; ignored\n");
			break;
		case KINK_NPTYPE_ISAKMP:
			if (f_debug_payload)
				kinkd_log(KLLV_DEBUG, "  KINK_ISAKMP\n");
			if (kh->isakmp == NULL) {
				kh->isakmp = get_kink_payload(p);
				if (in_encrypt)
					kh->encrypted.isakmp = 1;
			} else
				kinkd_log(KLLV_INFO,
				    "duplicate KINK_ISAKMP; "
				    "currently not supported, ignored\n");
			break;
		case KINK_NPTYPE_ENCRYPT:
			flag_kink_encrypt = 1;
			if (in_encrypt) {
				kinkd_log(KLLV_INFO,
				    "KINK_ENCRYPT in KINK_ENCRYPT; ignored\n");
				break;
			}
			if (f_debug_payload)
				kinkd_log(KLLV_DEBUG, "  KINK_ENCRYPT\n");
			if (kh->encrypt == NULL)
				kh->encrypt = get_kink_payload(p);
			else
				kinkd_log(KLLV_INFO,
				    "duplicate KINK_ENCRYPT; ignored\n");
			break;
		case KINK_NPTYPE_TGT_REQ:
			kinkd_log(KLLV_ERR,
			    "KINK_TGT_REQ not supported\n");
			kh->error_code = KINK_ERR_INTERR;
			return 1;
		case KINK_NPTYPE_TGT_REP:
			kinkd_log(KLLV_ERR,
			    "KINK_TGT_REP not supported\n");
			kh->error_code = KINK_ERR_INTERR;
			return 1;
		case KINK_NPTYPE_ERROR:
			if (f_debug_payload)
				kinkd_log(KLLV_DEBUG, "  KINK_ERROR\n");
			if (kh->error == NULL) {
				kh->error = get_kink_payload(p);
				if (in_encrypt)
					kh->encrypted.error = 1;
			} else
				kinkd_log(KLLV_INFO,
				    "duplicate KINK_ERROR; ignore\n");
			break;
		default:
			/* XXX KINK_PROTOERR? */
			kinkd_log(KLLV_INFO,
			    "unknown KINK payload type %d\n", nptype);
			kh->error_code = KINK_ERR_PROTOERR;
			return 1;
		}

		/* if (len == 0) then loops forever so adjust here */
		if (len < sizeof(*p))
			len = sizeof(*p);

		nptype = p->next_payload;
		p = (struct kink_payload *)ALIGN_PTR((char *)p + len, 4);
		remlen -= len;
		padding = ALIGN(len, 4) - len;
	}
	/*
	 * contents of KINK_ENCRYPT may be padded, so remaining byte
	 * does not necessarily mean an error.
	 */
	if (f_debug_payload)
		kinkd_log(KLLV_DEBUG,
		    "%d bytes remain in the packet\n", remlen);
	else if (remlen != 0 && !in_encrypt)
		kinkd_log(KLLV_INFO,
		    "%d bytes remain in the packet\n", remlen);

	return 0;

fail_trunc:
	kinkd_log(KLLV_ERR,
	    "truncated KINK payload in type %d (KINK_PROTOERR)\n", nptype);
	kh->error_code = KINK_ERR_PROTOERR;
	return 1;
}




int
kink_decode_get_msgtype(vchar_t *packet)
{
	struct kink_header *kheader;

	kheader = (struct kink_header *)packet->v;
	if (packet->l < sizeof(*kheader)) {
		kinkd_log(KLLV_INFO, "received packet is too short\n");
		return -1;
	}
	return kheader->type;
}

ssize_t
kink_decode_get_msglen(void *ptr, size_t len)
{
	struct kink_header *kheader;

	if (len < sizeof(*kheader)) {
		kinkd_log(KLLV_INFO, "received packet is too short\n");
		return -1;
	}
	kheader = (struct kink_header *)ptr;
	return ntohs(kheader->length);
}

unsigned int
kink_decode_get_xid(vchar_t *packet)
{
	struct kink_header *kheader;

	kheader = (struct kink_header *)packet->v;
	if (packet->l < sizeof(*kheader)) {
		/*
		 * This function is always called after
		 * kink_decode_get_msgtype(), so NOTREACHED here.
		 */
		kinkd_log(KLLV_FATAL,
		    "INTERNAL ERROR: received packet is too short\n");
		abort();
	}
	return ntohl(kheader->xid);
}

void *
kink_decode_get_cksum_ptr(vchar_t *packet)
{
	struct kink_header *kheader;

	kheader = (struct kink_header *)packet->v;
	/* packet->l >= sizeof(*kheader) is already guaranteed */
	if (packet->l < sizeof(*kheader) + kheader->cksum_len) {
		kinkd_log(KLLV_INFO, "received packet is too short\n");
		return NULL;
	}
	return packet->v + sizeof(*kheader);
}

size_t
kink_decode_get_cksum_len(vchar_t *packet)
{
	struct kink_header *kheader;

	kheader = (struct kink_header *)packet->v;
	/* packet->l >= sizeof(*kheader) is already guaranteed */
	return kheader->cksum_len;
}





/* XXX these make_*() functions should receive pointer and len? */
static vchar_t *
make_kink_ap_req(struct kink_handle *kh)
{
	vchar_t *buf;
	struct kink_pl_ap_req_b *p;

	if ((buf = vmalloc(sizeof(*p) + kh->krb_ap_req->l)) == NULL) {
		kinkd_log(KLLV_FATAL, "out of memory\n");
		exitreq = EXITREQ_NOMEM;
		return NULL;
	}
	p = (struct kink_pl_ap_req_b *)buf->v;

	p->epoch = htonl(kh->g->epoch);
	memcpy(buf->v + sizeof(*p), kh->krb_ap_req->v, kh->krb_ap_req->l);

	return buf;
}

static vchar_t *
make_kink_ap_rep(struct kink_handle *kh)
{
	vchar_t *buf;
	struct kink_pl_ap_rep_b *p;

	if ((buf = vmalloc(sizeof(*p) + kh->krb_ap_rep->l)) == NULL) {
		kinkd_log(KLLV_FATAL, "out of memory\n");
		exitreq = EXITREQ_NOMEM;
		return NULL;
	}
	p = (struct kink_pl_ap_rep_b *)buf->v;

	p->epoch = htonl(kh->g->epoch);
	memcpy(buf->v + sizeof(*p), kh->krb_ap_rep->v, kh->krb_ap_rep->l);

	return buf;
}

static vchar_t *
make_kink_isakmp(struct kink_handle *kh)
{
	vchar_t *buf;
	struct kink_pl_isakmp_b *p;

	if ((buf = vmalloc(sizeof(*p) + kh->in_isakmp->l)) == NULL) {
		kinkd_log(KLLV_FATAL, "out of memory\n");
		exitreq = EXITREQ_NOMEM;
		return NULL;
	}
	p = (struct kink_pl_isakmp_b *)buf->v;

	p->in_nptype = kh->isakmp_1sttype;
	p->qm_ver = KINK_QM_VERSION;
	p->reserved = 0;
	memcpy(buf->v + sizeof(*p), kh->in_isakmp->v, kh->in_isakmp->l);

	return buf;
}

static vchar_t *
make_kink_error(struct kink_handle *kh)
{
	vchar_t *buf;
	struct kink_pl_error_b *p;

	if ((buf = vmalloc(sizeof(*p))) == NULL) {
		kinkd_log(KLLV_FATAL, "out of memory\n");
		exitreq = EXITREQ_NOMEM;
		return NULL;
	}
	p = (struct kink_pl_error_b *)buf->v;

	p->error_code = htonl(kh->error_code);

	return buf;
}

static vchar_t *
make_krb_error(struct kink_handle *kh, int32_t bbkkret)
{
	vchar_t *buf;
	void *krb_error;
	size_t krb_error_len;
	int32_t my_bbkkret;

	my_bbkkret = bbkk_make_error(kh->g->context, bbkkret,
	    &krb_error, &krb_error_len);
	if (my_bbkkret != 0) {
		kinkd_log(KLLV_ERR,
		    "make_krb_error() failed: %s\n",
		    bbkk_get_err_text(kh->g->context, my_bbkkret));
		return NULL;
	}

	if ((buf = vmalloc(krb_error_len)) == NULL) {
		kinkd_log(KLLV_FATAL, "out of memory\n");
		exitreq = EXITREQ_NOMEM;
		free(krb_error);
		return NULL;
	}
	memcpy(buf->v, krb_error, krb_error_len);
	free(krb_error);

	return buf;
}



int
read_kink_ap_req(struct kink_handle *kh, vchar_t *buf)
{
	struct kink_pl_ap_req_b *p;

	if (buf->l < sizeof(*p)) {
		kinkd_log(KLLV_INFO, "too short KINK_AP_REQ\n");
		return 1;
	}
	p = (struct kink_pl_ap_req_b *)buf->v;

	kh->recv_epoch = ntohl(p->epoch);
	/* XXX copying KRB_AP_REQ is very wasteful */
	kh->krb_ap_req = vmalloc(buf->l - sizeof(*p));
	if (kh->krb_ap_req == NULL) {
		kinkd_log(KLLV_FATAL, "out of memory\n");
		exitreq = EXITREQ_NOMEM;
		return 1;
	}
	memcpy(kh->krb_ap_req->v, buf->v + sizeof(*p), buf->l - sizeof(*p));
	return 0;
}

int
read_kink_ap_rep(struct kink_handle *kh, vchar_t *buf)
{
	struct kink_pl_ap_rep_b *p;

	if (buf->l < sizeof(*p)) {
		kinkd_log(KLLV_INFO, "too short KINK_AP_REP\n");
		return 1;
	}
	p = (struct kink_pl_ap_rep_b *)buf->v;

	kh->recv_epoch = ntohl(p->epoch);
	/* XXX copying KRB_AP_REP is very wasteful */
	kh->krb_ap_rep = vmalloc(buf->l - sizeof(*p));
	if (kh->krb_ap_rep == NULL) {
		kinkd_log(KLLV_FATAL, "out of memory\n");
		exitreq = EXITREQ_NOMEM;
		return 1;
	}
	memcpy(kh->krb_ap_rep->v, buf->v + sizeof(*p), buf->l - sizeof(*p));
	return 0;
}

int
read_kink_isakmp(struct kink_handle *kh, vchar_t *buf)
{
	struct kink_pl_isakmp_b *p;

	if (buf->l < sizeof(*p)) {
		kinkd_log(KLLV_INFO, "too short KINK_ISAKMP\n");
		return 1;
	}
	p = (struct kink_pl_isakmp_b *)buf->v;
	if (p->qm_ver != KINK_QM_VERSION) {
		kinkd_log(KLLV_ERR,
		    "unsupported Quick Mode version 0x%02x\n", p->qm_ver);
		kh->error_code = KINK_ERR_BADQMVERS;
		return 1;
	}
	kh->isakmp_1sttype = p->in_nptype;
	kh->in_isakmp = vmalloc(buf->l - sizeof(*p));
	if (kh->in_isakmp == NULL) {
		kinkd_log(KLLV_FATAL, "out of memory\n");
		exitreq = EXITREQ_NOMEM;
		return 1;
	}
	memcpy(kh->in_isakmp->v, buf->v + sizeof(*p), buf->l - sizeof(*p));
	return 0;
}

int
read_kink_error(uint32_t *error_code, vchar_t *buf)
{
	struct kink_pl_error_b *p;

	if (buf->l < sizeof(*p)) {
		kinkd_log(KLLV_INFO, "too short KINK_ERROR\n");
		return 1;
	}
	p = (struct kink_pl_error_b *)buf->v;

	*error_code = ntohl(p->error_code);
	return 0;
}



/*
 * size does not include base header and checksum.
 */
static int
setpl_kink_header(struct payload_setter *ps,
    size_t size, struct kink_handle *kh, int type, int need_cksum)
{
	struct kink_header *kheader;
	size_t cksum_len;

	cksum_len = 0;
	if (need_cksum &&
	    (cksum_len = kink_get_checksum_size(kh)) == (size_t)-1)
		return 1;

	/* space for base header and checksum */
	size += sizeof(struct kink_header) + ALIGN(cksum_len, 4);
	/* XXX */
	size +=
	    3 * 3 +				/* 4-octet padding */
	    sizeof(struct kink_pl_encrypt) +	/* encrypt */
	    64 + 16;				/* enc overhead & padding */

	if ((ps->buf = vmalloc(size)) == NULL) {
		kinkd_log(KLLV_FATAL, "out of memory\n");
		exitreq = EXITREQ_NOMEM;
		return 1;
	}

	ps->p = ps->buf->v;

	kheader = (struct kink_header *)ps->p;
	kheader->type = type;
	kheader->ver = KINK_VERSION;
	kheader->length = 0;				/* calculated later */
	kheader->doi = htonl(1);	/* XXX IPsec DOI */
	kheader->xid = htonl(kh->xid);
	kheader->cksum_len = cksum_len;
	kheader->next_payload = KINK_NPTYPE_DONE;	/* provisional */
	kheader->flags = kh->flags;
	kheader->reserved = 0;

	ps->length_loc = &kheader->length;
	ps->nptype_loc = &kheader->next_payload;
	if (need_cksum)
		ps->cksum_ptr = ps->p + sizeof(struct kink_header);
	else
		ps->cksum_ptr = NULL;
	ps->encrypt = NULL;

	ps->p += sizeof(struct kink_header) + ALIGN(cksum_len, 4);
	return 0;
}

static void
setpl_kink_payload(struct payload_setter *ps,
    vchar_t *data, int nptype)
{
	struct kink_payload *payload;

	/* sanity check */
	if ((char *)ALIGN_PTR(ps->p, 4) + sizeof(*payload) + data->l >
	    ps->buf->v + ps->buf->l) {
		kinkd_log(KLLV_FATAL, "buffer overflow\n");
		abort();
	}

	/* align 4-octet boundary */
	while ((intptr_t)ps->p % 4 != 0)
		*ps->p++ = '\0';

	*ps->nptype_loc = nptype;
	payload = (struct kink_payload *)ps->p;
	ps->nptype_loc = &payload->next_payload;

	payload->next_payload = KINK_NPTYPE_DONE;	/* provisional */
	payload->reserved = 0;
	payload->length = htons(sizeof(*payload) + data->l);
	memcpy(ps->p + sizeof(*payload), data->v, data->l);

	ps->p += sizeof(*payload) + data->l;
}

static void
setpl_begin_encrypt_here(struct payload_setter *ps)
{
	struct kink_pl_encrypt *encrypt;

	/* sanity check */
	if ((char *)ALIGN_PTR(ps->p, 4) + sizeof(*encrypt) >
	    ps->buf->v + ps->buf->l) {
		kinkd_log(KLLV_FATAL, "buffer overflow\n");
		abort();
	}

	/* align 4-octet boundary */
	while ((intptr_t)ps->p % 4 != 0)
		*ps->p++ = '\0';

	*ps->nptype_loc = KINK_NPTYPE_ENCRYPT;

	/* set KINK_ENCRYPT as a normal payload */
	encrypt = (struct kink_pl_encrypt *)ps->p;
	encrypt->h.next_payload = KINK_NPTYPE_DONE;
	encrypt->h.reserved = 0;
	encrypt->h.length = 0;
	encrypt->b.in_nptype = KINK_NPTYPE_DONE;
	encrypt->b.reserved1 = 0;
	encrypt->b.reserved2 = 0;

	/* preserve the location of KINK_ENCRYPT header and continue */
	ps->nptype_loc = &encrypt->b.in_nptype;
	ps->encrypt = encrypt;
	ps->p += sizeof(*encrypt);
}

static int
setpl_do_encrypt(struct payload_setter *ps, struct kink_handle *kh)
{
	void *enc_ptr;
	size_t enc_len;
	int32_t bbkkret;

	if (ps->encrypt == NULL) {
		kinkd_log(KLLV_FATAL,
		    "INTERNAL ERROR: encryption header has not been set\n");
		abort();
	}

	/*
	 * encrypt region:  &(ps->encrypt.b) -- ps->p
	 */
	/* encrypt */
	bbkkret = bbkk_encrypt(kh->g->context, kh->auth_context,
	    &ps->encrypt->b, ps->p - (char *)&ps->encrypt->b,
	    &enc_ptr, &enc_len);
	if (bbkkret != 0) {
		kinkd_log(KLLV_ERR,
		    "bbkk_encrypt: %s\n",
		    bbkk_get_err_text(kh->g->context, bbkkret));
		return 1;
	}
	if (f_debug_crypt) {
		kinkd_log(KLLV_DEBUG, "KINK_ENCRYPT before encryption\n");
		kinkd_log_dump(KLLV_DEBUG,
		    &ps->encrypt->b, ps->p - (char *)&ps->encrypt->b);
	}

	/* sanity check */
	if ((char *)&ps->encrypt->b + enc_len >
	    ps->buf->v + ps->buf->l) {
		kinkd_log(KLLV_FATAL, "buffer overflow\n");
		free(enc_ptr);
		abort();
	}

	/* copy */
	memcpy(&ps->encrypt->b, enc_ptr, enc_len);
	ps->p = (char *)&ps->encrypt->b + enc_len;
	free(enc_ptr);

	/*
	 * complete KINK_ENCRYPT payload
	 */
	ps->encrypt->h.length = htons(ps->p - (char *)ps->encrypt);

	/*
	 * KINK_ENCRYPT MUST be the last payload so there is no
	 * need to restore nptype_loc.
	 */
	ps->nptype_loc = &ps->encrypt->h.next_payload;

	ps->encrypt = NULL;
	return 0;
}

static vchar_t *
setpl_finalize(struct payload_setter *ps, struct kink_handle *kh)
{
	size_t size;

	size = ps->p - ps->buf->v;

	*ps->length_loc = htons(size);
	ps->buf->l = size;

	if (ps->cksum_ptr != NULL &&
	    kink_calc_checksum(kh, ps->cksum_ptr, ps->buf) == NULL) {
		vfree(ps->buf);
		return NULL;
	}

	return ps->buf;
}

static void
setpl_cancel(struct payload_setter *ps)
{
	vfree(ps->buf);
}

/* payload->length MUST be < remaining length */
static vchar_t *
get_kink_payload(void *ptr)
{
	struct kink_payload *payload;
	vchar_t *buf;
	size_t len;

	payload = (struct kink_payload *)ptr;
	len = ntohs(payload->length);
	if (len < sizeof(*payload)) {
		kinkd_log(KLLV_WARNING, "too short KINK payload\n");
		return NULL;
	}
	if ((buf = vmalloc(len - sizeof(*payload))) == NULL) {
		kinkd_log(KLLV_FATAL, "out of memory\n");
		exitreq = EXITREQ_NOMEM;
		return NULL;
	}
	memcpy(buf->v, (char *)ptr + sizeof(*payload), buf->l);
	return buf;
}



/*
 * stringifiers
 */

const char *
kink_msgtype2str(int msgtype)
{
	switch (msgtype) {
	case KINK_MSGTYPE_RESERVED:
		return "RESERVED";
	case KINK_MSGTYPE_CREATE:
		return "CREATE";
	case KINK_MSGTYPE_DELETE:
		return "DELETE";
	case KINK_MSGTYPE_REPLY:
		return "REPLY";
	case KINK_MSGTYPE_GETTGT:
		return "GETTGT";
	case KINK_MSGTYPE_ACK:
		return "ACK";
	case KINK_MSGTYPE_STATUS:
		return "STATUS";
	default:
		return "UnknownMsgType";
	}
}
