/* $Id: main.c,v 1.51 2005/01/19 00:53:47 kamada Exp $ */
/*
 * Copyright (C) 2003-2005 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "config.h"

#include <errno.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "racoon.h"
#include "pathnames.h"
#include "utils.h"
#include "crypto_openssl.h"
#include "bbkk.h"
#include "kink_conf.h"
#include "handle.h"
#include "etchosts.h"
#include "session.h"


static const char kinkd_version[] = "[i-d=06, compat=6]";


static void usage(void);
static void version(void);


/* global options */
int f_foreground = 0;
int f_loglevel = KLLV_BASE;
char *config_file = RACOON2_CONFIG_FILE;

int f_debug_krb5 = 0;
int f_debug_ticketing = 0;
int f_debug_packet = 0;
int f_debug_crypt = 0;
int f_debug_peer = 0;
int f_debug_payload = 0;
int f_debug_pfkey = 0;


int
main(int argc, char *argv[])
{
	struct kink_global kg;
	int32_t bbkkret;
	int c;

	(void)get_myname(argv[0]);
	while ((c = getopt(argc, argv, "df:hD:FV")) != -1) {
		switch (c) {
		case 'd':
			f_loglevel = KLLV_DEBUG;
			break;
		case 'f':
			config_file = optarg;
			break;
		case 'h':
			usage();
			/* NOTREACHED */
		case 'D':
			if (strcmp(optarg, "krb5") == 0)
				f_debug_krb5 = 1;
			else if (strcmp(optarg, "ticketing") == 0)
				f_debug_ticketing = 1;
			else if (strcmp(optarg, "packet") == 0)
				f_debug_packet = 1;
			else if (strcmp(optarg, "crypt") == 0)
				f_debug_crypt = 1;
			else if (strcmp(optarg, "peer") == 0)
				f_debug_peer = 1;
			else if (strcmp(optarg, "payload") == 0)
				f_debug_payload = 1;
			else if (strcmp(optarg, "pfkey") == 0)
				f_debug_pfkey = 1;
			else {
				usage();
				/* NOTREACHED */
			}
			break;
		case 'F':
			f_foreground = 1;
			break;
		case 'V':
			version();
			/* NOTREACHED */
		default:
			usage();
			/* NOTREACHED */
		}
	}
	argc -= optind;
	argv += optind;
	if (argc != 0) {
		usage();
		/* NOTREACHED */
	}

	if (rbuf_init(8, 80, 4, 160, 4) == -1) {
		kinkd_log(KLLV_ERR, "rbuf_init() failed\n");
		exit(1);
	}
	plog_init(
	    f_loglevel < KLLV_DEBUG ? RCT_LOGMODE_NORMAL : RCT_LOGMODE_DEBUG,
	    NULL, get_myname(NULL) /* argv[0] */, 1, f_foreground);

	kinkd_log(KLLV_NOTICE,
	    "kinkd racoon2-%s %s\n", rc_version(), kinkd_version);
	kinkd_log(KLLV_NOTICE, "linked with %s, %s.\n",
	    bbkk_libversion(), crypto_libversion());

	if (rcf_read(config_file, 0) == -1) {
		kinkd_log(KLLV_ERR, "rcf_read() failed\n");
		exit(1);
	}
	load_etchosts();

	/*
	 * initialize global state
	 */
	kg.my_principal = get_default_principal();
	if (bbkk_init(&kg.context, kg.my_principal) != 0) {
		fprintf(stderr, "%s: failed to bbkk_init\n", get_myname(NULL));
		exit(1);
	}
	if ((bbkkret = bbkk_get_tgt(kg.context, kg.my_principal)) != 0) {
		fprintf(stderr, "%s: bbkk_get_tgt: %s\n",
		    get_myname(NULL), bbkk_get_err_text(kg.context, bbkkret));
		exit(1);
	}
	kg.epoch = time(NULL);
	kg.next_xid = 0;
	TAILQ_INIT(&kg.schedq);
	SLIST_INIT(&kg.peerlist);
	SLIST_INIT(&kg.deadpeerlist);
	kg.fd_pfkey = -1;
	kg.fd_rcnd = -1;

	if (!f_foreground && daemon(0, 0) == -1) {
		fprintf(stderr, "%s: daemon: %s\n",
		    get_myname(NULL), strerror(errno));
		exit(1);
	}

	session(&kg);
	exit(0);
}

static void
usage(void)
{
	printf("usage: %s [-dhFV] [-f config]\n"
	    "   -d  output debugging message.\n"
	    "   -h  help (this message).\n"
	    "   -F  run in foreground.\n"
	    "   -V  version.\n"
	    "   -f  specify the configuration file.\n",
	    get_myname(NULL));
	exit(1);
}

static void
version(void)
{
	printf("kinkd racoon2-%s %s\n", rc_version(), kinkd_version);
	printf("  linked with %s, %s.\n",
	    bbkk_libversion(), crypto_libversion());
	exit(1);
}
