/* $Id: scheduler.c,v 1.38 2004/12/15 07:36:33 kamada Exp $ */
/*
 * Copyright (C) 2003, 2004 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "config.h"

#include <sys/types.h>
#include <sys/time.h>

#ifdef MAKE_KINK_LIST_FILE
# include <errno.h>
# include <fcntl.h>
#endif
#include <inttypes.h>
#include <stdio.h>
#ifdef MAKE_KINK_LIST_FILE
# include <string.h>
# include <unistd.h>
#endif

#include "../lib/vmbuf.h"
#include "../lib/rc_type.h"
#include "../lib/rc_net.h"
#ifdef MAKE_KINK_LIST_FILE
# include "pathnames.h"
#endif
#include "utils.h"
#include "sockmisc.h"
#include "handle.h"
#include "peer.h"
#include "scheduler.h"
#include "proposal.h"


static void sched_insert(struct kink_handle *kh);
static struct kink_handle *sched_get_first(struct kink_global *kg);

#ifdef MAKE_KINK_LIST_FILE
static void make_list_file(struct kink_global *kg);
#endif


struct timeval *
scheduler(struct kink_global *kg, struct timeval *tv)
{
	struct kink_handle *kh;
	struct timeval cur;
	int ret;

#ifdef MAKE_KINK_LIST_FILE
	/* XXX too waste to make it each time */
	make_list_file(kg);
#endif

	for (;;) {
		kh = sched_get_first(kg);
		if (kh == NULL)
			return NULL;
		gettimeofday(&cur, NULL);
		timersub(&kh->next_retrans, &cur, tv);
		if (tv->tv_sec >= 0 && tv->tv_usec > 0)
			return tv;

		ret = (*kh->state->timeout_handler)(kh);
		if (ret != 0) {
			kinkd_log(KLLV_ERR,
			    "timeout callback failed: %s\n",
			    kh->state->strname);
			kinkd_log(KLLV_ERR,
			    "give up (p=%s)\n",
			    kh->peer != NULL ? kh->peer->remote_principal
			                     : "unknown");

			sched_unregister(kh);
			if (kh->state->cancel != NULL)
				(*kh->state->cancel)(kh);
			if (kh->ph2 != NULL)
				release_ph2(kh->ph2);
			release_handle(kh);
		}
	}
}

void
sched_register(struct kink_handle *kh)
{
	gettimeofday(&kh->next_retrans, NULL);
	kh->retrans_interval = kh->state->timer;

	sched_insert(kh);
}

void
sched_unregister(struct kink_handle *kh)
{
	TAILQ_REMOVE(&kh->g->schedq, kh, next_schedq);
}

/* XXX fixed back off */
void
sched_waitnext(struct kink_handle *kh)
{
	/* XXX gettimeofday here? */
	kh->retrans_interval *= 2;

	TAILQ_REMOVE(&kh->g->schedq, kh, next_schedq);
	sched_insert(kh);
}

void
sched_resettimer(struct kink_handle *kh)
{
	gettimeofday(&kh->next_retrans, NULL);
	kh->retrans_interval = kh->state->timer;

	TAILQ_REMOVE(&kh->g->schedq, kh, next_schedq);
	sched_insert(kh);
}

void
sched_resettimer_custom(struct kink_handle *kh, int sec)
{
	gettimeofday(&kh->next_retrans, NULL);
	kh->retrans_interval = sec;

	TAILQ_REMOVE(&kh->g->schedq, kh, next_schedq);
	sched_insert(kh);
}

struct kink_handle *
sched_get_by_kh(struct kink_global *kg, struct kink_handle *rkh)
{
	struct kink_handle *kh;

	TAILQ_FOREACH(kh, &kg->schedq, next_schedq) {
		if (kh == rkh)
			return kh;
	}
	return NULL;
}

/*
 * XXX
 * This is not a scheduler function, but kink_handle pool is currently
 * unified with scheduler.
 */
struct kink_handle *
sched_get_by_xid_side(struct kink_global *kg,
    uint32_t xid, int side)
{
	struct kink_handle *kh;

	TAILQ_FOREACH(kh, &kg->schedq, next_schedq) {
		if (kh->state->pseudo)
			continue;
		/*
		 * XXX STATUS and DELETE-responder does not have ph2...
		 * In such case, sched_get_by_xid_side is always called
		 * by initiators.  (But we want to implement DELETE-REPLY
		 * cache on responders, this will be a problem.)
		 */
		if (kh->ph2 == NULL) {
			/* XXX XXX */
#define INITIATOR 0
			if (kh->xid == xid && side == INITIATOR)
#undef INITIATOR
				return kh;
			continue;
		}
		if (kh->xid == xid &&
		    kh->ph2->side == side)
			return kh;
	}
	return NULL;
}

#ifdef CURRENTLY_NOT_USED
struct kink_handle *
sched_get_by_xid_side_ph2addr(struct kink_global *kg,
    uint32_t xid, int side, struct sockaddr *src, struct sockaddr *dst)
{
	struct kink_handle *kh;

	TAILQ_FOREACH(kh, &kg->schedq, next_schedq) {
		if (kh->state->pseudo || kh->ph2 == NULL)
			continue;
		if (kh->xid == xid &&
		    kh->ph2->side == side &&
		    rcs_cmpsa_wop(kh->ph2->src, src) == 0 &&
		    rcs_cmpsa_wop(kh->ph2->dst, dst) == 0)
			return kh;
	}
	return NULL;
}
#endif

struct kink_handle *
sched_get_by_xid_side_peer(struct kink_global *kg,
    uint32_t xid, int side, struct kink_peer *peer)
{
	struct kink_handle *kh;

	TAILQ_FOREACH(kh, &kg->schedq, next_schedq) {
		if (kh->state->pseudo || kh->ph2 == NULL)
			continue;
		if (kh->xid == xid &&
		    kh->ph2->side == side &&
		    kh->peer == peer)
			return kh;
	}
	return NULL;
}

struct kink_handle *
sched_get_by_saidx(struct kink_global *kg,
    struct sockaddr *src, struct sockaddr *dst,
    unsigned int proto_id, uint32_t spi)
{
	struct kink_handle *kh;

	TAILQ_FOREACH(kh, &kg->schedq, next_schedq) {
		if (kh->state->pseudo || kh->ph2 == NULL)
			continue;
		if (kh->ph2->approval != NULL) {
			if (match_saidx(kh->ph2->approval, proto_id, spi) == 0)
				return kh;
		} else if (kh->ph2->proposal != NULL) {
			if (match_saidx(kh->ph2->proposal, proto_id, spi) == 0)
				return kh;
		}
	}
	return NULL;
}

/* not by peer's contents but by pointer itself */
struct kink_handle *
sched_get_by_peer(struct kink_global *kg, struct kink_peer *peer)
{
	struct kink_handle *kh;

	TAILQ_FOREACH(kh, &kg->schedq, next_schedq) {
		if (kh->state->pseudo)
			continue;
		if (kh->peer == peer)
			return kh;
	}
	return NULL;
}



static void
sched_insert(struct kink_handle *kh)
{
	struct kink_handle *p;

	kh->next_retrans.tv_sec += kh->retrans_interval;

	TAILQ_FOREACH(p, &kh->g->schedq, next_schedq) {
		if (timercmp(&kh->next_retrans, &p->next_retrans, <))
			break;
	}
	if (p == NULL)
		/* queue is empty or all entries have shorter expiration */
		TAILQ_INSERT_TAIL(&kh->g->schedq, kh, next_schedq);
	else
		TAILQ_INSERT_BEFORE(p, kh, next_schedq);
}

static struct kink_handle *
sched_get_first(struct kink_global *kg)
{
	return TAILQ_FIRST(&kg->schedq);
}



void
print_kink_handles(struct kink_global *kg)
{
	struct kink_handle *kh;

	kinkd_log(KLLV_INFO, "kink_handle list\n");
	TAILQ_FOREACH(kh, &kg->schedq, next_schedq) {
		if (kh->state->pseudo) {
			kinkd_log(KLLV_INFO,
			    "- %p, pseudo, s=%s\n", kh, kh->state->strname);
			continue;
		}
		kinkd_log_susp(KLLV_INFO,
		    "- %p, s=%s, p=%s",
		    kh, kh->state->strname, kh->peer->remote_principal);
		if (kh->ph2 != NULL)
			kinkd_log_susp(KLLV_INFO,
			    ", src=%s, dst=%s",
			    rcs_sa2str(kh->ph2->src),
			    rcs_sa2str(kh->ph2->dst));
		kinkd_log_susp(KLLV_INFO, "\n");
		kinkd_log_flush();
	}
}


#ifdef DEBUG_THOROUGH_FREE
void
cleanup_sched(struct kink_global *kg)
{
	struct kink_handle *kh;

	while ((kh = TAILQ_FIRST(&kg->schedq)) != NULL) {
		TAILQ_REMOVE(&kg->schedq, kh, next_schedq);

		if (kh->state->pseudo)
			continue;

		if (kh->ph2 != NULL)
			release_ph2(kh->ph2);
		release_handle(kh);
	}
}
#endif


#ifdef MAKE_KINK_LIST_FILE
static void
make_list_file(struct kink_global *kg)
{
	struct kink_handle *kh;
	FILE *fp;
	int fd;

	if ((fd = open(KINK_LIST_FILE_TMP,
	    O_WRONLY | O_CREAT | O_EXCL, 0644)) == -1) {
		kinkd_log(KLLV_ERR,
		    "fopen: %s: %s\n", KINK_LIST_FILE_TMP, strerror(errno));
		return;
	}
	if ((fp = fdopen(fd, "w")) == NULL) {
		kinkd_log(KLLV_ERR,
		    "fopen: %s: %s\n", KINK_LIST_FILE_TMP, strerror(errno));
		unlink(KINK_LIST_FILE_TMP);
		return;
	}

	TAILQ_FOREACH(kh, &kg->schedq, next_schedq) {
		if (kh->state->pseudo)
			continue;
		/*
		 * mature check; XXX hyper layer violation.
		 * we don't know &state_* here.
		 */
		if (!(strcmp(kh->state->strname, "I_AGING") == 0 ||
		      strcmp(kh->state->strname, "I_STALE") == 0 ||
		      strcmp(kh->state->strname, "R_AGING") == 0 ||
		      strcmp(kh->state->strname, "R_STALE") == 0))
			continue;

		fprintf(fp, "%s,%s,%s,%lu\n",
		    kh->peer->remote_principal,
		    rcs_sa2str_wop(kh->ph2->src),
		    rcs_sa2str_wop(kh->ph2->dst),
		    (unsigned long)kh->tkt_endtime);
	}

	fclose(fp);
	rename(KINK_LIST_FILE_TMP, KINK_LIST_FILE);
}
#endif
