/* $Id: session.c,v 1.42 2005/01/19 00:53:47 kamada Exp $ */
/*
 * Copyright (C) 2003, 2004 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "config.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>

#include <netinet/in.h>

#include <errno.h>
#include <fcntl.h>
#include <ifaddrs.h>
#include <netdb.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "../lib/vmbuf.h"
#include "../lib/rc_type.h"
#include "../lib/rc_net.h"
#include "../lib/rbuf.h"
#include "../lib/plog.h"
#include "../lib/if_spmd.h"
#include "pathnames.h"
#include "utils.h"
#include "sockmisc.h"
#include "bbkk.h"
#include "kink_conf.h"
#include "etchosts.h"
#include "handle.h"
#include "peer.h"
#include "scheduler.h"
#include "pfkey.h"
#include "dpd.h"
#include "base.h"
#include "session.h"


static void handle_exitreq(void);
static void cleanup(void);

static int init_signals(void);
#ifdef DEBUG_THOROUGH_FREE
static void cleanup_signals(void);
#endif
static void sig_handler(int sig);
static void handle_signals(void);

static int make_pidfile(void);
static void cleanup_pidfile(void);

static int periodic_update_addrs(struct kink_handle *kh);
static int update_addrs(void);
static struct kink_addr *pickup_kink_addr(struct sockaddr *src);
#ifdef DEBUG_THOROUGH_FREE
static void cleanup_addrs(void);
#endif

static void make_fd_set(int *nfds, fd_set *rfds, int fd_pfkey, int fd_rcnd);

#ifdef SIGINFO
static void print_info(void);
static void print_kink_addrs(void);
#endif

static void callback_delete(rc_type satype,
    uint32_t spi, struct sockaddr *src, struct sockaddr *dst);
static void callback_acquire(rc_type satype, uint32_t seq,
    uint32_t slid, struct sockaddr *src, struct sockaddr *dst);
static void callback_expire(rc_type satype, rc_type samode,
    uint32_t spi, struct sockaddr *src, struct sockaddr *dst);


static struct kink_global *g_kg;
static char *pidfile = NULL;
static pid_t pidfile_pid = 0;
static int sigreq[2];		/* pipe for signal request */
static int chaddrreq = 0;	/* address change request */
int exitreq = 0;		/* exit request */

static LIST_HEAD(addrlist_st, kink_addr) addrlist;


void
session(struct kink_global *kg)
{
	int fd_pfkey, fd_rcnd;
	fd_set rfds0, rfds;
	int ret, nfds;
	struct kink_addr *ka;

	/* XXX used out of this function*/
	g_kg = kg;

	if (make_pidfile() != 0)
		exit(1);
	if (init_signals() != 0) {
		cleanup();
		exit(1);
	}

	/* initial open of KINK sockets */
	LIST_INIT(&addrlist);
	(void)update_addrs();

	/* open PF_KEY socket */
	kg->fd_pfkey = fd_pfkey = pfkey_init();
	if (fd_pfkey == -1) {
		cleanup();
		exit(1);
	}
	pk_setcallback_delete(&callback_delete);
	pk_setcallback_acquire(&callback_acquire);
	pk_setcallback_expire(&callback_expire);

	/* open racoon I/F socket */
	kg->fd_rcnd = fd_rcnd = spmif_init();
	/* currently racoon is optional so that error is not checked here */

	/* prepare descriptor list for select() */
	make_fd_set(&nfds, &rfds0, fd_pfkey, fd_rcnd);

	periodic_update_addrs(NULL);

	/*
	 * main loop
	 */
	kinkd_log(KLLV_DEBUG, "entering main loop\n");
	for (;;) {
		struct timeval tv, *tvp;

		tvp = scheduler(kg, &tv);
		handle_exitreq();
		if (chaddrreq) {
			make_fd_set(&nfds, &rfds0, fd_pfkey, fd_rcnd);
			chaddrreq = 0;
		}

		rfds = rfds0;
		ret = select(nfds, &rfds, NULL, NULL, tvp);
		if (ret == -1) {
			/*
			 * select() may return with EINTR even if SA_RESTART
			 * is set. (implementation-defined)
			 */
			if (errno == EINTR)
				;		/* do nothing */
			else {
				kinkd_log(KLLV_FATAL, "select: %s\n",
				    strerror(errno));
				exitreq = EXITREQ_OTHER;
			}
		} else if (FD_ISSET(fd_pfkey, &rfds)) {
			pfkey_handler(fd_pfkey);
		} else if (fd_rcnd != -1 && FD_ISSET(fd_rcnd, &rfds)) {
			ret = spmif_handler(fd_rcnd);
			if (ret != 0) {
				spmif_clean(fd_rcnd);
				kg->fd_rcnd = fd_rcnd = -1;
				/* reload fd_set */
				make_fd_set(&nfds, &rfds0, fd_pfkey, fd_rcnd);
			}
		} else if (FD_ISSET(sigreq[0], &rfds)) {
			handle_signals();
		} else {
			LIST_FOREACH(ka, &addrlist, next) {
				if (FD_ISSET(ka->fd, &rfds)) {
					receive(kg, ka);
					handle_exitreq();
				}
			}
		}
		handle_exitreq();

		dpd_expunge_dead(kg);
		handle_exitreq();
	}
}

static void
handle_exitreq(void)
{
	switch (exitreq) {
	case 0:
		return;
	case EXITREQ_NOMEM:		/* what should we do? */
		exitreq = 0;
		return;
	default:
		break;
	}

	kinkd_log(KLLV_NOTICE, "kinkd exiting\n");
	cleanup();
	exit(0);
}

static void
cleanup(void)
{
	if (g_kg->fd_rcnd >= 0) {
		(void)spmif_post_quit(g_kg->fd_rcnd);
		spmif_clean(g_kg->fd_rcnd);
	}
#ifdef DEBUG_THOROUGH_FREE
	cleanup_sched(g_kg);
	cleanup_peers(g_kg);
	cleanup_addrs();
	cleanup_pfkey();
#endif
	cleanup_pidfile();
#ifdef MAKE_KINK_LIST_FILE
	unlink(KINK_LIST_FILE);
#endif

#ifdef DEBUG_THOROUGH_FREE
	cleanup_signals();
	/* XXX main.c layer */
	cleanup_etchosts();
	(void)bbkk_fini(g_kg->context);
	free(g_kg->my_principal);
	rcf_clean();
	plog_clean();
	rbuf_clean();
#endif
}


static int
init_signals(void)
{
	static const struct sigaction sa0;
	struct sigaction sa;

	if (pipe(sigreq) == -1) {
		kinkd_log(KLLV_FATAL, "pipe: %s\n", strerror(errno));
		return 1;
	}

	sa = sa0;
	sa.sa_handler = &sig_handler;
	sa.sa_flags = SA_RESTART;
	if (sigaction(SIGHUP, &sa, NULL) == -1) {
		kinkd_log(KLLV_FATAL, "sigaction: %s\n", strerror(errno));
		goto fail;
	}
	if (sigaction(SIGINT, &sa, NULL) == -1) {
		kinkd_log(KLLV_FATAL, "sigaction: %s\n", strerror(errno));
		goto fail;
	}
	if (sigaction(SIGTERM, &sa, NULL) == -1) {
		kinkd_log(KLLV_FATAL, "sigaction: %s\n", strerror(errno));
		goto fail;
	}
	if (sigaction(SIGUSR1, &sa, NULL) == -1) {
		kinkd_log(KLLV_FATAL, "sigaction: %s\n", strerror(errno));
		goto fail;
	}
#ifdef SIGINFO
	if (sigaction(SIGINFO, &sa, NULL) == -1) {
		kinkd_log(KLLV_FATAL, "sigaction: %s\n", strerror(errno));
		goto fail;
	}
#endif
	return 0;

fail:
	close(sigreq[1]);
	close(sigreq[0]);
	return 1;
}

#ifdef DEBUG_THOROUGH_FREE
static void
cleanup_signals(void)
{
	close(sigreq[1]);
	close(sigreq[0]);
}
#endif

static void
sig_handler(int sig)
{
	char sigbyte;

	sigbyte = sig;
	(void)write(sigreq[1], &sigbyte, sizeof(sigbyte));
	/* we cannot call kinkd_log, so ignore errors */
}

static void
handle_signals(void)
{
	char sigbyte;

	if (read(sigreq[0], &sigbyte, sizeof(sigbyte)) == -1) {
		kinkd_log(KLLV_ERR,
		    "cannot read signal: %s\n", strerror(errno));
		return;
	}

	switch (sigbyte) {
	case SIGHUP:
		kinkd_log(KLLV_NOTICE, "SIGHUP: not implemented\n");
		break;
	case SIGINT:
		kinkd_log(KLLV_NOTICE, "SIGINT: shutdown\n");
		exitreq = EXITREQ_USER;
		break;
	case SIGTERM:
		kinkd_log(KLLV_NOTICE, "SIGTERM: shutdown\n");
		exitreq = EXITREQ_USER;
		break;
	case SIGUSR1:
		/* experimental STATUS flooding */
		flood_status(g_kg);
		break;
#ifdef SIGINFO
	case SIGINFO:
		print_info();
		break;
#endif
	default:
		kinkd_log(KLLV_ERR, "unexpected signal %d\n", sigbyte);
		break;
	}
}


static int
make_pidfile(void)
{
	char pidstr[16];
	size_t len;
	const char *myname;
	int fd;

	if (pidfile != NULL) {
		kinkd_log(KLLV_ERR, "make_pidfile() is called twice\n");
		return 1;
	}

	pidfile_pid = getpid();
	myname = get_myname(NULL);
	len = strlen(PIDFILE_DIR) + 1 + strlen(myname) + 4 + 1;
	if ((pidfile = (char *)malloc(len)) == NULL) {
		kinkd_log(KLLV_FATAL, "out of memory\n");
		return 1;
	}
	snprintf(pidfile, len, "%s/%s.pid", PIDFILE_DIR, myname);

	if ((fd = open(pidfile, O_WRONLY | O_CREAT | O_EXCL, 0644)) == -1) {
		kinkd_log(KLLV_FATAL, "%s: cannot make PID file\n", pidfile);
		free(pidfile);
		pidfile = NULL;
		return 1;
	}
	snprintf(pidstr, sizeof(pidstr), "%d\n", pidfile_pid);
	if (write(fd, pidstr, strlen(pidstr)) == -1) {
		kinkd_log(KLLV_FATAL, "%s: cannot write PID file\n", pidfile);
		cleanup_pidfile();
		return 1;
	}
	close(fd);

	return 0;
}

static void
cleanup_pidfile(void)
{
	/* kinkd doesn't fork(2), but just to be safe... */
	if (pidfile_pid != getpid())
		return;

	if (pidfile != NULL) {
		unlink(pidfile);
		free(pidfile);
		pidfile = NULL;
	}
}


static int
periodic_update_addrs(struct kink_handle *kh)
{
	static struct kink_state state = {
		1, "UPDATE_ADDRS",
		&periodic_update_addrs, 10,
		NULL, NULL
	};
	static struct kink_handle skh;

	if (kh == NULL) {
		/* initial invocation */
		skh.g = g_kg;
		skh.state = &state;
		sched_register(&skh);
	} else {
		(void)update_addrs();
		sched_resettimer(kh);
	}
	return 0;
}

static int
update_addrs(void)
{
	static const struct kink_addr ka0;
	int fd, on = 1;
	struct sockaddr *sa;
	struct ifaddrs *ifp0, *ifp;
	in_addr_t in_addr;
	struct in6_addr *in6_addr;
	struct kink_addr *ka;

	LIST_FOREACH(ka, &addrlist, next) {
		ka->alive = 0;
	}

	if (getifaddrs(&ifp0) == -1) {
		kinkd_log(KLLV_ERR, "getifaddrs: %s\n", strerror(errno));
		return 1;
	}
	for (ifp = ifp0; ifp != NULL; ifp = ifp->ifa_next) {
		sa = ifp->ifa_addr;

		switch (sa->sa_family) {
		case AF_INET:
			in_addr =
			    ntohl(((struct sockaddr_in *)sa)->sin_addr.s_addr);
			if (in_addr == INADDR_ANY)
				continue;
			else if (IN_CLASSA(in_addr) &&
			    (in_addr & IN_CLASSA_NET) ==
			    (IN_LOOPBACKNET << IN_CLASSA_NSHIFT))
				continue;
			break;
		case AF_INET6:
			in6_addr =
			    &((struct sockaddr_in6 *)sa)->sin6_addr;
			if (IN6_IS_ADDR_UNSPECIFIED(in6_addr))
				continue;
			else if (IN6_IS_ADDR_LOOPBACK(in6_addr))
				continue;
			break;
		default:
#if 0
			kinkd_log(KLLV_DEBUG,
			    "unknown address %s (AF=%d) on %s\n",
			    rcs_sa2str(sa), sa->sa_family, ifp->ifa_name);
#endif
			continue;
		}
		fix_scope_id_ref_ifname(sa, ifp->ifa_name);

		if (!match_saddr_with_config(sa, ifp->ifa_name))
			continue;

		/* search */
		LIST_FOREACH(ka, &addrlist, next)
			if (rcs_cmpsa_wop(ka->sa, sa) == 0)	/* no wop? */
				break;
		if (ka != NULL) {		/* match */
			ka->alive = 1;
			continue;
		}

		/* not match, so create new kink_addr */
		if ((ka = (struct kink_addr *)malloc(sizeof(*ka))) == NULL) {
			kinkd_log(KLLV_FATAL, "out of memory\n");
			freeifaddrs(ifp0);
			exitreq = EXITREQ_NOMEM;
			return 1;
		}
		*ka = ka0;
		if ((ka->sa = rcs_sadup(sa)) == NULL) {
			kinkd_log(KLLV_FATAL, "out of memory\n");
			free(ka);
			freeifaddrs(ifp0);
			exitreq = EXITREQ_NOMEM;
			return 1;
		}
		kinkd_log(KLLV_INFO, "binding %s\n", rcs_sa2str(ka->sa));

		fd = socket(ka->sa->sa_family, SOCK_DGRAM, 0);
		if (fd < 0) {
			kinkd_log(KLLV_INFO,
			    "socket: %s; skip\n", strerror(errno));
			free(ka->sa);
			free(ka);
			continue;
		}
		if (ka->sa->sa_family == PF_INET6 &&
		    setsockopt(fd, IPPROTO_IPV6, IPV6_V6ONLY,
		    &on, sizeof(on)) == -1) {
			kinkd_log(KLLV_INFO,
			    "setsockopt: %s; skip\n", strerror(errno));
			close(fd);
			free(ka->sa);
			free(ka);
			continue;
		}
		if (setsockopt_bypass(fd, ka->sa->sa_family) == -1)
			kinkd_log(KLLV_INFO, "cannot set bypass policy\n");
		if (bind(fd, ka->sa, COMPAT_SA_LEN(ka->sa)) == -1) {
			kinkd_log(KLLV_INFO,
			    "bind: %s; skip\n", strerror(errno));
			close(fd);
			free(ka->sa);
			free(ka);
			continue;
		}

		ka->fd = fd;
		ka->alive = 1;
		LIST_INSERT_HEAD(&addrlist, ka, next);
		chaddrreq = 1;
	}
	freeifaddrs(ifp0);

	/* close */
again:
	LIST_FOREACH(ka, &addrlist, next) {
		if (ka->alive || (ka->refcnt != 0))
			continue;
		kinkd_log(KLLV_INFO, "unbinding %s\n", rcs_sa2str(ka->sa));
		LIST_REMOVE(ka, next);
		close(ka->fd);
		free(ka->sa);
		free(ka);
		chaddrreq = 1;
		goto again;
	}
	return 0;
}

static struct kink_addr *
pickup_kink_addr(struct sockaddr *src)
{
	struct kink_addr *ka;

	LIST_FOREACH(ka, &addrlist, next) {
		if (rcs_cmpsa_wop(ka->sa, src) == 0)
			return ka;
	}
	return NULL;
}

#ifdef DEBUG_THOROUGH_FREE
static void
cleanup_addrs(void)
{
	struct kink_addr *ka;

	while ((ka = LIST_FIRST(&addrlist)) != NULL) {
		LIST_REMOVE(ka, next);

		close(ka->fd);
		free(ka->sa);
		free(ka);
	}
}
#endif


static void
make_fd_set(int *nfds, fd_set *rfds, int fd_pfkey, int fd_rcnd)
{
	struct kink_addr *ka;

#define ADD_FD(fd, set, nfds) do {					\
		if ((fd) >= FD_SETSIZE) {				\
			kinkd_log(KLLV_FATAL, "fd_set overrun\n");	\
			exit(1);					\
		}							\
		if ((fd) >= (nfds))					\
			(nfds) = (fd) + 1;				\
		FD_SET((fd), (set));					\
	} while (0 /* CONSTCOND */)

	*nfds = 0;
	FD_ZERO(rfds);
	ADD_FD(fd_pfkey, rfds, *nfds);
	if (fd_rcnd != -1)
		ADD_FD(fd_rcnd, rfds, *nfds);
	ADD_FD(sigreq[0], rfds, *nfds);
	LIST_FOREACH(ka, &addrlist, next) {
		ADD_FD(ka->fd, rfds, *nfds);
	}

#undef ADD_FD
}


#ifdef SIGINFO
static void
print_info(void)
{
	kinkd_log(KLLV_INFO, "SIGINFO\n");
	print_kink_addrs();
	print_kink_peers(g_kg);
	print_kink_handles(g_kg);
}

static void
print_kink_addrs(void)
{
	struct kink_addr *ka;

	kinkd_log(KLLV_INFO, "kink_addr list\n");
	LIST_FOREACH(ka, &addrlist, next) {
		kinkd_log(KLLV_INFO,
		    "- fd=%d, %srefcnt=%d, addr=%s\n",
		    ka->fd,
		    ka->alive ? "" : "dead, ",
		    ka->refcnt,
		    rcs_sa2str(ka->sa));
	}
}
#endif


static void
callback_delete(rc_type satype,
    uint32_t spi, struct sockaddr *src, struct sockaddr *dst)
{
	delete(g_kg, satype, spi, src, dst);
}

static void
callback_acquire(rc_type satype, uint32_t seq,
    uint32_t slid, struct sockaddr *src, struct sockaddr *dst)
{
	acquire(g_kg, satype, seq, slid, src, dst, pickup_kink_addr(src));
}

static void
callback_expire(rc_type satype, rc_type samode,
    uint32_t spi, struct sockaddr *src, struct sockaddr *dst)
{
	expire(g_kg, satype, samode, spi, src, dst);
}
