/* $Id: if_pfkeyv2.c,v 1.27 2005/01/07 09:02:08 sakane Exp $ */

/*	$KAME: pfkey.c,v 1.138 2003/06/30 11:01:18 sakane Exp $	*/

/*
 * Copyright (C) 1995, 1996, 1997, and 1998 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>

#ifdef HAVE_NET_PFKEYV2_H
# include <net/pfkeyv2.h>
# include <netinet6/ipsec.h>
#else
# include <stdint.h>
# include <linux/pfkeyv2.h>
# include <linux/ipsec.h>
#endif
#include "pfkeyv2aux.h"
#include <netinet/in.h>

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#ifdef HAVE_STDARG_H
#include <stdarg.h>
#else
#include <varargs.h>
#endif
#include <errno.h>

#include "racoon.h"
#include "cfsetup.h"

static struct rcpfk_cb *cb;
static pid_t pid = 0;
static int f_noharm = 0;

static struct sadb_supported *supported_map_auth = NULL;
static struct sadb_supported *supported_map_enc = NULL;
#if defined(PFK_IPCOMP_SUPPORTED)
static struct sadb_supported *supported_map_comp = NULL;
#endif

static vchar_t *rcpfk_recv __P((struct rcpfk_msg *));
static int rcpfk_send_spdaddx __P((struct rcpfk_msg *, int));
static int rcpfk_check_ext_content __P((struct sadb_ext *, caddr_t));
static int rcpfk_open __P((struct rcpfk_msg *));
static int rcpfk_close __P((struct rcpfk_msg *));
static int rcpfk_send __P((struct rcpfk_msg *, vchar_t *));
static void rcpfk_seterror __P((struct rcpfk_msg *, int, const char *fmt, ...));

static int rcpfk_set_sadbmsg __P((vchar_t **, struct rcpfk_msg *, int));
static int rcpfk_set_sadbsa __P((vchar_t **, struct rcpfk_msg *));
static int rcpfk_set_sadbaddress __P((vchar_t **, struct rcpfk_msg *, int));
static int rcpfk_set_sadbkey __P((vchar_t **, struct rcpfk_msg *, int));
static int rcpfk_set_sadblifetime __P((vchar_t **, struct rcpfk_msg *, int));
static int rcpfk_set_sadbxsa2 __P((vchar_t **, struct rcpfk_msg *));
static int rcpfk_set_sadbxpolicy __P((vchar_t **, struct rcpfk_msg *, int));
static int rcpfk_set_sadbxpolicy_transport __P((vchar_t **, struct rcpfk_msg *, int));
static int rcpfk_set_sadbxpolicy_tunnel __P((vchar_t **, struct rcpfk_msg *, int));

static int rcpfk_recv_getspi __P((caddr_t *, struct rcpfk_msg *));
static int rcpfk_send_addx __P((struct rcpfk_msg *, int));
static int rcpfk_recv_update __P((caddr_t *, struct rcpfk_msg *));
static int rcpfk_recv_add __P((caddr_t *, struct rcpfk_msg *));
static int rcpfk_recv_expire __P((caddr_t *, struct rcpfk_msg *));
static int rcpfk_recv_acquire __P((caddr_t *, struct rcpfk_msg *));
static int rcpfk_recv_delete __P((caddr_t *, struct rcpfk_msg *));
static int rcpfk_recv_register __P((caddr_t *, struct rcpfk_msg *));
static int set_supported_algorithm __P((caddr_t, struct sadb_supported **));
static int rcpfk_recv_spdupdate __P((caddr_t *, struct rcpfk_msg *));
static int rcpfk_recv_spdadd __P((caddr_t *, struct rcpfk_msg *));
static int rcpfk_recv_spddelete __P((caddr_t *, struct rcpfk_msg *));
static int rcpfk_recv_spddelete2 __P((caddr_t *, struct rcpfk_msg *));
static int rcpfk_recv_spdexpire __P((caddr_t *, struct rcpfk_msg *));
static int rcpfk_recv_spdget __P((caddr_t *, struct rcpfk_msg *));
static struct sadb_alg *findsupportedalg __P((struct sadb_supported *, int));

static struct pfkey_msgtype {
	char *name;
	int (*recvfunc) __P((caddr_t *, struct rcpfk_msg *));
} rcpfk_msg[] = {
	{ "",			0, },
	{ "GETSPI",		rcpfk_recv_getspi, },
	{ "UPDATE",		rcpfk_recv_update, },
	{ "ADD",		rcpfk_recv_add, },
	{ "DELETE",		rcpfk_recv_delete, },
	{ "GET",		0, },
	{ "ACQUIRE",		rcpfk_recv_acquire, },
	{ "REGISTER",		rcpfk_recv_register, },
	{ "EXPIRE",		rcpfk_recv_expire, },
	{ "FLUSH",		0, },
	{ "DUMP",		0, },
	{ "X_PROMISC",		0, },
	{ "X_PCHANGE",		0, },
	{ "X_SPDUPDATE",	rcpfk_recv_spdupdate, },
	{ "X_SPDADD",		rcpfk_recv_spdadd, },
	{ "X_SPDDELETE",	rcpfk_recv_spddelete, },
	{ "X_SPDGET",		rcpfk_recv_spdget, },
	{ "X_SPDACQUIRE",	0, },
	{ "X_SPDDUMP",		0, },
	{ "X_SPDFLUSH",		0, },
	{ "X_SPDSETIDX",	0, },
	{ "X_SPDEXPIRE",	rcpfk_recv_spdexpire, },
	{ "X_SPDDELETE2",	rcpfk_recv_spddelete2, },
};

/*
 * PF_KEY packet handler
 * IN: rcpfk_con must be allocated.
 * OUT:
 *    0: success
 *   -1: fail
 */
int
rcpfk_handler(struct rcpfk_msg *rc)
{
	vchar_t *msg;
	struct sadb_msg *base;
	struct sadb_ext *ext;
	caddr_t p, ep;
	caddr_t mhp[SADB_EXT_MAX + 1];
	int i;

	/* receive pfkey message. */
	if ((msg = rcpfk_recv(rc)) == NULL)
		return -1;

	/* initialize */
	for (i = 0; i < sizeof(mhp)/sizeof(mhp[0]); i++)
		mhp[i] = 0;
	mhp[0] = msg->v;
	p = (caddr_t)msg->v;
	ep = p + msg->l;

	/* skip base header */
	p += sizeof(struct sadb_msg);

	while (p < ep) {
		ext = (struct sadb_ext *)p;

		/* length check */
		if (ep < p + sizeof(*ext) ||
		    PFKEY_EXTLEN(ext) < sizeof(*ext) ||
		    ep < p + PFKEY_EXTLEN(ext)) {
			rcpfk_seterror(rc, EINVAL,
			    "invalid pfkey extension format");
    err:
			vfree(msg);
			return -1;
		}

		/* duplicate check */
		if (mhp[ext->sadb_ext_type] != 0) {
			rcpfk_seterror(rc, EINVAL,
			    "duplicate pfkey extension type=%d",
			    ext->sadb_ext_type);
			goto err;
		}

		/* length check in the content */
		if (rcpfk_check_ext_content(ext, ep)) {
			rcpfk_seterror(rc, EINVAL,
			    "invalid pfkey extension type=%d",
			    ext->sadb_ext_type);
			goto err;
		}

		mhp[ext->sadb_ext_type] = (caddr_t)ext;

		p += PFKEY_EXTLEN(ext);
	}
	if (p != ep) {
		rcpfk_seterror(rc, EINVAL, "invalid pfkey extension format");
		goto err;
	}

	base = (struct sadb_msg *)msg->v;

	if (base->sadb_msg_type <=0 ||
	    ARRAYLEN(rcpfk_msg) <= base->sadb_msg_type) {
		rcpfk_seterror(rc, ENOTSUP,
		    "unknown message type %d", base->sadb_msg_type);
		goto err;
	}

	/* return when error occured. */
	if (base->sadb_msg_errno) {
		rcpfk_seterror(rc, base->sadb_msg_errno,
		    "error at the kernel on %s, %s",
		    rcpfk_msg[base->sadb_msg_type].name,
		    strerror(base->sadb_msg_errno));
		goto err;
	}

	/* the message has to be processed or not ? */
	if (base->sadb_msg_pid != pid &&
	  !(base->sadb_msg_type == SADB_DELETE ||
	    base->sadb_msg_type == SADB_ACQUIRE ||
	    base->sadb_msg_type == SADB_EXPIRE)) {
		rcpfk_seterror(rc, ENOENT, "%s message is not interesting "
		    "because pid %d is not mine",
		    rcpfk_msg[base->sadb_msg_type].name,
		    base->sadb_msg_pid);
		goto err;
	}

	if (rcpfk_msg[base->sadb_msg_type].recvfunc == 0) {
		rcpfk_seterror(rc, ENOTSUP, "command %s not supported",
		    rcpfk_msg[base->sadb_msg_type].name);
		goto err;
	}

	/* each function probably call the callback function */
	if ((rcpfk_msg[base->sadb_msg_type].recvfunc)(mhp, rc))
		goto err;

	vfree(msg);

	return 0;
}

/*
 *
 */
static vchar_t *
rcpfk_recv(struct rcpfk_msg *rc)
{
	vchar_t *buf;
	struct sadb_msg base;
	int len, reallen;

	len = recv(rc->so, (caddr_t)&base, sizeof(base), MSG_PEEK);
	if (len < 0) {
		rcpfk_seterror(rc, errno, "%s", strerror(errno));
		return NULL;
	} else if (len < sizeof(base)) {
		rcpfk_seterror(rc, EINVAL, "invalid message length");
		return NULL;
	}

	reallen = PFKEY_UNUNIT64(base.sadb_msg_len);
	if ((buf = vmalloc(reallen)) == NULL) {
		rcpfk_seterror(rc, errno, "%s", strerror(errno));
		return NULL;
	}

	len = recv(rc->so, buf->v, buf->l, MSG_PEEK);
	if (len < 0) {
		rcpfk_seterror(rc, errno, "%s", strerror(errno));
    err:
		vfree(buf);
		return NULL;
	} else if (len != reallen) {
		rcpfk_seterror(rc, EINVAL, "invalid message length");
		goto err;
	}

	len = recv(rc->so, buf->v, buf->l, 0);
	if (len < 0) {
		rcpfk_seterror(rc, errno, "%s", strerror(errno));
		goto err;
	} else if (len != reallen) {
		rcpfk_seterror(rc, EINVAL, "invalid message length");
		goto err;
	}

	return buf;
}

static int
rcpfk_check_ext_content(struct sadb_ext *ext, caddr_t ep)
{
	struct sadb_address *addr;
	struct sockaddr *sa;
	struct sadb_supported *sup;

	/* length check of the content */
	switch (ext->sadb_ext_type) {
	case SADB_EXT_SA:
	case SADB_EXT_LIFETIME_CURRENT:
	case SADB_EXT_LIFETIME_HARD:
	case SADB_EXT_LIFETIME_SOFT:
	case SADB_EXT_KEY_AUTH:
	case SADB_EXT_KEY_ENCRYPT:
	case SADB_EXT_IDENTITY_SRC:
	case SADB_EXT_IDENTITY_DST:
	case SADB_EXT_SENSITIVITY:
	case SADB_EXT_PROPOSAL:
	case SADB_EXT_SPIRANGE:
	case SADB_X_EXT_POLICY:
	case SADB_X_EXT_SA2:
#ifdef SADB_X_EXT_TAG
	case SADB_X_EXT_TAG:
#endif
		break;
	case SADB_EXT_SUPPORTED_AUTH:
	case SADB_EXT_SUPPORTED_ENCRYPT:
		sup = (struct sadb_supported *)ext;
		if ((PFKEY_UNUNIT64(sup->sadb_supported_len) - sizeof(*sup)) %
		    sizeof(struct sadb_alg) != 0)
			return -1;
		break;
	case SADB_EXT_ADDRESS_SRC:
	case SADB_EXT_ADDRESS_DST:
	case SADB_EXT_ADDRESS_PROXY:
		addr = (struct sadb_address *)ext;
		sa = (struct sockaddr *)(addr + 1);
		if (ep < (caddr_t)sa + SA_LEN(sa))
			return -1;
		break;
	default:
		return -1;
	}

	return 0;
}

/*
 * PF_KEY initialization
 * IN: rc->flags, cb
 * OUT: rc->so
 */
int
rcpfk_init(struct rcpfk_msg *rc, struct rcpfk_cb *cb0)
{
	/* set pid */
	pid = getpid();

	/* set flag */
	if (rc->flags & PFK_FLAG_NOHARM)
		f_noharm++;

	/* open the PF_KEY I/F */
	if (rcpfk_open(rc))
		return -1;

	/* register the socket to each type */
	rc->satype = RCT_SATYPE_AH;
	if (rcpfk_send_register(rc) || rcpfk_handler(rc))
		return -1;

	rc->satype = RCT_SATYPE_ESP;
	if (rcpfk_send_register(rc) || rcpfk_handler(rc))
		return -1;

	rc->satype = RCT_SATYPE_IPCOMP;
	if (rcpfk_send_register(rc) || rcpfk_handler(rc))
		return -1;

	/* initialize cb */
	cb = cb0;

	return 0;
}

int
rcpfk_clean(struct rcpfk_msg *rc)
{
	if (supported_map_auth != NULL) {
		rc_free(supported_map_auth);
		supported_map_auth = NULL;
	}
	if (supported_map_enc != NULL) {
		rc_free(supported_map_enc);
		supported_map_enc = NULL;
	}
#if defined(PFK_IPCOMP_SUPPORTED)
	if (supported_map_comp != NULL) {
		rc_free(supported_map_comp);
		supported_map_comp = NULL;
	}
#endif

	return rcpfk_close(rc);
}

static int
rcpfk_open(struct rcpfk_msg *rc)
{
	const int len = RCPFK_SOCKBUFSIZE;

	if ((rc->so = socket(PF_KEY, SOCK_RAW, PF_KEY_V2)) == -1) {
		rcpfk_seterror(rc, errno, "%s", strerror(errno));
		return -1;
	}

	/*
	 * This is a temporary workaround for KAME PR 154.
	 * Don't really care even if it fails.
	 */
	if (setsockopt(rc->so, SOL_SOCKET, SO_SNDBUF, &len, sizeof(len)) &&
	    setsockopt(rc->so, SOL_SOCKET, SO_RCVBUF, &len, sizeof(len))) {
		rcpfk_seterror(rc, errno, "%s", strerror(errno));
		rcpfk_close(rc);
		return -1;
	}

	return 0;
}

static int
rcpfk_close(struct rcpfk_msg *rc)
{
	if (close(rc->so) == -1) {
		rcpfk_seterror(rc, errno, "%s", strerror(errno));
		return -1;
	}

	return 0;
}

static int
rcpfk_send(struct rcpfk_msg *rc, vchar_t *buf)
{
	/*
	 * set final message length (XXX here?)
	 */
	((struct sadb_msg *)buf->v)->sadb_msg_len = PFKEY_UNIT64(buf->l);

	if (send(rc->so, buf->v, buf->l, 0) == -1) {
		rcpfk_seterror(rc, errno, "%s", strerror(errno));
		return -1;
	}

	return 0;
}

static void
rcpfk_seterror(struct rcpfk_msg *rc, int eno, const char *fmt, ...)
{
	va_list ap;

	rc->eno = eno;

	va_start(ap, fmt);
	vsnprintf(rc->estr, sizeof(rc->estr), fmt, ap);
	va_end(ap);
}


/*
 * sending modules
 */
/*
 * send a SADB_GETSPI to kernel.
 *     <base, address, SPI range>
 */
int
rcpfk_send_getspi(struct rcpfk_msg *rc)
{
	vchar_t *buf = 0;

	if (rcpfk_set_sadbmsg(&buf, rc, SADB_GETSPI)) {
    err:
		if (buf)
			vfree(buf);
		return -1;
	}

	if (rcpfk_set_sadbxsa2(&buf, rc))
		goto err;

	if (rcpfk_set_sadbaddress(&buf, rc, SADB_EXT_ADDRESS_SRC))
		goto err;

	if (rcpfk_set_sadbaddress(&buf, rc, SADB_EXT_ADDRESS_DST))
		goto err;

	if (rcpfk_send(rc, buf)) {
		vfree(buf);
		return -1;
	}

	vfree(buf);
	return 0;
}

/*
 * send a SADB_UPDATE or a SADB_ADD to kernel.
 *     <base, SA, (lifetime(HSC),) address(SD), (address(P),)
 *         key(AE), (identity(SD),) (sensitivity)>
 */
static int
rcpfk_send_addx(struct rcpfk_msg *rc, int type)
{
	vchar_t *buf = 0;

	if (rcpfk_set_sadbmsg(&buf, rc, type)) {
    err:
		if (buf)
			vfree(buf);
		return -1;
	}

	if (rcpfk_set_sadbsa(&buf, rc))
		goto err;

	if (rcpfk_set_sadbxsa2(&buf, rc))
		goto err;

	if (rcpfk_set_sadblifetime(&buf, rc, SADB_EXT_LIFETIME_HARD))
		goto err;

	if (rcpfk_set_sadblifetime(&buf, rc, SADB_EXT_LIFETIME_SOFT))
		goto err;

	if (rcpfk_set_sadbaddress(&buf, rc, SADB_EXT_ADDRESS_SRC))
		goto err;

	if (rcpfk_set_sadbaddress(&buf, rc, SADB_EXT_ADDRESS_DST))
		goto err;

	if (rc->satype != RCT_SATYPE_AH &&
	    rcpfk_set_sadbkey(&buf, rc, SADB_EXT_KEY_ENCRYPT))
		goto err;

	if (rcpfk_set_sadbkey(&buf, rc, SADB_EXT_KEY_AUTH))
		goto err;

	if (rcpfk_send(rc, buf)) {
		vfree(buf);
		return -1;
	}

	vfree(buf);
	return 0;
}

int
rcpfk_send_update(struct rcpfk_msg *rc)
{
	return rcpfk_send_addx(rc, SADB_UPDATE);
}

int
rcpfk_send_add(struct rcpfk_msg *rc)
{
	return rcpfk_send_addx(rc, SADB_ADD);
}

/*
 * send a SADB_DELETE to kernel.
 *     <base, SA(*), address(SD)>
 */
int
rcpfk_send_delete(struct rcpfk_msg *rc)
{
	vchar_t *buf = 0;

	if (rcpfk_set_sadbmsg(&buf, rc, SADB_DELETE)) {
    err:
		if (buf)
			vfree(buf);
		return -1;
	}

	/*
	 * when it sends a SADB_DELETE without spi to the kernel.  This is 
	 * the "delete all" request (an extension also present in Solaris)
	 */
	if (rc->spi != 0) {
		if (rcpfk_set_sadbsa(&buf, rc))
			goto err;
	}

	if (rcpfk_set_sadbaddress(&buf, rc, SADB_EXT_ADDRESS_SRC))
		goto err;

	if (rcpfk_set_sadbaddress(&buf, rc, SADB_EXT_ADDRESS_DST))
		goto err;

	if (rcpfk_send(rc, buf)) {
		vfree(buf);
		return -1;
	}

	vfree(buf);
	return 0;
}

/*
 * send an error against ACQUIRE message to kenrel.
 *     <base>
 */
int
rcpfk_send_acquire(struct rcpfk_msg *rc)
{
	vchar_t *buf = 0;

	if (rcpfk_set_sadbmsg(&buf, rc, SADB_ACQUIRE)) {
		if (buf)
			vfree(buf);
		return -1;
	}

	((struct sadb_msg *)buf->v)->sadb_msg_errno = rc->eno;

	if (rcpfk_send(rc, buf)) {
		vfree(buf);
		return -1;
	}

	vfree(buf);
	return 0;
}

/*
 * sending SADB_REGISTER message to the kernel.
 *     <base>
 */
int
rcpfk_send_register(struct rcpfk_msg *rc)
{
	vchar_t *buf = 0;

	if (rcpfk_set_sadbmsg(&buf, rc, SADB_REGISTER)) {
		if (buf)
			vfree(buf);
		return -1;
	}

	if (rcpfk_send(rc, buf)) {
		vfree(buf);
		return -1;
	}

	vfree(buf);
	return 0;
}

/*
 * send a SADB_SPDUPDATE/SADB_SPDADD to kernel.
 *     <base, SA, SA2, lifetime(H), address(SD), x_policy>
 */
static int
rcpfk_send_spdaddx(struct rcpfk_msg *rc, int type)
{
	vchar_t *buf = 0;
	struct sockaddr *sa_addr_bak;

	if (rcpfk_set_sadbmsg(&buf, rc, type)) {
    err:
		if (buf)
			vfree(buf);
		return -1;
	}

	if (rcpfk_set_sadbxsa2(&buf, rc))
		goto err;

	sa_addr_bak = rc->sa_src;
	rc->sa_src = rc->sp_src;
	if (rcpfk_set_sadbaddress(&buf, rc, SADB_EXT_ADDRESS_SRC))
		goto err;
	rc->sa_src = sa_addr_bak;

	sa_addr_bak = rc->sa_dst;
	rc->sa_dst = rc->sp_dst;
	if (rcpfk_set_sadbaddress(&buf, rc, SADB_EXT_ADDRESS_DST))
		goto err;
	rc->sa_dst = sa_addr_bak;

	if (rcpfk_set_sadblifetime(&buf, rc, SADB_EXT_LIFETIME_HARD))
		goto err;

	if (rcpfk_set_sadbxpolicy(&buf, rc, SADB_X_SPDUPDATE))
		goto err;

	((struct sadb_msg *)buf->v)->sadb_msg_satype = SADB_SATYPE_UNSPEC;

	if (rcpfk_send(rc, buf)) {
		vfree(buf);
		return -1;
	}

	vfree(buf);
	return 0;
}

int
rcpfk_send_spdupdate(struct rcpfk_msg *rc)
{
	return rcpfk_send_spdaddx(rc, SADB_X_SPDUPDATE);
}

int
rcpfk_send_spdadd(struct rcpfk_msg *rc)
{
	return rcpfk_send_spdaddx(rc, SADB_X_SPDADD);
}

/*
 * send a SADB_SPDDELETE to kernel.
 */
int
rcpfk_send_spddelete(struct rcpfk_msg *rc)
{
	return 0;
}

/*
 * send a SADB_SPDDELETE2 to kernel.
 *     <base, x_policy>
 */
int
rcpfk_send_spddelete2(struct rcpfk_msg *rc)
{
	vchar_t *buf = 0;

	if (rcpfk_set_sadbmsg(&buf, rc, SADB_X_SPDDELETE2)) {
    err:
		if (buf)
			vfree(buf);
		return -1;
	}

	if (rcpfk_set_sadbxsa2(&buf, rc))
		goto err;

	if (rcpfk_set_sadbxpolicy(&buf, rc, SADB_X_SPDDELETE2))
		goto err;

	((struct sadb_msg *)buf->v)->sadb_msg_satype = SADB_SATYPE_UNSPEC;

	if (rcpfk_send(rc, buf)) {
		vfree(buf);
		return -1;
	}

	vfree(buf);
	return 0;
}

/*
 * make a sadb_msg
 */
static int
rcpfk_set_sadbmsg(vchar_t **msg, struct rcpfk_msg *rc, int type)
{
	vchar_t *buf;
	struct sadb_msg *p;
	int len;

	len = sizeof(*p);
	if ((buf = vmalloc(len)) == NULL) {
		rcpfk_seterror(rc, errno, "%s", strerror(errno));
		return -1;
	}

	p = (struct sadb_msg *)buf->v;
	p->sadb_msg_version = PF_KEY_V2;
	p->sadb_msg_type = type;
	p->sadb_msg_errno = 0;
	switch (type) {
	case SADB_X_SPDUPDATE:
	case SADB_X_SPDADD:
	case SADB_X_SPDDELETE:
	/* XXX other SADB_X_SPD* ? */
		p->sadb_msg_satype = SADB_SATYPE_UNSPEC;
		break;
	default:
		p->sadb_msg_satype = rct2pfk_satype(rc->satype);
		break;
	}
	p->sadb_msg_len = 0;	/* must be update before it is sent */
	p->sadb_msg_reserved = 0;
	p->sadb_msg_seq = rc->seq;
	p->sadb_msg_pid = (u_int32_t)pid;

	*msg = buf;
	return 0;
}

/*
 * append a sadb_sa to the buffer.
 * OUT: total length of the buffer
 */
static int
rcpfk_set_sadbsa(vchar_t **msg, struct rcpfk_msg *rc)
{
	vchar_t *buf;
	struct sadb_sa *p;
	int len, prevlen, extlen;

	extlen = sizeof(struct sadb_sa);
	prevlen = (*msg)->l;
	len = prevlen + extlen;
	if ((buf = vreallocf(*msg, len)) == NULL) {
		rcpfk_seterror(rc, errno, "%s", strerror(errno));
		return -1;
	}

	p = (struct sadb_sa *)(buf->v + prevlen);
	p->sadb_sa_len = PFKEY_UNIT64(extlen);
	p->sadb_sa_exttype = SADB_EXT_SA;
	p->sadb_sa_spi = rc->spi;
	p->sadb_sa_replay = rc->wsize;
	p->sadb_sa_state = SADB_SASTATE_LARVAL;
	p->sadb_sa_auth = rct2pfk_authtype(rc->authtype);
	if (rc->satype == RCT_SATYPE_AH)
		p->sadb_sa_encrypt = SADB_EALG_NONE;
	else
		p->sadb_sa_encrypt = rct2pfk_enctype(rc->enctype);
	p->sadb_sa_flags = rc->saflags;

	*msg = buf;
	return 0;
}

/*
 * append a sadb_address to the buffer.
 * OUT: total length of the buffer
 */
static int
rcpfk_set_sadbaddress(vchar_t **msg, struct rcpfk_msg *rc, int type)
{
	vchar_t *buf;
	struct sadb_address *p;
	struct sockaddr *sa;
	int pref;
	int len, prevlen, extlen;

	switch (type) {
	case SADB_EXT_ADDRESS_SRC:
		sa = rc->sa_src;
		pref = rc->pref_src;
		break;
	case SADB_EXT_ADDRESS_DST:
		sa = rc->sa_dst;
		pref = rc->pref_dst;
		break;
	case SADB_EXT_ADDRESS_PROXY:
		/* proxy can not be used */
	default:
		rcpfk_seterror(rc, EINVAL, "invalid address type=%d", type);
		return -1;
	}
	extlen = sizeof(struct sadb_address) + PFKEY_ALIGN8(SA_LEN(sa));
	prevlen = (*msg)->l;
	len = prevlen + extlen;
	if ((buf = vreallocf(*msg, len)) == NULL) {
		rcpfk_seterror(rc, errno, "%s", strerror(errno));
		return -1;
	}

	p = (struct sadb_address *)(buf->v + prevlen);
	p->sadb_address_len = PFKEY_UNIT64(extlen);
	p->sadb_address_exttype = type & 0xffff;
	p->sadb_address_proto = rct2pfk_proto(rc->ul_proto) & 0xff;
	p->sadb_address_prefixlen = pref;
	p->sadb_address_reserved = 0;
	memcpy(p + 1, sa, SA_LEN(sa));

	*msg = buf;
	return 0;
}

/*
 * set sadb_key structure after clearing buffer with zero.
 * OUT: the pointer of buf + len.
 */
static int
rcpfk_set_sadbkey(vchar_t **msg, struct rcpfk_msg *rc, int type)
{
	vchar_t *buf;
	struct sadb_key *p;
	int keytype;
	size_t keylen;
	caddr_t key;
	int len, prevlen, extlen;

	switch (type) {
	case SADB_EXT_KEY_AUTH:
		keytype = rct2pfk_authtype(rc->authtype);
		key = rc->authkey;
		keylen = rc->authkeylen;
		break;
	case SADB_EXT_KEY_ENCRYPT:
		keytype = rct2pfk_enctype(rc->enctype);
		key = rc->enckey;
		keylen = rc->enckeylen;
		break;
	default:
		rcpfk_seterror(rc, EINVAL, "invalid key type=%d", type);
		return -1;
	}
	extlen = sizeof(struct sadb_key) + PFKEY_ALIGN8(keylen);
	prevlen = (*msg)->l;
	len = prevlen + extlen;
	if ((buf = vreallocf(*msg, len)) == NULL) {
		rcpfk_seterror(rc, errno, "%s", strerror(errno));
		return -1;
	}

	p = (struct sadb_key *)(buf->v + prevlen);
	p->sadb_key_len = PFKEY_UNIT64(extlen);
	p->sadb_key_exttype = type & 0xffff;
	p->sadb_key_bits = keylen << 3;
	p->sadb_key_reserved = 0;
	memcpy(p + 1, key, keylen);

	*msg = buf;
	return 0;
}

/*
 * set sadb_lifetime structure after clearing buffer with zero.
 * OUT: the pointer of buf + len.
 */
static int
rcpfk_set_sadblifetime(vchar_t **msg, struct rcpfk_msg *rc, int type)
{
	vchar_t *buf;
	struct sadb_lifetime *p;
	u_int64_t lft_time, lft_bytes;
	int len, prevlen, extlen;

	switch (type) {
	case SADB_EXT_LIFETIME_SOFT:
		lft_time = rc->lft_soft_time;
		lft_bytes = rc->lft_soft_bytes;
		break;
	case SADB_EXT_LIFETIME_HARD:
		lft_time = rc->lft_hard_time;
		lft_bytes = rc->lft_hard_bytes;
		break;
	default:
		rcpfk_seterror(rc, EINVAL, "invalid lifetime type=%d", type);
		return -1;
	}
	extlen = sizeof(*p);
	prevlen = (*msg)->l;
	len = prevlen + extlen;
	if ((buf = vreallocf(*msg, len)) == NULL) {
		rcpfk_seterror(rc, errno, "%s", strerror(errno));
		return -1;
	}

	p = (struct sadb_lifetime *)(buf->v + prevlen);
	p->sadb_lifetime_len = PFKEY_UNIT64(extlen);
	p->sadb_lifetime_exttype = type & 0xffff;
	p->sadb_lifetime_allocations = 0;
	p->sadb_lifetime_bytes = lft_bytes;
	p->sadb_lifetime_addtime = lft_time;
	p->sadb_lifetime_usetime = 0;

	*msg = buf;
	return 0;
}

/*
 * copy secasvar data into sadb_x_sa2.
 * `buf' must has been allocated sufficiently.
 */
static int
rcpfk_set_sadbxsa2(vchar_t **msg, struct rcpfk_msg *rc)
{
	vchar_t *buf;
	struct sadb_x_sa2 *p;
	int len, prevlen, extlen;

	extlen = sizeof(struct sadb_x_sa2);
	prevlen = (*msg)->l;
	len = prevlen + extlen;
	if ((buf = vreallocf(*msg, len)) == NULL) {
		rcpfk_seterror(rc, errno, "%s", strerror(errno));
		return -1;
	}

	p = (struct sadb_x_sa2 *)(buf->v + prevlen);
	p->sadb_x_sa2_len = PFKEY_UNIT64(extlen);
	p->sadb_x_sa2_exttype = SADB_X_EXT_SA2;
	p->sadb_x_sa2_mode = rct2pfk_samode(rc->samode);
	p->sadb_x_sa2_reqid = rc->reqid;

	*msg = buf;
	return 0;
}

/*
 * available policies:
 *     ESP transport
 *         esp/transport//require
 *     ESP tunnel
 *         esp/tunnel/a-b/require
 *     AH transport
 *         ah/transport//require
 *     AH tunnel
 *         ah/tunnel/a-b/require
 *     AH+ESP transport (IP|AH|ESP|ULP)
 *         esp/transport//require ah/transport//require
 *     AH+ESP tunnel    (IP1|AH|ESP|IP2|ULP)
 *         kame: esp/tunnel/a-b/require ah/transport/a-b/require
 */
static int
rcpfk_set_sadbxpolicy(vchar_t **msg, struct rcpfk_msg *rc, int type)
{
	switch (rc->samode) {
	case RCT_IPSM_TRANSPORT:
		return rcpfk_set_sadbxpolicy_transport(msg, rc, type);
	case RCT_IPSM_TUNNEL:
		return rcpfk_set_sadbxpolicy_tunnel(msg, rc, type);
	default:
		rcpfk_seterror(rc, EINVAL, "invalid samode=%d", rc->samode);
		return -1;
	}
}

#define SETXISR(p, l, t, m) \
do { \
	(p)->sadb_x_ipsecrequest_len = PFKEY_ALIGN8((l)); \
	(p)->sadb_x_ipsecrequest_proto = rct2ipproto_satype((t)); \
	(p)->sadb_x_ipsecrequest_mode = rct2pfk_samode((m)); \
	(p)->sadb_x_ipsecrequest_level = \
	    rct2pfk_seclevel(RCT_IPSL_REQUIRE); \
	(p)->sadb_x_ipsecrequest_reqid = 0; \
} while (0)

static int
rcpfk_set_sadbxpolicy_transport(vchar_t **msg, struct rcpfk_msg *rc, int type)
{
	vchar_t *buf;
	struct sadb_x_policy *xpl;
	struct sadb_x_ipsecrequest *xisr;
	int len, prevlen, extlen;

	extlen = sizeof(*xpl);
	if (type != SADB_X_SPDDELETE) {
		switch (rc->satype) {
		case RCT_SATYPE_AH:
		case RCT_SATYPE_ESP:
		case RCT_SATYPE_IPCOMP:
			extlen += PFKEY_ALIGN8(sizeof(*xisr));
			break;
		case RCT_SATYPE_AH_ESP:
		case RCT_SATYPE_AH_IPCOMP:
		case RCT_SATYPE_ESP_IPCOMP:
			extlen += PFKEY_ALIGN8(2 * sizeof(*xisr));
			break;
		case RCT_SATYPE_AH_ESP_IPCOMP:
			extlen += PFKEY_ALIGN8(3 * sizeof(*xisr));
			break;
		default:
			rcpfk_seterror(rc, EINVAL,
			    "invalid satype=%d", rc->satype);
			return -1;
		}
	}
	prevlen = (*msg)->l;
	len = prevlen + extlen;
	if ((buf = vreallocf(*msg, len)) == NULL) {
		rcpfk_seterror(rc, errno, "%s", strerror(errno));
		return -1;
	}

	xpl = (struct sadb_x_policy *)(buf->v + prevlen);
	xpl->sadb_x_policy_len = PFKEY_UNIT64(extlen);
	xpl->sadb_x_policy_exttype = SADB_X_EXT_POLICY;
	xpl->sadb_x_policy_type = IPSEC_POLICY_IPSEC;
	xpl->sadb_x_policy_dir = rct2pfk_dir(rc->dir);
	xpl->sadb_x_policy_id = rc->slid;

	if (type == SADB_X_SPDDELETE)
		goto end;

	xisr = (struct sadb_x_ipsecrequest *)(xpl + 1);
	switch (rc->satype) {
	case RCT_SATYPE_AH:
	case RCT_SATYPE_ESP:
	case RCT_SATYPE_IPCOMP:
		SETXISR(xisr, sizeof(*xisr), rc->satype, rc->samode);
		break;
	case RCT_SATYPE_AH_ESP:
		SETXISR(xisr, sizeof(*xisr), RCT_SATYPE_ESP, rc->samode);
		xisr++;
		SETXISR(xisr, sizeof(*xisr), RCT_SATYPE_AH, rc->samode);
		break;
	case RCT_SATYPE_AH_IPCOMP:
		SETXISR(xisr, sizeof(*xisr), RCT_SATYPE_IPCOMP, rc->samode);
		xisr++;
		SETXISR(xisr, sizeof(*xisr), RCT_SATYPE_AH, rc->samode);
		break;
	case RCT_SATYPE_ESP_IPCOMP:
		SETXISR(xisr, sizeof(*xisr), RCT_SATYPE_ESP, rc->samode);
		xisr++;
		SETXISR(xisr, sizeof(*xisr), RCT_SATYPE_IPCOMP, rc->samode);
		break;
	case RCT_SATYPE_AH_ESP_IPCOMP:
		SETXISR(xisr, sizeof(*xisr), RCT_SATYPE_ESP, rc->samode);
		xisr++;
		SETXISR(xisr, sizeof(*xisr), RCT_SATYPE_IPCOMP, rc->samode);
		xisr++;
		SETXISR(xisr, sizeof(*xisr), RCT_SATYPE_AH, rc->samode);
		break;
	default:
		rcpfk_seterror(rc, EINVAL, "invalid satype=%d", rc->satype);
		*msg = buf;	/* because it will be released by the caller */
		return -1;
	}

    end:
	*msg = buf;
	return 0;
}

static int
rcpfk_set_sadbxpolicy_tunnel(vchar_t **msg, struct rcpfk_msg *rc, int type)
{
	vchar_t *buf;
	struct sadb_x_policy *xpl;
	struct sadb_x_ipsecrequest *xisr;
	int len, prevlen, extlen;

	extlen = sizeof(*xpl);
	if (type != SADB_X_SPDDELETE) {
		switch (rc->satype) {
		case RCT_SATYPE_AH:
		case RCT_SATYPE_ESP:
		case RCT_SATYPE_IPCOMP:
			len = sizeof(*xisr) +
			    SA_LEN(rc->sa_src) + SA_LEN(rc->sa_dst);
			extlen += PFKEY_ALIGN8(len);
			break;
		case RCT_SATYPE_AH_ESP:
		case RCT_SATYPE_AH_IPCOMP:
		case RCT_SATYPE_ESP_IPCOMP:
			len = 2 * sizeof(*xisr) +
			    SA_LEN(rc->sa_src) + SA_LEN(rc->sa_dst);
			extlen += PFKEY_ALIGN8(len);
			break;
		case RCT_SATYPE_AH_ESP_IPCOMP:
			len = 3 * sizeof(*xisr) +
			    SA_LEN(rc->sa_src) + SA_LEN(rc->sa_dst);
			extlen += PFKEY_ALIGN8(len);
			break;
		default:
			rcpfk_seterror(rc, EINVAL,
			    "invalid satype=%d", rc->satype);
			return -1;
		}
	}
	prevlen = (*msg)->l;
	len = prevlen + extlen;
	if ((buf = vreallocf(*msg, len)) == NULL) {
		rcpfk_seterror(rc, errno, "%s", strerror(errno));
		return -1;
	}

	xpl = (struct sadb_x_policy *)(buf->v + prevlen);
	xpl->sadb_x_policy_len = PFKEY_UNIT64(extlen);
	xpl->sadb_x_policy_exttype = SADB_X_EXT_POLICY;
	xpl->sadb_x_policy_type = IPSEC_POLICY_IPSEC;
	xpl->sadb_x_policy_dir = rct2pfk_dir(rc->dir);
	xpl->sadb_x_policy_id = rc->slid;

	if (type == SADB_X_SPDDELETE)
		goto end;

	xisr = (struct sadb_x_ipsecrequest *)(xpl + 1);
	len =  sizeof(*xisr) + SA_LEN(rc->sa_src) + SA_LEN(rc->sa_dst);
	switch (rc->satype) {
	case RCT_SATYPE_AH:
	case RCT_SATYPE_ESP:
	case RCT_SATYPE_IPCOMP:
		SETXISR(xisr, len, rc->satype, RCT_IPSM_TUNNEL);
		break;
	case RCT_SATYPE_AH_ESP:
		SETXISR(xisr, len, RCT_SATYPE_ESP, RCT_IPSM_TUNNEL);
		xisr++;
		SETXISR(xisr, sizeof(*xisr), RCT_SATYPE_AH,
		    RCT_IPSM_TRANSPORT);
		break;
	case RCT_SATYPE_AH_IPCOMP:
		SETXISR(xisr, len, RCT_SATYPE_IPCOMP, RCT_IPSM_TUNNEL);
		xisr++;
		SETXISR(xisr, sizeof(*xisr), RCT_SATYPE_AH,
		    RCT_IPSM_TRANSPORT);
		break;
	case RCT_SATYPE_ESP_IPCOMP:
		SETXISR(xisr, len, RCT_SATYPE_ESP, RCT_IPSM_TUNNEL);
		xisr++;
		SETXISR(xisr, sizeof(*xisr), RCT_SATYPE_IPCOMP,
		    RCT_IPSM_TRANSPORT);
		break;
	case RCT_SATYPE_AH_ESP_IPCOMP:
		SETXISR(xisr, len, RCT_SATYPE_ESP, RCT_IPSM_TUNNEL);
		xisr++;
		SETXISR(xisr, sizeof(*xisr), RCT_SATYPE_IPCOMP,
		    RCT_IPSM_TRANSPORT);
		xisr++;
		SETXISR(xisr, sizeof(*xisr), RCT_SATYPE_AH,
		    RCT_IPSM_TRANSPORT);
		break;
	default:
		rcpfk_seterror(rc, EINVAL, "invalid satype=%d", rc->satype);
		*msg = buf;	/* because it will be release the caller */
		return -1;
	}

    end:
	*msg = buf;
	return 0;
}


/*
 * receiving modules
 */

/*
 * receive SADB_GETSPI from kernel.
 * OUT:
 *     rc->spi
 *     rc->sa_src
 *     rc->sa_dst
 */
static int
rcpfk_recv_getspi(caddr_t *mhp, struct rcpfk_msg *rc)
{
	struct sadb_msg *base;
	struct sadb_sa *sa;
	struct sockaddr *src, *dst;

	/* validity check */
	if (mhp[0] == 0 ||
	    mhp[SADB_EXT_SA] == 0 ||
	    mhp[SADB_EXT_ADDRESS_SRC] == 0 ||
	    mhp[SADB_EXT_ADDRESS_DST] == 0) {
		rcpfk_seterror(rc, EINVAL,
		    "inappropriate GETSPI message passed");
		return -1;
	}
	base = (struct sadb_msg *)mhp[0];
	sa = (struct sadb_sa *)mhp[SADB_EXT_SA];
	src = PFKEY_ADDR_SADDR(mhp[SADB_EXT_ADDRESS_SRC]);
	dst = PFKEY_ADDR_SADDR(mhp[SADB_EXT_ADDRESS_DST]);

	rc->seq = base->sadb_msg_seq;
	rc->satype = pfk2rct_satype(base->sadb_msg_satype);
	rc->spi = sa->sadb_sa_spi;
	rc->sa_src = (struct sockaddr *)&rc->sa_src_storage;
	rc->sa_dst = (struct sockaddr *)&rc->sa_dst_storage;
	memcpy(rc->sa_src, src, SA_LEN(src));
	memcpy(rc->sa_dst, dst, SA_LEN(dst));

	if (cb->cb_getspi != 0 && cb->cb_getspi(rc) < 0)
		return -1;

	return 0;
}

/*
 * receive SADB_UPDATE from kernel.
 * OUT:
 */
static int
rcpfk_recv_update(caddr_t *mhp, struct rcpfk_msg *rc)
{
	struct sadb_msg *base;
	struct sadb_sa *sa;
	struct sockaddr *src, *dst;
	u_int8_t samode;

	/* ignore this message in the case of the local test mode. */
	if (f_noharm)
		return 0;

	/* validity check */
	if (mhp[0] == 0 ||
	    mhp[SADB_EXT_SA] == 0 ||
	    mhp[SADB_EXT_ADDRESS_SRC] == 0 ||
	    mhp[SADB_EXT_ADDRESS_DST] == 0) {
		rcpfk_seterror(rc, EINVAL,
		    "inappropriate UPDATE message passed");
		return -1;
	}
	base = (struct sadb_msg *)mhp[0];
	sa = (struct sadb_sa *)mhp[SADB_EXT_SA];
	src = PFKEY_ADDR_SADDR(mhp[SADB_EXT_ADDRESS_SRC]);
	dst = PFKEY_ADDR_SADDR(mhp[SADB_EXT_ADDRESS_DST]);

	rc->seq = base->sadb_msg_seq;
	rc->satype = pfk2rct_satype(base->sadb_msg_satype);
	rc->spi = sa->sadb_sa_spi;
	rc->sa_src = (struct sockaddr *)&rc->sa_src_storage;
	rc->sa_dst = (struct sockaddr *)&rc->sa_dst_storage;
	memcpy(rc->sa_src, src, SA_LEN(src));
	memcpy(rc->sa_dst, dst, SA_LEN(dst));
	samode = mhp[SADB_X_EXT_SA2] == 0 ?
	    IPSEC_MODE_ANY :
	    ((struct sadb_x_sa2 *)mhp[SADB_X_EXT_SA2])->sadb_x_sa2_mode;
	rc->samode = samode == IPSEC_MODE_ANY ? 0 : pfk2rct_samode(samode);

	if (cb->cb_update != 0 && cb->cb_update(rc) < 0)
		return -1;

	return 0;
}

/*
 * receive SADB_ADD from kernel.
 * OUT:
 */
static int
rcpfk_recv_add(caddr_t *mhp, struct rcpfk_msg *rc)
{
	struct sadb_msg *base;
	struct sadb_sa *sa;
	struct sockaddr *src, *dst;
	u_int8_t samode;

	/* ignore this message in the case of the local test mode. */
	if (f_noharm)
		return 0;

	/* validity check */
	if (mhp[0] == 0 ||
	    mhp[SADB_EXT_SA] == 0 ||
	    mhp[SADB_EXT_ADDRESS_SRC] == 0 ||
	    mhp[SADB_EXT_ADDRESS_DST] == 0) {
		rcpfk_seterror(rc,
		    EINVAL, "inappropriate ADD message passed");
		return -1;
	}
	base = (struct sadb_msg *)mhp[0];
	src = PFKEY_ADDR_SADDR(mhp[SADB_EXT_ADDRESS_SRC]);
	dst = PFKEY_ADDR_SADDR(mhp[SADB_EXT_ADDRESS_DST]);
	sa = (struct sadb_sa *)mhp[SADB_EXT_SA];

	rc->seq = base->sadb_msg_seq;
	rc->satype = pfk2rct_satype(base->sadb_msg_satype);
	rc->spi = sa->sadb_sa_spi;
	rc->sa_src = (struct sockaddr *)&rc->sa_src_storage;
	rc->sa_dst = (struct sockaddr *)&rc->sa_dst_storage;
	memcpy(rc->sa_src, src, SA_LEN(src));
	memcpy(rc->sa_dst, dst, SA_LEN(dst));
	samode = mhp[SADB_X_EXT_SA2] == 0 ?
	    IPSEC_MODE_ANY :
	    ((struct sadb_x_sa2 *)mhp[SADB_X_EXT_SA2])->sadb_x_sa2_mode;
	rc->samode = samode == IPSEC_MODE_ANY ? 0 : pfk2rct_samode(samode);

	if (cb->cb_add != 0 && cb->cb_add(rc) < 0)
		return -1;

	return 0;
}

/*
 * receive SADB_EXPIRE from kernel.
 * OUT:
 */
static int
rcpfk_recv_expire(caddr_t *mhp, struct rcpfk_msg *rc)
{
	struct sadb_msg *base;
	struct sadb_sa *sa;
	struct sockaddr *src, *dst;
	u_int8_t samode;

	/* ignore this message in the case of the local test mode. */
	if (f_noharm)
		return 0;

	/* validity check */
	if (mhp[0] == 0 ||
	    mhp[SADB_EXT_SA] == 0 ||
	    mhp[SADB_EXT_ADDRESS_SRC] == 0 ||
	    mhp[SADB_EXT_ADDRESS_DST] == 0 ||
	    mhp[SADB_EXT_LIFETIME_CURRENT] == 0 ||
	    (mhp[SADB_EXT_LIFETIME_HARD] != 0 &&
	     mhp[SADB_EXT_LIFETIME_SOFT] != 0)) {
		rcpfk_seterror(rc, EINVAL,
		    "inappropriate EXPIRE message passed");
		return -1;
	}
	base = (struct sadb_msg *)mhp[0];
	sa = (struct sadb_sa *)mhp[SADB_EXT_SA];
	src = PFKEY_ADDR_SADDR(mhp[SADB_EXT_ADDRESS_SRC]);
	dst = PFKEY_ADDR_SADDR(mhp[SADB_EXT_ADDRESS_DST]);

	rc->seq = base->sadb_msg_seq;
	rc->satype = pfk2rct_satype(base->sadb_msg_satype);
	rc->spi = sa->sadb_sa_spi;
	rc->sa_src = (struct sockaddr *)&rc->sa_src_storage;
	rc->sa_dst = (struct sockaddr *)&rc->sa_dst_storage;
	memcpy(rc->sa_src, src, SA_LEN(src));
	memcpy(rc->sa_dst, dst, SA_LEN(dst));
	samode = mhp[SADB_X_EXT_SA2] == 0 ?
	    IPSEC_MODE_ANY :
	    ((struct sadb_x_sa2 *)mhp[SADB_X_EXT_SA2])->sadb_x_sa2_mode;
	rc->samode = samode == IPSEC_MODE_ANY ? 0 : pfk2rct_samode(samode);
	rc->expired = mhp[SADB_EXT_LIFETIME_HARD] != 0 ? 2 : 1;

	if (cb->cb_expire != 0 && cb->cb_expire(rc) < 0)
		return -1;

	return 0;
}

/*
 * receive SADB_ACQUIRE from kernel.
 * OUT:
 */
static int
rcpfk_recv_acquire(caddr_t *mhp, struct rcpfk_msg *rc)
{
	struct sadb_msg *base;
	struct sadb_x_policy *xpl;
	struct sockaddr *src, *dst;

	/* ignore this message in the case of the local test mode. */
	if (f_noharm)
		return 0;

	/* validity check */
	if (mhp[0] == 0 ||
	    mhp[SADB_EXT_ADDRESS_SRC] == 0 ||
	    mhp[SADB_EXT_ADDRESS_DST] == 0 ||
	    mhp[SADB_X_EXT_POLICY] == 0) {
		rcpfk_seterror(rc, EINVAL,
		    "inappropriate ACQUIRE message passed");
		return -1;
	}
	base = (struct sadb_msg *)mhp[0];
	xpl = (struct sadb_x_policy *)mhp[SADB_X_EXT_POLICY];
	src = PFKEY_ADDR_SADDR(mhp[SADB_EXT_ADDRESS_SRC]);
	dst = PFKEY_ADDR_SADDR(mhp[SADB_EXT_ADDRESS_DST]);

	/* ignore if type is not IPSEC_POLICY_IPSEC */
	if (xpl->sadb_x_policy_type != IPSEC_POLICY_IPSEC) {
		rcpfk_seterror(rc, 0, "ignore ACQUIRE message "
		    "bacause the type is not IPsec");
		return 0;
	}

	/* ignore it if src is multicast address */
	if ((dst->sa_family == AF_INET &&
	    IN_MULTICAST(ntohl(((struct sockaddr_in *)dst)->sin_addr.s_addr)))
#ifdef INET6
	    || (dst->sa_family == AF_INET6 &&
	    IN6_IS_ADDR_MULTICAST(&((struct sockaddr_in6 *)dst)->sin6_addr))
#endif
	) {
		rcpfk_seterror(rc, 0, "ignore ACQUIRE message "
		    "due to a multicast address");
		return 0;
	}

	rc->seq = base->sadb_msg_seq;
	rc->satype = pfk2rct_satype(base->sadb_msg_satype);
	rc->sa_src = (struct sockaddr *)&rc->sa_src_storage;
	rc->sa_dst = (struct sockaddr *)&rc->sa_dst_storage;
	memcpy(rc->sa_src, src, SA_LEN(src));
	memcpy(rc->sa_dst, dst, SA_LEN(dst));
#if 0	/* does acquire have sa2? */
	rc->samode = pfk2rct_samode(mhp[SADB_X_EXT_SA2] == 0
		? IPSEC_MODE_ANY
		: ((struct sadb_x_sa2 *)mhp[SADB_X_EXT_SA2])->sadb_x_sa2_mode);
#endif
	rc->slid = xpl->sadb_x_policy_id >> 4;	/* XXX quick hack */

	if (cb->cb_acquire != 0 && cb->cb_acquire(rc) < 0)
		return -1;

	return 0;
}

/*
 * receive SADB_DELETE from kernel.
 * OUT:
 */
static int
rcpfk_recv_delete(caddr_t *mhp, struct rcpfk_msg *rc)
{
	struct sadb_msg *base;
	struct sadb_sa *sa;
	struct sockaddr *src, *dst;

	/* ignore this message in the case of the local test mode. */
	if (f_noharm)
		return 0;

	/* validity check */
	if (mhp[0] == 0 ||
	    mhp[SADB_EXT_SA] == 0 ||
	    mhp[SADB_EXT_ADDRESS_SRC] == 0 ||
	    mhp[SADB_EXT_ADDRESS_DST] == 0) {
		rcpfk_seterror(rc, EINVAL,
		    "inappropriate DELETE message passed");
		return -1;
	}
	base = (struct sadb_msg *)mhp[0];
	sa = (struct sadb_sa *)mhp[SADB_EXT_SA];
	src = PFKEY_ADDR_SADDR(mhp[SADB_EXT_ADDRESS_SRC]);
	dst = PFKEY_ADDR_SADDR(mhp[SADB_EXT_ADDRESS_DST]);

	rc->seq = base->sadb_msg_seq;
	rc->satype = pfk2rct_satype(base->sadb_msg_satype);
	rc->spi = sa->sadb_sa_spi;
	rc->sa_src = (struct sockaddr *)&rc->sa_src_storage;
	rc->sa_dst = (struct sockaddr *)&rc->sa_dst_storage;
	memcpy(rc->sa_src, src, SA_LEN(src));
	memcpy(rc->sa_dst, dst, SA_LEN(dst));
#if 0	/* does delete have sa2? */
	rc->samode = pfk2rct_samode(mhp[SADB_X_EXT_SA2] == 0
		? IPSEC_MODE_ANY
		: ((struct sadb_x_sa2 *)mhp[SADB_X_EXT_SA2])->sadb_x_sa2_mode);
#endif

	if (cb->cb_delete != 0 && cb->cb_delete(rc) < 0)
		return -1;

	return 0;
}

/*
 * receive SADB_REGISTER from kernel.
 * OUT:
 */
static int
rcpfk_recv_register(caddr_t *mhp, struct rcpfk_msg *rc)
{
	struct sadb_msg *base;

	/* validity check */
	if (mhp[0] == 0) {
		rcpfk_seterror(rc, EINVAL,
		    "an invalid REGISTER message was passed");
		return -1;
	}
	base = (struct sadb_msg *)mhp[0];

	if (mhp[SADB_EXT_SUPPORTED_AUTH]) {
		if (set_supported_algorithm(mhp[SADB_EXT_SUPPORTED_AUTH],
		    &supported_map_auth)) {
    err:
			rcpfk_seterror(rc, 0, "%s",
			    strerror(base->sadb_msg_errno));
			return -1;
		}
	}
	if (mhp[SADB_EXT_SUPPORTED_ENCRYPT]) {
		if (set_supported_algorithm(mhp[SADB_EXT_SUPPORTED_ENCRYPT],
		    &supported_map_enc))
			goto err;
	}
#if defined(PFK_IPCOMP_SUPPORTED)
	if (mhp[SADB_EXT_X_SUPPORTED_IPCOMP]) {
		if (set_supported_algorithm(mhp[SADB_EXT_X_SUPPORTED_IPCOMP],
		    &supported_map_comp))
			goto err;
	}
#endif

	return 0;
}

static int
set_supported_algorithm(caddr_t m, struct sadb_supported **dstsup)
{
	struct sadb_supported *srcsup = (struct sadb_supported *)m;
	struct sadb_supported *sup;
	size_t len;

	len = PFKEY_EXTLEN(srcsup);
	if ((sup = rc_malloc(len)) == NULL)
		return -1;
	memcpy(sup, srcsup, len);

	if (*dstsup != NULL)
		rc_free(*dstsup);
	*dstsup = sup;

	return 0;
}

/*
 * NOTE: in case of the MIP6 security, the kernel passes the information
 * that the addresses changed by its protocol.  in this case, racoon
 * must change the addresses of the KMP session OR tells changing the
 * addresses to an appropriate daemon.
 */
static int
rcpfk_recv_spdupdate(caddr_t *mhp, struct rcpfk_msg *rc)
{
	struct sadb_x_policy *xpl;

	/* validity check */
	if (mhp[0] == 0 ||
	    mhp[SADB_X_EXT_POLICY] == 0) {
		rcpfk_seterror(rc, EINVAL,
		    "inappropriate SPDUPDATE message passed");
		return -1;
	}
	xpl = (struct sadb_x_policy *)mhp[SADB_X_EXT_POLICY];

	rc->slid = xpl->sadb_x_policy_id >> 4;	/* XXX quick hack */

	if (cb->cb_spdupdate != 0 && cb->cb_spdupdate(rc) < 0)
		return -1;

	return 0;
}

static int
rcpfk_recv_spdadd(caddr_t *mhp, struct rcpfk_msg *rc)
{
	struct sadb_x_policy *xpl;

	/* validity check */
	if (mhp[0] == 0 ||
	    mhp[SADB_X_EXT_POLICY] == 0) {
		rcpfk_seterror(rc, EINVAL,
		    "inappropriate SPDADD message passed");
		return -1;
	}
	xpl = (struct sadb_x_policy *)mhp[SADB_X_EXT_POLICY];

	rc->slid = xpl->sadb_x_policy_id >> 4;	/* XXX quick hack */

	if (cb->cb_spdadd != 0 && cb->cb_spdadd(rc) < 0)
		return -1;

	return 0;
}

static int
rcpfk_recv_spddelete(caddr_t *mhp, struct rcpfk_msg *rc)
{
	struct sadb_x_policy *xpl;

	/* validity check */
	if (mhp[0] == 0 ||
	    mhp[SADB_X_EXT_POLICY] == 0) {
		rcpfk_seterror(rc, EINVAL,
		    "inappropriate SPDDELETE message passed");
		return -1;
	}
	xpl = (struct sadb_x_policy *)mhp[SADB_X_EXT_POLICY];

	rc->slid = xpl->sadb_x_policy_id >> 4;	/* XXX quick hack */

	if (cb->cb_spddelete != 0 && cb->cb_spddelete(rc) < 0)
		return -1;

	return 0;
}

static int
rcpfk_recv_spddelete2(caddr_t *mhp, struct rcpfk_msg *rc)
{
	struct sadb_x_policy *xpl;

	/* validity check */
	if (mhp[0] == 0 ||
	    mhp[SADB_X_EXT_POLICY] == 0) {
		rcpfk_seterror(rc, EINVAL,
		    "inappropriate SPDDELETE message passed");
		return -1;
	}
	xpl = (struct sadb_x_policy *)mhp[SADB_X_EXT_POLICY];

	rc->slid = xpl->sadb_x_policy_id >> 4;	/* XXX quick hack */

	if (cb->cb_spddelete2 != 0 && cb->cb_spddelete2(rc) < 0)
		return -1;

	return 0;
}

static int
rcpfk_recv_spdexpire(caddr_t *mhp, struct rcpfk_msg *rc)
{
	struct sadb_x_policy *xpl;

	/* validity check */
	if (mhp[0] == 0 ||
	    mhp[SADB_X_EXT_POLICY] == 0) {
		rcpfk_seterror(rc, EINVAL,
		    "inappropriate SPDEXPIRE message passed");
		return -1;
	}
	xpl = (struct sadb_x_policy *)mhp[SADB_X_EXT_POLICY];

	rc->slid = xpl->sadb_x_policy_id >> 4;	/* XXX quick hack */

	if (cb->cb_spdexpire != 0 && cb->cb_spdexpire(rc) < 0)
		return -1;

	return 0;
}

static int
rcpfk_recv_spdget(caddr_t *mhp, struct rcpfk_msg *rc)
{
	/* validity check */
	if (mhp[0] == 0) {
		rcpfk_seterror(rc, EINVAL,
		    "inappropriate SPDGET message passed");
		return -1;
	}

	if (cb->cb_spdget != 0 && cb->cb_spdget(rc) < 0)
		return -1;

	return 0;
}

/*
 * check the algorithm is supported or not.
 * OUT:
 *    0      = supported
 *    others = not supported
 */
int
rcpfk_supported_auth(int algtype)
{
	int type;

	if ((type = rct2pfk_authtype(algtype)) == -1)
		return -1;

	return findsupportedalg(supported_map_auth, type) != NULL ? 1 : 0;
}

int
rcpfk_supported_enc(int algtype)
{
	int type;

	if ((type = rct2pfk_enctype(algtype)) == -1)
		return -1;

	return findsupportedalg(supported_map_enc, type) != NULL ? 1 : 0;
}

#if 0
int
rcpfk_supported_comp(int algtype)
{
	int type;

	if ((type = rct2pfk_comptype(algtype)) == -1)
		return -1;

	return findsupportedalg(supported_map_comp, type) != NULL ? 1 : 0;
}
#endif

static struct sadb_alg *
findsupportedalg(struct sadb_supported *sup, int alg_id)
{
	int tlen = PFKEY_UNUNIT64(sup->sadb_supported_len) - sizeof(*sup);
	struct sadb_alg *a = (struct sadb_alg *)(sup + 1);

	while (tlen > 0) {
		if (tlen < sizeof(*a)) {
			/* invalid format */
			break;
		}
		if (a->sadb_alg_id == alg_id)
			return a;

		tlen -= sizeof(*a);
		a++;
	}

	return NULL;
}
