/* $Id: main.c,v 1.86 2005/01/27 06:45:24 mk Exp $ */
/*
 * Copyright (C) 2004 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "spmd_includes.h"

#define DOMAIN_PORT 53

#ifdef SPMD_DEBUG
# define DPRINTF(format, args...) spmd_log(SPMD_L_DEBUG2, format, ## args)
#else
# define DPRINTF(format, args...)
#endif

/* ----- global ---- */
char spmd_version[] = SPMD_VERSION;
int spmd_foreground;
int spmd_nss;

/* static */
static void do_kill(void);
static void do_daemon(void);
static void print_version(void);
static void print_usage(void);
int check_nsswitchconf(void);


/*----- options -----*/
enum {
	__OPT_MIN = 0xff,
	OPT_CONFIG,
	OPT_FW,
	OPT_DEBUG_R,
	OPT_DEBUG_A,
	OPT_KILL,
	OPT_VERSION,
	OPT_HELP,
	__OPT_MAX
};

#ifdef HAVE_GETOPT_LONG
static struct option const longoptions[] = 
{
	{"config-file", required_argument, (int *)0, OPT_CONFIG},
	{"foreground",  no_argument,       (int *)0, OPT_FW},
	{"debug",	no_argument,       (int *)0, OPT_DEBUG_R},
	{"DEBUG",	required_argument, (int *)0, OPT_DEBUG_A},
	{"kill",	no_argument,	   (int *)0, OPT_KILL},
	{"version",     no_argument,       (int *)0, OPT_VERSION},
	{"help",	no_argument,       (int *)0, OPT_HELP},
	{0, 0, 0, 0}
};
#endif /* HAVE_GETOPT_LONG */
static const char *shortoptions = "f:p:D:VFdkh";

int
main(int argc, char **argv)
{
	int c;
#ifdef HAVE_GETOPT_LONG
	int option_index;
#endif /* HAVE_GETOPT_LONG */
	struct dns_server *dns = NULL;
	char config[PATH_MAX]; 
	int run_as_dns_proxy = 1;

	/* init global variables */
	spmd_foreground = 0;
	spmd_loglevel = SPMD_L_DEFLT;

	/* init local variables */
	memset(config, 0, sizeof(config));
	//strncpy(config, RACOON2_CONFIG_FILE, sizeof(config)-1); /* XXX */
	strlcpy(config, RACOON2_CONFIG_FILE, sizeof(config)); /* XXX */

	/* init libracoon */
	plog_init(0, NULL, argv[0], 1, 1);	/* XXX arg vals are correct? */
	if (rbuf_init(8, 80, 4, 160, 4) == -1) {/* XXX arg vals are correct? */
		spmd_log(SPMD_L_ERR, "rbuf_init() failed");
		exit(EXIT_FAILURE);
	}

	/* parse options */
#ifdef HAVE_GETOPT_LONG
	while ((c = getopt_long(argc, argv, shortoptions, 
					longoptions, &option_index)) != EOF) {
#else
	while ((c = getopt(argc, argv, shortoptions)) != EOF) {
#endif /* HAVE_GETOPT_LONG */
		switch (c) {
		case 'f':
		case OPT_CONFIG:
			//strncpy(config, optarg, sizeof(config)-1);
			strlcpy(config, optarg, sizeof(config));
			break;
		case 'F':
		case OPT_FW:
			spmd_foreground = 1;
			break;
		case 'd':
		case OPT_DEBUG_R:
			spmd_loglevel++;
			break;
		case 'D':
		case OPT_DEBUG_A:
			spmd_loglevel = atoi(optarg);
			if (spmd_loglevel <= SPMD_L_MIN || spmd_loglevel >= SPMD_L_MAX) {
				print_usage();
				exit(EXIT_FAILURE);
			}
			break;
		case 'k':
		case OPT_KILL:
			do_kill();
			exit(EXIT_SUCCESS);
			break;
		case 'V':
		case OPT_VERSION:
			print_version();
			exit(EXIT_SUCCESS);
			break;
		case 'h':
		case OPT_HELP:
		default:
			print_usage();
			exit(EXIT_SUCCESS);
			break;
		}
	}

	/* parse config */
	if (rcf_read(config, 0) < 0) {
		spmd_log(SPMD_L_ERR, "failed to parse config file:%s", config);
		exit(EXIT_FAILURE);
	}

	if (!spmd_foreground) {
		do_daemon();
	}

	spmd_log(SPMD_L_NOTICE, "Racoon Spmd - Security Policy Management Daemon - Started");
	spmd_log(SPMD_L_NOTICE, "Spmd Version: %s", rc_version());
	
	/*---------- task ----------*/
	task_init();

	/* run as DNS proxy ? */
	if (rcf_spmd_resolver()) {  /* get dns server addresses */
		struct dns_server *ds;
		struct sockaddr *sa;
		struct rc_addrlist *ns = NULL;
		struct rc_addrlist *nsp = NULL;

		run_as_dns_proxy = 1;
		
		if (rcf_get_resolvers(&ns)<0) {
			spmd_log(SPMD_L_ERR, "can not get resolver address");
			spmd_exit(EXIT_FAILURE);
		}

		for (nsp=ns; nsp; nsp=nsp->next) {
			if (nsp->type != RCT_ADDR_INET) {
				spmd_log(SPMD_L_ERR, "error: resolver address must be numeric");
				continue;
			}
			ds = dns_alloc();
			sa = &ds->sock.sa;
			sa->sa_family = nsp->a.ipaddr->sa_family;
			if (sa->sa_family == AF_INET) {
				struct sockaddr_in *sin;
				sin = (struct sockaddr_in *)sa;
				memcpy(sin, nsp->a.ipaddr, sizeof(*sin));
				if (nsp->port == 0)
					sin->sin_port = htons(DOMAIN_PORT);
				else
					sin->sin_port = htons(nsp->port);
			} else if (sa->sa_family == AF_INET6) {
				struct sockaddr_in6 *sin6;
				sin6 = (struct sockaddr_in6 *)sa;
				memcpy(sin6, nsp->a.ipaddr, sizeof(*sin6));
				if (nsp->port == 0)
					sin6->sin6_port = htons(DOMAIN_PORT);
				else
					sin6->sin6_port = htons(nsp->port);
			}
			if (dns) {
				dns = ds;
			} else {
				ds->next = dns;
				dns = ds;
			}
		}
		rcs_free_addrlist(ns);
		if (!dns) {
			spmd_log(SPMD_L_ERR, "error: no available dns server address");
			spmd_exit(EXIT_FAILURE);
		}
	} else { /* no */
		run_as_dns_proxy = 0;
	}

	/*---------- nsswitch.conf ----------*/
	spmd_nss = check_nsswitchconf();
	if (spmd_nss < 0) {
		spmd_log(SPMD_L_ERR, "read error: %s", NSSWITCH_CONF_FILE);
		spmd_exit(EXIT_FAILURE);
	}

	/*---------- dns proxy ----------*/
	if ( run_as_dns_proxy && (spmd_nss & NSS_DNS) ) {
		/*-------- setup dns socket --------*/
		dnsl_init();
		if (dns) {
			struct dns_server *head = dns;

			while (dns) {
				dns->s = setup_dns_sock(&dns->sock.sa);
				if (dns->s < 0) {
					spmd_log(SPMD_L_ERR, "can't setup dns socket");
					spmd_exit(EXIT_FAILURE);
				}
				dns=dns->next;
			}
			dnsl_add(head);
		} else {
			spmd_log(SPMD_L_ERR, "dns server: not specified");
			spmd_exit(EXIT_FAILURE);
		}
		spmd_add_dns_task();
	}
	
	/*-------- setup resolver sockets --------*/
	if ( run_as_dns_proxy) {
		struct rc_addrlist *ns_bounds = NULL;
		if (rcf_get_dns_queries(&ns_bounds)<0) {
			spmd_log(SPMD_L_ERR, "read error: dns query addresses: %s", config);
			spmd_exit(EXIT_FAILURE);
		}
		spmd_init_resolver_sock(ns_bounds);
		rcs_free_addrlist(ns_bounds);
	}

	/*--------- Policy ----------*/
	spmd_pfkey_init();

	/*---------- hosts file ----------*/
	//if ( run_as_dns_proxy && (spmd_nss & NSS_FILES) ) {
	if (spmd_nss & NSS_FILES) {
		if (hosts_cache_update() < 0)
			spmd_exit(EXIT_FAILURE);
	}

	/*--------- FQDN ----------*/
	if (run_as_dns_proxy)
		fqdn_query_task_register();

	/*--------- shell ----------*/
	shell_init();

	/*--------- setup signals ----------*/
	init_signal();

	/*--------- main loop ----------*/
	if (task_loop() < 0) {
		spmd_log(SPMD_L_ERR, "fatal error");
		spmd_exit(EXIT_FAILURE);
	}

	rcf_clean();
	return 0;
}

/* ---------- */
static void
print_version(void)
{
	fprintf(stderr,
		"Version: %s\n", rc_version());
	return;
}

static void
print_usage(void)
{
        fprintf(stderr,
                "Options:\n"
#ifdef HAVE_GETOPT_LONG
		"\t-V|--version            : show version number\n"
		"\t-f|--config-file [FILE] : specify config file\n"
		"\t-d|--debug              : increase debug level\n"
		"\t-D|--DEBUG  Debug Level : specify debug level(1-6)\n"
		"\t-F|--foreground         : run foreground\n"
		"\t-k|--kill               : kill running spmd\n"
                "\t-h|--help               : show this help\n"
#else
		"\t-V             : show version number\n"
		"\t-f             : specify config file\n"
		"\t-d             : increase debug level\n"
		"\t-D Debug Level : specify debug level(1-6)\n"
		"\t-F             : run foreground\n"
		"\t-k             : kill running spmd\n"
                "\t-h             : show this help\n"
#endif
		);
	return;
}

void
spmd_exit(int status)
{
	int err;

	spmd_log(SPMD_L_NOTICE, "Exit");
	if (!spmd_foreground) {
		err = unlink(SPMD_PID_FILE);
		if (err < 0) {
			spmd_log(SPMD_L_ERR, "failed to unlink file '%s' : %s",
				SPMD_PID_FILE, strerror(errno));
		}
	}

	exit(status);
}

static void
do_kill(void)
{
	pid_t pid;

	pid = read_pid(SPMD_PID_FILE);
	if (pid < 0) {
		spmd_log(SPMD_L_ERR, "can't read pidfile");
		return ;
	}
	kill(pid, SIGTERM);
	return;
}

static void
do_daemon(void)
{
	openlog("spmd", LOG_PID, LOG_DAEMON);
	if (daemon(0, 0) < 0) { 
		perror("daemon()"); 
#ifdef __linux__ /* glibc specific ? */
		if (errno == 0) {
			spmd_log(SPMD_L_ERR, 
			"/dev/null may not be a character device with the expected major and minor numbers, check please");
		}
#endif
		exit(EXIT_FAILURE); 
	} 
	umask(0);
	if (write_pid(SPMD_PID_FILE, getpid()) < 0) {
		spmd_log(SPMD_L_ERR, "can't write pid file");
		spmd_log(SPMD_L_ERR, "already running? <pid=%d>", read_pid(SPMD_PID_FILE));
		exit(EXIT_FAILURE); 
	}
	return;
}

int
check_nsswitchconf(void)
{
#if defined(HAVE_NSSWITCH_CONF) /* For Linux, FeeBSD5 and NetBSD */

	FILE *fp;
	char buf[BUFSIZ];
	char *ap, *cp, *ep;
	int ret=0;

	fp = fopen(NSSWITCH_CONF_FILE, "r");
	if (!fp) 
		return -1;

	while ( (ap=fgets(buf, sizeof(buf), fp)) ) {
		if (*ap == '#')
			continue;

		cp = strpbrk(ap, "#\n");
		if (!cp) /* strange */
			continue;
		*cp = '\0';

		cp = strpbrk(ap, ":");
		if (!cp) /* empty ? */
			continue;
		*cp = '\0';
		cp++;

		if (strncasecmp(ap, "hosts", strlen(ap))) 
			continue;

		while (cp && *cp) {
			while (*cp == ' ' || *cp == '\t')
				cp++;
			ep = cp;
			cp = strpbrk(cp, " \t");
			if (cp) {
				*cp = '\0';
				cp++;
			}

			if (!strncasecmp(ep, "files", strlen(ap))) {
				spmd_log(SPMD_L_INFO, "\'files\' found in nsswitch.conf hosts line, we will read hosts file");
				ret |= NSS_FILES;
			}

			if (!strncasecmp(ep, "dns", strlen(ap))) {
				spmd_log(SPMD_L_INFO, "\'dns\' found in nsswitch.conf hosts line, we will start dns proxy service");
				ret |= NSS_DNS;
			}
		}
	}

#elif defined(HAVE_HOST_CONF) /* For FreeBSD4 */

	FILE *fp;
	char buf[BUFSIZ];
	char *ap, *ep;
	int k;
	int ret=0;
	struct {
		char *key;
		int flag;
	} nsk[] = {
		{ "files",	NSS_FILES, },
		{ "hosts",	NSS_FILES, },
		{ "hosttable",	NSS_FILES, },
		{ "httable",	NSS_FILES, },
		{ "bind",	NSS_DNS, },
		{ "dns",	NSS_DNS, },
		{ "domain",	NSS_DNS, },
	};

	fp = fopen(NSSWITCH_CONF_FILE, "r");
	if (!fp) 
		return -1;

	while ( (ap=fgets(buf, sizeof(buf), fp)) ) {
		if (*ap == '#')
			continue;

		ep = strpbrk(ap, "\t #\n");
		if (!ep) /* empty ? */
			continue;
		*ep = '\0';

		for (k = 0; k < sizeof(nsk)/sizeof(nsk[0]); k++) {
			if (!strncasecmp(ap, nsk[k].key, strlen(ap))) {
				spmd_log(SPMD_L_INFO, "\'%s\' found in %s",
					nsk[k].key, NSSWITCH_CONF_FILE);
				ret |= nsk[k].flag;
			}
		}
	}
#elif defined(HAVE_LOOKUP_IN_RESOLV_CONF) /* For OpenBSD */
	/* 
	 * If the lookup keyword is not used 
	 * in the system's resolv.conf file 
	 * then the assumed order is "bind file".
	 */
	FILE *fp;
	char buf[BUFSIZ];
	char *ap, *cp, *ep;
	int ret=0;

	fp = fopen(NSSWITCH_CONF_FILE, "r");
	if (!fp) {
		/* 
		 * The system's resolv.conf file does not exist, 
		 * then the only database used is "file".
		 */
		spmd_log(SPMD_L_INFO, "can not open %s file. we will read only hosts file.", NSSWITCH_CONF_FILE);
		ret |= NSS_FILES;
		return ret;
	}

	while ( (ap=fgets(buf, sizeof(buf), fp)) ) {
		if (*ap == '#')
			continue;

		cp = strpbrk(ap, "#\n");
		if (!cp) /* strange */
			continue;
		*cp = '\0';

		cp = strpbrk(ap, " \t");
		if (!cp) /* empty ? */
			continue;
		*cp = '\0';
		cp++;

		if (strncasecmp(ap, "lookup", strlen(ap))) 
			continue;

		while (cp && *cp) {
			while (*cp == ' ' || *cp == '\t')
				cp++;
			ep = cp;
			cp = strpbrk(cp, " \t");
			if (cp) {
				*cp = '\0';
				cp++;
			}

			if (!strncasecmp(ep, "file", strlen(ap))) {
				spmd_log(SPMD_L_INFO, "\'file\' found in %s lookup line, we will read hosts file", NSSWITCH_CONF_FILE);
				ret |= NSS_FILES;
			}

			if (!strncasecmp(ep, "bind", strlen(ap))) {
				spmd_log(SPMD_L_INFO, "\'bind\' found in %s lookup line, we will start dns proxy service", NSSWITCH_CONF_FILE);
				ret |= NSS_DNS;
			}
		}
	}
#else
#error "Please try to port check_nsswitchconf() to your OS"
#endif

	fclose(fp);

	return ret;
}
