/* $Id: shell.c,v 1.59 2005/01/24 06:59:31 fukumoto Exp $ */ 
/*
 * Copyright (C) 2003 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "spmd_includes.h"

#define DEFAULT_BACKLOG 5

#define SPMD_SHELL_BUFSIZ	512

struct shell_sock {
	struct shell_sock *next;
	int s;
	union { /* we dont use sockaddr_storage. */
		struct sockaddr sa;
		struct sockaddr_in sin;
		struct sockaddr_in6 sin6;
		struct sockaddr_un sun;
	} sock;
};
static struct shell_sock *shhead = NULL;

static int shell_banner(int s);
static int shell_sock_open_sa(const struct sockaddr *sa);
static struct sockaddr *shell_build_sock_unix(const char *path);
static int shell_sock_open_file(const struct sockaddr *sa);
int shell_accept(struct task *t);
static int shell_interpreter(struct task *t);
static int shell_cmd_handler(int sh_argc, char **sh_argv, struct task *t);
static int shell_unknown_handler(int sh_argc, char **sh_argv, struct task *t);
static int shell_login_handler(int sh_argc, char **sh_argv, struct task *t);
static int shell_ns_handler(int sh_argc, char **sh_argv, struct task *t);
static int shell_fqdn_handler(int sh_argc, char **sh_argv, struct task *t);
static int shell_policy_handler(int sh_argc, char **sh_argv, struct task *t);
static int shell_stat_handler(int sh_argc, char **sh_argv, struct task *t);
static int shell_quit_handler(int sh_argc, char **sh_argv, struct task *t);

enum command { LOGIN, NS, FQDN, POLICY, STAT, QUIT, EXIT, __END };
typedef struct {
	enum command cmd;
	char	string[16];
	int	(*func)(int sh_argc, char **sh_argv, struct task *t);
} shell_handler_t;
static shell_handler_t sh_hdl[] = {
	{ LOGIN,"LOGIN", shell_login_handler, },
	{ NS, "NS", shell_ns_handler, },
	{ FQDN, "FQDN", shell_fqdn_handler, },
	{ POLICY, "POLICY", shell_policy_handler, },
	{ STAT, "STAT", shell_stat_handler, },
	{ QUIT, "QUIT", shell_quit_handler, },
	{ EXIT, "EXIT", shell_quit_handler, },
	{ __END, "", NULL, },
};

static int
shell_banner(int s)
{
	char buf[SPMD_SHELL_BUFSIZ];
	char status[5] = "220 ";
	struct utsname uts;
	struct timeval tv;
	struct tm *tm;
	char date[32];
	int n,len;
	int ret=0;

	gettimeofday(&tv, 0);
	tm = localtime(&tv.tv_sec);
	strftime(date, sizeof(date), "%a, %d %b %Y %T %z", tm);

	uname(&uts);

	snprintf(buf, sizeof(buf), "%s%s Spmd %s %s %s\r\n", 
			status, uts.nodename, spmd_version, uts.sysname, date);

	len = strlen(buf);
	n = write(s, buf, len);
	if (n!=len)
		ret = -1;
	
	return ret;
}

static int
shell_sock_open_sa(const struct sockaddr *sa)
{
	int on = 1;
	int s;
	int backlog = DEFAULT_BACKLOG;

	s = socket(sa->sa_family, SOCK_STREAM, 0);
	if (s<0) {
		spmd_log(SPMD_L_ERR, "%s", strerror(errno));
		goto fin; 
	} 

	if (sa->sa_family == AF_INET6) {
		if (setsockopt(s, IPPROTO_IPV6,IPV6_V6ONLY, &on, sizeof(on)) < 0) { 
			spmd_log(SPMD_L_ERR, "%s", strerror(errno));
			close(s);
			s = -1;
			goto fin;
		}
		if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) < 0) { 
			spmd_log(SPMD_L_ERR, "%s", strerror(errno));
			close(s);
			s = -1;
			goto fin;
		}
		if (bind(s, sa, rcs_getsalen(sa)) < 0) {
			spmd_log(SPMD_L_ERR, "%s", strerror(errno));
			close(s);
			s = -1;
			goto fin;
		}
		if (listen(s, backlog) < 0) {
			spmd_log(SPMD_L_ERR, "%s", strerror(errno));
			close(s);
			s = -1;
			goto fin;
		}
	}
	else if (sa->sa_family == AF_INET) {
		if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) < 0) { 
			spmd_log(SPMD_L_ERR, "%s", strerror(errno));
			close(s);
			s = -1;
			goto fin;
		}
		if (bind(s, sa, rcs_getsalen(sa)) < 0) {
			spmd_log(SPMD_L_ERR, "%s", strerror(errno));
			close(s);
			s = -1;
			goto fin;
		}
		if (listen(s, backlog) < 0) {
			spmd_log(SPMD_L_ERR, "%s", strerror(errno));
			close(s);
			s = -1;
			goto fin;
		}
	}

fin:
	return s;
}

static struct sockaddr *
shell_build_sock_unix(const char *path)
{
	struct sockaddr_un *sun = NULL;

	sun = spmd_calloc(M_SHELL, sizeof(struct sockaddr_un));
	if (!sun) {
		spmd_log(SPMD_L_ERR, "can not allocate memory");
		return NULL;
	}

	if (strlen(path) >= sizeof(sun->sun_path)) {
		spmd_log(SPMD_L_ERR, "path too long");
		spmd_free(M_SHELL, sun);
		return NULL;
	}

	sun->sun_family = AF_UNIX;
	strcpy(sun->sun_path, path);
#ifdef HAVE_SA_LEN
	sun->sun_len = SUN_LEN(sun);
#endif
	unlink(path);

	return (struct sockaddr *)sun;
}

static int
shell_sock_open_file(const struct sockaddr *sa)
{
	int s = -1;
	int backlog = DEFAULT_BACKLOG;

	if (!sa) {
		goto fin;
	}

	s = socket(PF_UNIX, SOCK_STREAM, 0);
	if (s<0) {
		spmd_log(SPMD_L_ERR, "%s", strerror(errno));
		goto fin; 
	} 

	if (bind(s, sa, SUN_LEN((struct sockaddr_un *)sa)) < 0) {
		spmd_log(SPMD_L_ERR, "%s", strerror(errno));
		close(s);
		s = -1;
		goto fin;
	}
	if (listen(s, backlog) < 0) {
		spmd_log(SPMD_L_ERR, "%s", strerror(errno));
		close(s);
		s = -1;
		goto fin;
	}

fin:
	return s;
}

void
shell_init(void)
{
	struct shell_sock *sh, *p;
	struct task *t;
	struct rc_addrlist *rcl_top = NULL, *rcl;
	int fd = -1;
	struct sockaddr *sa = NULL;


	rcf_get_spmd_interfaces(&rcl_top);
	/* XXX check rcl_top */
	for (rcl=rcl_top; rcl; rcl=rcl->next) {
		switch (rcl->type) {
		case RCT_ADDR_INET:
			//sa = rcs_sadup(rcl->a.ipaddr);
			fd = shell_sock_open_sa(rcl->a.ipaddr);
			if (fd < 0) {
				//rc_free(sa);
				continue;
			}
			break;
		case RCT_ADDR_FILE:
			sa = shell_build_sock_unix(vmem2str(rcl->a.vstr));
			fd = shell_sock_open_file(sa);
			if (fd < 0) {
				spmd_free(M_SHELL, sa);
				continue;
			}
			break;
		case RCT_ADDR_FQDN:
			{
				char *fqdn = NULL;
				char portstr[16];
				struct addrinfo hints, *res0, *res;
				int gai_err;
				char host[NI_MAXHOST];

				fqdn = (char *)vmem2str(rcl->a.vstr);
				memset(portstr, 0, sizeof(portstr));
				if (rcl->port == 0) {
					snprintf(portstr, sizeof(portstr), "%d", SPMD_SHELL_PORT);
				} else {
					snprintf(portstr, sizeof(portstr), "%d", rcl->port);
				}

				memset(&hints, 0, sizeof(hints));
				hints.ai_family = PF_UNSPEC;
				hints.ai_socktype = SOCK_STREAM;
				gai_err = getaddrinfo(fqdn, portstr, &hints, &res0);
				if (gai_err < 0) {
					spmd_log(SPMD_L_ERR, "%s", gai_strerror(gai_err));
					continue;
				}
				for (res=res0; res; res=res->ai_next) {
					getnameinfo(res->ai_addr, res->ai_addrlen, host, sizeof(host), NULL, 0, NI_NUMERICHOST);
					if (res->ai_family == AF_INET) {
						if (!strncmp(host, "::1", strlen(host))) {
							continue;
						}
					} else if (res->ai_family == AF_INET) {
						if (!strncmp(host, "127.0.0.1", strlen(host))) {
							continue;
						}
					} else {
						continue;
					}
					fd = shell_sock_open_sa(res->ai_addr);
					/* XXX */
					sh = (struct shell_sock *)spmd_calloc(M_SHELL, sizeof(struct shell_sock));
					sh->s = fd;
					memcpy(&sh->sock.sa, rcl->a.ipaddr, rcs_getsalen(rcl->a.ipaddr));
					if (!shhead) {
						shhead = sh;
					} else {
						p = shhead;
						while (p->next)
							p=p->next;
						p->next = sh;
					}
				}
				freeaddrinfo(res0);
				continue;
			}
			break;
		default:
			continue;
			break;
		}
		
		sh = (struct shell_sock *)spmd_calloc(M_SHELL, sizeof(struct shell_sock));
		sh->s = fd;
		memcpy(&sh->sock.sa, rcl->a.ipaddr, rcs_getsalen(rcl->a.ipaddr));
		if (!shhead) {
			shhead = sh;
		} else {
			p = shhead;
			while (p->next)
				p=p->next;
			p->next = sh;
		}
	}

	sh = shhead;
	while (sh) {
		t = task_alloc(0); 
		t->fd = sh->s;
		t->flags = 0;
		t->sa = &sh->sock.sa; 
		t->salen = sizeof(sh->sock);/* cant use SPMD_SALEN() */
		t->func = shell_accept;
		task_list_add(t, &spmd_task_root->read);
		sh=sh->next;
	}

	if (!shhead) {
		spmd_log(SPMD_L_ERR, "can not initialize racoon interface");
		spmd_exit(EXIT_FAILURE);
	}

	return;

}


int
shell_accept(struct task *t)
{
	int cli_sock;
	int on = 1;
	int rtn;
	struct task *newt;

        int s = t->fd;
	struct sockaddr *sa = t->sa;
	socklen_t salen  = t->salen;

	if ( (cli_sock = accept(s, (struct sockaddr *)sa, &salen)) < 0) {
		spmd_log(SPMD_L_ERR, "accept(2) failed");
		return -1;
	}

	rtn = setsockopt(cli_sock, SOL_SOCKET, SO_KEEPALIVE, &on, sizeof (on));
	if (rtn < 0) {
		spmd_log(SPMD_L_ERR, "setsockopt(SO_KEEPALIVE) failed");
		close(cli_sock);
		return -1;
	}
	rtn = setsockopt(cli_sock, IPPROTO_TCP, TCP_NODELAY, &on, sizeof (on));
	if (rtn < 0 && (sa->sa_family != AF_UNIX) ) {
		spmd_log(SPMD_L_ERR, "setsockopt(TCP_NODELAY) failed");
		close(cli_sock);
		return -1;
	}

	/* display banner message */
	shell_banner(cli_sock);

	newt = task_alloc(0);
	newt->fd = cli_sock;
	newt->func = shell_interpreter;
	task_list_add(newt, &spmd_task_root->read);

	/* re-add myself */
	newt = task_alloc(0);
	newt->fd = s;
	newt->flags = 0;
	newt->sa = t->sa;
	newt->salen = t->salen;
	newt->func = shell_accept;
	task_list_add(newt, &spmd_task_root->read);

	return 0;
}


static int
shell_interpreter(struct task *t)
{
	int n;
	char cmd[SPMD_SHELL_BUFSIZ];
	char *buf, *bufhead;
	char *cp;
	int sh_argc=0;
	char **sh_argv;
	int ret;
	int s = t->fd;

	memset(cmd, 0, sizeof(cmd));
	n = read(s, cmd, sizeof(cmd));
	if (n<=0) {
		spmd_log(SPMD_L_ERR, "socket:%d, no read", s);
		close(s); /* kill CLOSE_WAIT */
		return -1;
	}

	bufhead = buf = (char *)spmd_calloc(M_TMP, n);

	cp = cmd;

	/* skip head space */
	while (*cp && isblank(*cp)) cp++;

	if (!*cp) { /* empty line */
		return -1;
	}

	sh_argv = (char **)spmd_malloc(M_TMP, sizeof(*sh_argv)*n);
	if (sh_argv == NULL) {
		spmd_log(SPMD_L_ERR, "%s: can not allocate memory", __FUNCTION__);
		return -1;
	}
	sh_argv[sh_argc] = buf;
	while (*cp) {
		if (isspace(*cp)) {
			while (*cp && isblank(*cp)) cp++;
			*buf = '\0';
			if (*cp == '\r') 
				break;
			buf++;
			++sh_argc;
			sh_argv[sh_argc] = buf; /* new element */
		}
		if (*cp == '\r') {
			*buf = '\0';
			break;
		} 
		*buf++ = *cp++;
	}
	sh_argc++;


	ret = shell_cmd_handler(sh_argc, sh_argv, t);

	spmd_free(M_TMP, bufhead);
	spmd_free(M_TMP, sh_argv);

	if (ret<0)
		spmd_log(SPMD_L_DEBUG, "shell command failed");
	
	return 0;
}

/* return value:
 *              0: success
 *              1: exit
 *             -1: error
 */
static int
shell_cmd_handler(int sh_argc, char **sh_argv, struct task *t)
{
	int i;
	int ret;


	for (i=0;i<__END;i++) {
		if (!strncasecmp(sh_argv[0], sh_hdl[i].string, strlen(sh_hdl[i].string))) {
			ret = sh_hdl[i].func(sh_argc-1, sh_argv+1, t);
			goto fin;
		}
	}

	ret = shell_unknown_handler(sh_argc-1, sh_argv+1, t);

fin:
	return ret;
}

static int
shell_unknown_handler(int sh_argc, char **sh_argv, struct task *t)
{
	char buf[SPMD_SHELL_BUFSIZ];
	char status[5] = "502 ";
	int s = t->fd;
	struct task *newt;
	int n,len;
	int ret=0;

	//strncpy(status, "502 ", sizeof(status)-1);
	strlcpy(status, "502", sizeof(status));
	snprintf(buf, sizeof(buf), "%sCommand not implemented\r\n", status);
	len = strlen(buf);
	n = write(s, buf, len);
	if (n!=len)
		ret = -1;

	newt = task_alloc(0);
	newt->fd = t->fd;
	newt->authenticated = t->authenticated;
	newt->func = shell_interpreter;
	task_list_add(newt, &spmd_task_root->read);
	return ret;
}

/* This function is TEMPORARY! 
 * I'll replace this with a read-config function.
 */
#define SPMD_MAX_PASSWD		16
char *
cfg_get_password(void)
{
	char password[SPMD_MAX_PASSWD] = "racoon2"; /* XXX */
	return (spmd_strdup(M_TMP, password));
}
/* ditto */
int
spmd_passwd_check(char *str)
{
	size_t ret;
	size_t len,plen,slen;
	char *passwd = cfg_get_password();

	if (!str||!passwd) {
		ret = -1;
		goto fin;
	}

	plen = strlen(passwd);
	slen = strlen(str);

	len = slen >= plen ? plen : -1;
	if (len < 0) {
		ret = -1;
		goto fin;
	}

	ret = strncmp(passwd, str, len); 

fin:
	spmd_free(M_TMP, passwd);

	return ret;
}

static int 
shell_login_handler(int sh_argc, char **sh_argv, struct task *t)
{
	char buf[SPMD_SHELL_BUFSIZ];
	char status[5] = "250 ";
	int ret=0;
	int s = t->fd;
	struct task *newt;
	int n,len;

	memset(buf, 0, sizeof(buf));

	if (sh_argc != 1) {
		goto err;
	}

	if (!spmd_passwd_check(sh_argv[0])) {
		//strncpy(status, "250 ", sizeof(status)-1);
		strlcpy(status, "250 ", sizeof(status));
		snprintf(buf, sizeof(buf), "%sOK\r\n", status);
		t->authenticated=1;
		goto fin;
	}

err:
	//strncpy(status, "550 ", sizeof(status)-1);
	strlcpy(status, "550 ", sizeof(status));
	snprintf(buf, sizeof(buf), "%sFAILED\r\n", status);

fin:
	len = strlen(buf);
	n = write(s, buf, len);
	if (n != len)
		ret = -1;

	newt = task_alloc(0);
	newt->fd = t->fd;
	newt->authenticated = t->authenticated;
	newt->func = shell_interpreter;
	task_list_add(newt, &spmd_task_root->read);

	return ret; 
}

static int
shell_ns_handler(int sh_argc, char **sh_argv, struct task *t)
{
	char buf[SPMD_SHELL_BUFSIZ];
	char addr[INET6_ADDRSTRLEN];
	struct dns_server *dns = NULL;
	struct task *newt, *newt2;
	int err;
	char status[5] = "250-";
	struct addrinfo hints, *res;
	int ret = 0;
	int s = t->fd;
	int n, len;

	memset(buf, 0, sizeof(buf));

	if (NEED_LOGIN(t)) {
		//strncpy(status, "550 ", sizeof(status)-1);
		strlcpy(status, "550 ", sizeof(status));
		snprintf(buf, sizeof(buf), "%sNeed Login\r\n", status);
		goto wfin;
	}

	if (!(spmd_nss & NSS_DNS)) {
		//strncpy(status, "550 ", sizeof(status)-1);
		strlcpy(status, "550 ", sizeof(status));
		snprintf(buf, sizeof(buf), "%sOperation Failed, check %s\r\n", status, NSSWITCH_CONF_FILE);
		goto wfin;
	}

	if (sh_argc == 1) { 
		if (!strncasecmp(sh_argv[0], "LIST", strlen("LIST"))) { 
			dns = dsl->live;
			do {
				err = getnameinfo((struct sockaddr *)&dns->sock.sa, SPMD_SALEN(&dns->sock.sa),
								addr, sizeof(addr), NULL, 0, NI_NUMERICHOST);
				if (err) {
					//strncpy(status, "550 ", sizeof(status)-1);
					strlcpy(status, "550 ", sizeof(status));
					snprintf(buf, sizeof(buf), "%sInternal Error\r\n", status);
					goto wfin; 
				}
					
				if (dns->next == dsl->live)
					status[3] = ' ';
				snprintf(buf, sizeof(buf), "%s%s\r\n", status, addr);
				len = strlen(buf);
				n = write(s, buf, len);
				if (n!=len)
					ret = -1;
				dns = dns->next;
			} while (dns != dsl->live);
			ret = 0;
			goto fin;
		} else {
			//strncpy(status, "500 ", sizeof(status)-1);
			strlcpy(status, "500 ", sizeof(status));
			snprintf(buf, sizeof(buf), "%sSyntax Error\r\n", status);
			ret = -1;
			goto wfin;
		}
	} else if (sh_argc == 2) {
		if (!strncasecmp(sh_argv[0], "ADD", strlen("ADD"))) { 
			memset(&hints, 0, sizeof(hints));
			hints.ai_socktype = SOCK_DGRAM;
			hints.ai_flags = AI_NUMERICHOST;
			err = getaddrinfo(sh_argv[1], NULL, &hints, &res);
			if (err) {
				spmd_log(SPMD_L_ERR, "%s", gai_strerror(err));
				//strncpy(status, "550 ", sizeof(status)-1);
				strlcpy(status, "550 ", sizeof(status));
				snprintf(buf, sizeof(buf), "%sOperation Failed\r\n", status);
				ret = -1;
				goto wfin;
			}
			dns = dnsl_find(res->ai_addr);
			if (dns) {
				//strncpy(status, "551 ", sizeof(status)-1);
				strlcpy(status, "551 ", sizeof(status));
				snprintf(buf, sizeof(buf), "%s%s Already Exists\r\n", status, sh_argv[1]);
				ret = -1;
				goto wfin;
			}

			newt = task_alloc_dns(res->ai_addr);
			if (!newt) {
				//strncpy(status, "550 ", sizeof(status)-1);
				strlcpy(status, "550 ", sizeof(status));
				snprintf(buf, sizeof(buf), "%sOperation Failed\r\n", status);
				ret = -1;
				goto wfin;
			}
			dns = newt->dns;

			task_list_add(newt, &spmd_task_root->read);

			dsl->live = dns;

			//strncpy(status, "250 ", sizeof(status)-1);
			strlcpy(status, "250 ", sizeof(status));
			snprintf(buf, sizeof(buf), "%s%s Added\r\n", status, sh_argv[1]);
			goto wfin;
		} else if (!strncasecmp(sh_argv[0], "DELETE", strlen("DELETE"))) {
			memset(&hints, 0, sizeof(hints));
			hints.ai_socktype = SOCK_DGRAM;
			hints.ai_flags = AI_NUMERICHOST;
			err = getaddrinfo(sh_argv[1], NULL, &hints, &res);
			if (err) {
				spmd_log(SPMD_L_ERR, "%s", gai_strerror(err));
				//strncpy(status, "550 ", sizeof(status)-1);
				strlcpy(status, "550 ", sizeof(status));
				snprintf(buf, sizeof(buf), "%sOperation Failed\r\n", status);
				ret = -1;
				goto wfin;
			}
			dns = dnsl_find(res->ai_addr);
			if (!dns) {
				//strncpy(status, "551 ", sizeof(status)-1);
				strlcpy(status, "551 ", sizeof(status));
				snprintf(buf, sizeof(buf), "%sDoes Not Exist\r\n", status);
				ret = -1;
				goto wfin;
			}
			freeaddrinfo(res);
			if (dnsl_del(dns) < 0) {
				//strncpy(status, "550 ", sizeof(status)-1);
				strlcpy(status, "550 ", sizeof(status));
				snprintf(buf, sizeof(buf), "%sInternal Error or At least need one dns server in the system\r\n", status);
				ret = -1;
				goto wfin;
			}
			task_list_del(dns->t, &spmd_task_root->read);
			{
				struct task *delt, *delq;
				delt = dns->t;

				if (spmd_task_root->delq == NULL) { 
					spmd_task_root->delq = delt; 
				} else { 
					delq = spmd_task_root->delq; 
					while (delq->next) 
						delq =  delq->next; 
					delq->next = delt; 
				}
			}
			dns->t->dns_deleted = 1;
			//strncpy(status, "250 ", sizeof(status)-1);
			strlcpy(status, "250 ", sizeof(status));
			snprintf(buf, sizeof(buf), "%s%s Deleted\r\n", status, sh_argv[1]);
			goto wfin;
		} else if (!strncasecmp(sh_argv[0], "CHANGE", strlen("CHANGE"))) {
			memset(&hints, 0, sizeof(hints));
			hints.ai_socktype = SOCK_DGRAM;
			hints.ai_flags = AI_NUMERICHOST;
			err = getaddrinfo(sh_argv[1], NULL, &hints, &res);
			if (err) {
				spmd_log(SPMD_L_ERR, "%s", gai_strerror(err));
				//strncpy(status, "550 ", sizeof(status)-1);
				strlcpy(status, "550 ", sizeof(status));
				snprintf(buf, sizeof(buf), "%sOperation Failed\r\n", status);
				ret = -1;
				goto wfin;
			}
			dns = dnsl_find(res->ai_addr);
			if (!dns) {
				//strncpy(status, "550 ", sizeof(status)-1);
				strlcpy(status, "550 ", sizeof(status));
				snprintf(buf, sizeof(buf), "%sCan not find %s in the our list.\r\n", status, sh_argv[1]);
				ret = -1;
				goto wfin;
			}
			freeaddrinfo(res);
			dsl->live = dns;
			//strncpy(status, "250 ", sizeof(status)-1);
			strlcpy(status, "250 ", sizeof(status));
			snprintf(buf, sizeof(buf), "%s%s changed.\r\n", status, sh_argv[1]);
			goto wfin;
		} else {
			//strncpy(status, "500 ", sizeof(status)-1);
			strlcpy(status, "500 ", sizeof(status));
			snprintf(buf, sizeof(buf), "%sSyntax Error\r\n", status);
			ret = -1;
			goto wfin;
		}
	} else {
		//strncpy(status, "500 ", sizeof(status)-1);
		strlcpy(status, "500 ", sizeof(status));
		snprintf(buf, sizeof(buf), "%sSyntax Error\r\n", status);
		ret = -1;
		goto wfin;
	}

wfin:
	len = strlen(buf);
	n = write(s, buf, len);
	if (n!=len)
		ret = -1;
fin:
	newt2 = task_alloc(0);
	newt2->fd = t->fd;
	newt2->authenticated = t->authenticated;
	newt2->func = shell_interpreter;
	task_list_add(newt2, &spmd_task_root->read);
	return ret;
}

static int
shell_fqdn_handler(int sh_argc, char **sh_argv, struct task *t)
{
	int ret=0;
	char status[5] = "250-";
	char buf[SPMD_SHELL_BUFSIZ];
	struct task *newt;
	struct addrinfo hints, *res;
	struct cache_entry *ce;
	struct fqdn_list *fl;
	int fqdn_len;
	int err;
	int s = t->fd;
	int n,len;
	char fqdn[MAX_NAME_LEN];


	memset(buf, 0, sizeof(buf));

	if (NEED_LOGIN(t)) {
		//strncpy(status, "550 ", sizeof(status)-1);
		strlcpy(status, "550 ", sizeof(status));
		snprintf(buf, sizeof(buf), "%sneed login\r\n", status);
		goto fin;
	}

	if (sh_argc == 1) {
		if (!strncasecmp(sh_argv[0], "LIST", strlen("LIST"))) { 
			fl = get_fqdn_db_top();
			if (!fl) { /* no fqdn registered */
				//strncpy(status, "251 ", sizeof(status)-1);
				strlcpy(status, "251 ", sizeof(status));
				snprintf(buf, sizeof(buf), "%sNo FQDN Registered\r\n", status);
				goto fin;
			}
			while (fl) { 
				if (fl->next != NULL) {
					snprintf(buf, sizeof(buf), "%s%s\r\n", status, fl->fqdn);
				} else {
					status[3] = ' ';
					snprintf(buf, sizeof(buf), "%s%s\r\n", status, fl->fqdn);
				}
				len = strlen(buf);
				n = write(s, buf, len);
				if (n!=len)
					ret = -1;
				fl=fl->next;
			}
			goto fin2;
		} else {
			goto serr;
		}
	} else if (sh_argc == 2) {
		if (!strncasecmp(sh_argv[0], "QUERY", strlen("QUERY"))) { 
			memset(&hints, 0, sizeof(hints));
			hints.ai_socktype = SOCK_DGRAM;
			hints.ai_flags = AI_NUMERICHOST;
			err = getaddrinfo(sh_argv[1], NULL, &hints, &res);
			if (err) {
				spmd_log(SPMD_L_ERR, "%s", gai_strerror(err));
				//strncpy(status, "550 ", sizeof(status)-1);
				strlcpy(status, "550 ", sizeof(status));
				snprintf(buf, sizeof(buf), "%sOperation Failed\r\n", status);
				ret = -1;
				goto fin;
			}
			ce = find_cache_entry(res->ai_addr);
			if (!ce) {
				//strncpy(status, "501 ", sizeof(status)-1);
				strlcpy(status, "501 ", sizeof(status));
				snprintf(buf, sizeof(buf), "%sNot Cached\r\n", status);
				goto fin;
			}
			freeaddrinfo(res);
			fl = ce->fltop;
			if (!fl) {
				//strncpy(status, "550 ", sizeof(status)-1);
				strlcpy(status, "550 ", sizeof(status));
				snprintf(buf, sizeof(buf), "%sInternal Error\r\n", status);
				goto fin;
			}
			while (fl) {
				if (fl->next != NULL) {
					snprintf(buf, sizeof(buf), "%s%s\r\n", status, fl->fqdn);
				} else {
					status[3] = ' ';
					snprintf(buf, sizeof(buf), "%s%s\r\n", status, fl->fqdn);
				}
				len = strlen(buf);
				n = write(s, buf, len);
				if (n!=len)
					ret = -1;
				fl=fl->next;
			}
			goto fin2;
		} else if (!strncasecmp(sh_argv[0], "ADD", strlen("ADD"))) { 
			fqdn_len = strlen(sh_argv[1]);
			if (fqdn_len < MAX_NAME_LEN) {
				//strncpy(fqdn, sh_argv[1], fqdn_len+1);
				strlcpy(fqdn, sh_argv[1], sizeof(fqdn));
				if (fqdn[fqdn_len-1] != '.') {
					if ((fqdn_len+1) < MAX_NAME_LEN) {
						//strncat(fqdn, ".", 2);
						strlcat(fqdn, ".", sizeof(fqdn));
					} else {
						goto serr;
					}
				}
				fqdn_len = strlen(fqdn);
				fl = find_fqdn_db(fqdn, fqdn_len);
				if (fl) {
					//strncpy(status, "550 ", sizeof(status)-1);
					strlcpy(status, "550 ", sizeof(status));
					snprintf(buf, sizeof(buf), "%sFQDN Already Registered\r\n", status);
				} else {
					add_fqdn_db(fqdn, fqdn_len);
					hosts_cache_update();
					snprintf(buf, sizeof(buf), "%sFQDN Registered\r\n", status);
				}
				goto fin;
			} else  {
				goto serr;
			}
		} else if (!strncasecmp(sh_argv[0], "DELETE", strlen("DELETE"))) { 
			fqdn_len = strlen(sh_argv[1]);
			if (fqdn_len < MAX_NAME_LEN) {
				//strncpy(fqdn, sh_argv[1], fqdn_len+1);
				strlcpy(fqdn, sh_argv[1], sizeof(fqdn));
				if (fqdn[fqdn_len-1] != '.') {
					if ((fqdn_len+1) < MAX_NAME_LEN) {
						//strncat(fqdn, ".", 2);
						strlcat(fqdn, ".", sizeof(fqdn));
					} else {
						goto serr;
					}
				}
				fqdn_len = strlen(fqdn);
				fl = find_fqdn_db(fqdn, fqdn_len);
				if (fl) {
					del_fqdn_db(fl);
					del_cache_entry_by_fqdn(fqdn, fqdn_len);
					snprintf(buf, sizeof(buf), "%sFQDN Deregistered\r\n", status);
				} else {
					//strncpy(status, "550 ", sizeof(status)-1);
					strlcpy(status, "550 ", sizeof(status));
					snprintf(buf, sizeof(buf), "%sNo such a FQDN exists\r\n", status);
				}
				goto fin;
			} else {
				goto serr;
			}
		} else {
			goto serr;
		}
	} else {
		goto serr;
	}

serr:
	//strncpy(status, "500 ", sizeof(status)-1);
	strlcpy(status, "500 ", sizeof(status));
	snprintf(buf, sizeof(buf), "%sSyntax Error\r\n", status);

fin:
	len = strlen(buf);
	n = write(s, buf, len);
	if (n!=len)
		ret = -1;

fin2:
	newt = task_alloc(0);
	newt->fd = t->fd;
	newt->authenticated = t->authenticated;
	newt->func = shell_interpreter;
	task_list_add(newt, &spmd_task_root->read);
	return ret;
}

/* temporary code */
static int
shell_policy_handler(int sh_argc, char **sh_argv, struct task *t)
{
	char status[5] = "250-";
	char buf[SPMD_SHELL_BUFSIZ];
	struct task *newt;
	int s = t->fd;
	int n,len;
	int ret=0;
	int selector_index;
	uint64_t lifetime = 0;
	struct addrinfo hints, *sres, *dres;
	int gai_err;
	struct rcpfk_msg *rc = &pfkey_container;
	rc_type samode;
	struct rcf_selector *sl = NULL;

	memset(buf, 0, sizeof(buf));

	if (NEED_LOGIN(t)) {
		//strncpy(status, "550 ", sizeof(status)-1);
		strlcpy(status, "550 ", sizeof(status));
		snprintf(buf, sizeof(buf), "%sneed login\r\n", status);
		goto fin;
	}

	if (sh_argc != 6 && sh_argc != 8) {  /* XXX */
		//strncpy(status, "500 ", sizeof(status)-1);
		strlcpy(status, "500 ", sizeof(status));
		snprintf(buf, sizeof(buf), "%sSyntax Error\r\n", status);
		goto fin;
	}

	if (!strncasecmp(sh_argv[0], "ADD", strlen("ADD"))) { 
		/* XXX XXX XXX */
		selector_index = atoi(sh_argv[1]);
		lifetime = atoll(sh_argv[2]); /* XXX 64-bit */
		if (!strncasecmp(sh_argv[3], "TUNNEL", strlen("TUNNEL"))) {
			samode = RCT_IPSM_TUNNEL;
			/* N/A */
			//strncpy(status, "550 ", sizeof(status)-1);
			strlcpy(status, "550 ", sizeof(status));
			snprintf(buf, sizeof(buf), "%sOperation Failed\r\n", status);
			ret = -1;
			goto fin;
		} else if (!strncasecmp(sh_argv[3], "TRANSPORT", strlen("TRANSPORT"))) {
			samode = RCT_IPSM_TRANSPORT;
		} else {
			//strncpy(status, "500 ", sizeof(status)-1);
			strlcpy(status, "500 ", sizeof(status));
			snprintf(buf, sizeof(buf), "%sSyntax Error\r\n", status);
			ret = -1;
			goto fin;
		}
		
		/* src */
		memset(&hints, 0, sizeof(hints));
		hints.ai_socktype = SOCK_DGRAM;
		hints.ai_flags = AI_NUMERICHOST;
		gai_err = getaddrinfo(sh_argv[4], NULL, &hints, &sres);
		if (gai_err) {
			spmd_log(SPMD_L_ERR, "%s", gai_strerror(gai_err));
			//strncpy(status, "550 ", sizeof(status)-1);
			strlcpy(status, "550 ", sizeof(status));
			snprintf(buf, sizeof(buf), "%sOperation Failed\r\n", status);
			ret = -1;
			goto fin;
		}
		/* dst */
		memset(&hints, 0, sizeof(hints));
		hints.ai_socktype = SOCK_DGRAM;
		hints.ai_flags = AI_NUMERICHOST;
		gai_err = getaddrinfo(sh_argv[5], NULL, &hints, &dres);
		if (gai_err) {
			spmd_log(SPMD_L_ERR, "%s", gai_strerror(gai_err));
			//strncpy(status, "550 ", sizeof(status)-1);
			strlcpy(status, "550 ", sizeof(status));
			snprintf(buf, sizeof(buf), "%sOperation Failed\r\n", status);
			ret = -1;
			freeaddrinfo(sres);
			goto fin;
		}
		/* *** Do actual policy operations *** */
		/* src to dst policy */
		rcf_get_selector(selector_index, &sl);
		sl_to_rc_wo_addr(sl, rc);

		rc->lft_hard_time = lifetime;

		rc->sp_src = sres->ai_addr;
		if (sres->ai_family == AF_INET) {
			rc->pref_src = 32;
			((struct sockaddr_in *)rc->sp_src)->sin_port = htons(sl->src->port);
		} else if (sres->ai_family == AF_INET6) {
			rc->pref_src = 128;
			((struct sockaddr_in6 *)rc->sp_src)->sin6_port = htons(sl->src->port);
		}
		rc->sp_dst = dres->ai_addr;
		if (dres->ai_family == AF_INET) {
			rc->pref_dst = 32;
			((struct sockaddr_in *)rc->sp_dst)->sin_port = htons(sl->dst->port);
		} else if (dres->ai_family == AF_INET6) {
			rc->pref_dst = 128;
			((struct sockaddr_in6 *)rc->sp_dst)->sin6_port = htons(sl->dst->port);
		}
		if (spmd_spd_update(rc)<0) {
			//strncpy(status, "550 ", sizeof(status)-1);
			strlcpy(status, "550 ", sizeof(status));
			snprintf(buf, sizeof(buf), "%sOperation Failed\r\n", status);
			rcf_free_selector(sl);
			goto fin;
		}
		/* dst to src policy */
		rc->slid++;
		rc->dir = (rc->dir == RCT_DIR_INBOUND) ? RCT_DIR_OUTBOUND : RCT_DIR_INBOUND;
		rc->sp_dst = sres->ai_addr;
		rc->sp_src = dres->ai_addr;
		if (spmd_spd_update(rc)<0) {
			//strncpy(status, "550 ", sizeof(status)-1);
			strlcpy(status, "550 ", sizeof(status));
			snprintf(buf, sizeof(buf), "%sOperation Failed\r\n", status);
			rcf_free_selector(sl);
			goto fin;
		}

		rcf_free_selector(sl);

		freeaddrinfo(sres);
		freeaddrinfo(dres);

		//strncpy(status, "250 ", sizeof(status)-1);
		strlcpy(status, "250 ", sizeof(status));
		snprintf(buf, sizeof(buf), "%sPolicy Added\r\n", status);

		goto fin;
	} else {
		//strncpy(status, "500 ", sizeof(status)-1);
		strlcpy(status, "500 ", sizeof(status));
		snprintf(buf, sizeof(buf), "%sSyntax Error\r\n", status);
		goto fin;
	}

fin:
	len = strlen(buf);
	n = write(s, buf, len);
	if (n!=len)
		ret = -1;

	newt = task_alloc(0);
	newt->fd = t->fd;
	newt->authenticated = t->authenticated;
	newt->func = shell_interpreter;
	task_list_add(newt, &spmd_task_root->read);
	return ret;
}

static int 
shell_stat_handler(int sh_argc, char **sh_argv, struct task *t)
{
	char buf[SPMD_SHELL_BUFSIZ];
	char status[5] = "250-";
	int i;
	int s = t->fd;
	struct task *newt;
	int ret=0;
	int n, len;

	memset(buf, 0, sizeof(buf));

	if (NEED_LOGIN(t)) {
		//strncpy(status, "550 ", sizeof(status)-1);
		strlcpy(status, "550 ", sizeof(status));
		snprintf(buf, sizeof(buf), "%sneed login\r\n", status);
		len = strlen(buf);
		n = write(s, buf, len);
		if (n!=len)
			ret = -1;
		goto fin;
	}

	if (sh_argc != 0) { 
		//strncpy(status, "500 ", sizeof(status)-1);
		strlcpy(status, "500 ", sizeof(status));
		snprintf(buf, sizeof(buf), "%sSyntax Error\r\n", status);
		len = strlen(buf);
		n = write(s, buf, len);
		if (n!=len)
			ret = -1;
		goto fin;
	}

	for (i=0;i<Q_END;i++) {
		//strncpy(status, "250-", sizeof(status)-1);
		strlcpy(status, "250-", sizeof(status));
		snprintf(buf, sizeof(buf), "%s%s = %d\r\n", status, qstat[i].name, qstat[i].number);
		len = strlen(buf);
		n = write(s, buf, len);
		if (n!=len)
			ret = -1;
	}

	for (i=0;i<M_END;i++) {
		//strncpy(status, "250-", sizeof(status)-1);
		strlcpy(status, "250-", sizeof(status));
		snprintf(buf, sizeof(buf), "%s%s = %d\r\n", status, mstat[i].name, mstat[i].allocated);
		len = strlen(buf);
		n = write(s, buf, len);
		if (n!=len)
			ret = -1;
	}

	for (i=0;i<C_END;i++) {
		//strncpy(status, "250-", sizeof(status)-1);
		strlcpy(status, "250-", sizeof(status));
		if ((i+1) == C_END) status[3] = ' '; /* end check */
		snprintf(buf, sizeof(buf), "%s%s = %d\r\n", status, cstat[i].name, cstat[i].number);
		len = strlen(buf);
		n = write(s, buf, len);
		if (n!=len)
			ret = -1;
	}

fin:
	newt = task_alloc(0);
	newt->fd = t->fd;
	newt->authenticated = t->authenticated;
	newt->func = shell_interpreter;
	task_list_add(newt, &spmd_task_root->read);

	return ret;
}

static int
shell_quit_handler(int sh_argc, char **sh_argv, struct task *t)
{
	char buf[SPMD_SHELL_BUFSIZ];
	char status[5] = "250 ";
	int s = t->fd;
	int n,len;
	int ret=0;

	snprintf(buf, sizeof(buf), "%sBYE\r\n", status);
	len = strlen(buf);
	n = write(s, buf, len);
	if (n!=len)
		ret = -1;

	close(s);
	return ret;
}

