/* $Id: spmd_pfkey.c,v 1.28 2005/01/07 09:02:09 sakane Exp $ */
/*
 * Copyright (C) 2004 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "spmd_includes.h"

#ifdef SPMD_DEBUG
# define DPRINTF(format, args...) spmd_log(SPMD_L_DEBUG2, format, ## args)
#else
# define DPRINTF(format, args...)
#endif

#define SAT_AH		1
#define SAT_ESP		2
#define SAT_IPCOMP	4

int pfkey_sock;
struct rcpfk_msg pfkey_container;

/*
 * sp_queue_top-->spq-->spq-->NULL
 *                 |
 *                 +-->fl
 *                 |
 *                 +-->fl
 */
struct sp_queue {
	struct sp_queue *next;
	int sl_index;

	rc_type src_type;
	union {
		struct fqdn_list *src_fl;
		struct sockaddr *src_sa;
	} src;

	rc_type dst_type;
	union {
		struct fqdn_list *dst_fl;
		struct sockaddr *dst_sa;
	} dst;
};

int sp_queue_add(int sl_index, const struct rc_addrlist *src, const struct rc_addrlist *dst);
struct sp_queue *sp_queue_search(int sl_index);

static struct rcpfk_cb pfkey_callback;

static int pfkey_receiver(struct task *t);

static int spmd_sp_init(struct rcf_selector *sl, struct rcpfk_msg *rc);

static int spmd_pfkey_spdadd(struct rcpfk_msg *rc);
static int spmd_pfkey_spddelete(struct rcpfk_msg *rc);

static void spmd_rcpfk_cont_sock_free(struct rcpfk_msg *rc);


int
spmd_pfkey_init(void)
{
	struct rcf_selector *sl_head = NULL;
	struct rcf_selector *sl = NULL;
	int spd_add_skip = 0;
	char *fqdn_str = NULL;
	size_t fqdn_strlen = 0;
	struct rc_addrlist *rcals = NULL, *rcald = NULL;
	struct task *t = NULL;

	memset(&pfkey_container, 0, sizeof(pfkey_container));
	memset(&pfkey_callback, 0, sizeof(pfkey_callback));
	/* set callback */
	pfkey_callback.cb_spdadd = &spmd_pfkey_spdadd;
	pfkey_callback.cb_spddelete = &spmd_pfkey_spddelete;

	if (rcpfk_init(&pfkey_container, &pfkey_callback) < 0) {
		return -1;
	}

	pfkey_sock = pfkey_container.so;

	if (rcf_get_selectorlist(&sl_head) < 0) {
		return -1;
	}

	/* set initial policy */
	for (sl = sl_head;sl;sl=sl->next) {
		spd_add_skip = 0;
		if (!sl->pl)
			continue;
		rcals = sl->src;
		rcald = sl->dst;
		if (rcals->type == RCT_ADDR_FQDN && rcald->type == RCT_ADDR_FQDN) {
			fqdn_str = (char *)vmem2str(rcals->a.vstr); /* add src to fqdn_db */
			fqdn_strlen = strlen(fqdn_str);
			add_fqdn_db(fqdn_str, fqdn_strlen);
			fqdn_str = (char *)vmem2str(rcald->a.vstr); /* add dst to fqdn_db */
			fqdn_strlen = strlen(fqdn_str);
			add_fqdn_db(fqdn_str, fqdn_strlen);
			/* add sp_queue */
			sp_queue_add(sl->sl_index, rcals, rcald);
			spd_add_skip=1;
		}
		else if (rcals->type == RCT_ADDR_FQDN && rcald->type == RCT_ADDR_INET) {
			fqdn_str = (char *)vmem2str(rcals->a.vstr); /* add src to fqdn_db */
			fqdn_strlen = strlen(fqdn_str);
			add_fqdn_db(fqdn_str, fqdn_strlen);
			/* add sp_queue */
			sp_queue_add(sl->sl_index, rcals, rcald);
			spd_add_skip=1;
		} 
		else if (rcals->type == RCT_ADDR_INET && rcald->type == RCT_ADDR_FQDN) {
			fqdn_str = (char *)vmem2str(rcald->a.vstr); /* add dst to fqdn_db */
			fqdn_strlen = strlen(fqdn_str);
			add_fqdn_db(fqdn_str, fqdn_strlen);
			/* add sp_queue */
			sp_queue_add(sl->sl_index, rcals, rcald);
			spd_add_skip=1;
		}

		if (spd_add_skip) {
			continue;
		}
		if (spmd_sp_init(sl, &pfkey_container) < 0) {
			continue;
		}
	}

	rcf_free_selector(sl_head);

	t = task_alloc(0);
	t->fd = pfkey_sock;
	t->msg = &pfkey_container;
	t->func = pfkey_receiver;
	task_list_add(t, &spmd_task_root->read);
	return 0;
}

static int
pfkey_receiver(struct task *t)
{
	struct task *newt = NULL;

	DPRINTF("%s:called", __FUNCTION__);

	/* just discard */
	rcpfk_handler((struct rcpfk_msg *)t->msg);


	/* re-add myself */
	newt = task_alloc(0);
	newt->fd = t->fd;
	newt->msg = t->msg;
	newt->flags = t->flags;
	newt->func = t->func;
	task_list_add(newt, &spmd_task_root->read);

	return 0;
}

/* just add only non-FQDN policies */
static int
spmd_sp_init(struct rcf_selector *sl, struct rcpfk_msg *rc)
{
	struct rcf_policy *pl = NULL;
	struct rcf_ipsec *ips = NULL;
	struct rc_addrlist *al = NULL;
	struct rc_addrlist *ipal = NULL;
	uint32_t satype = 0;
	const char *macro;

	if (!sl->pl) {
		return -1;
	}
	pl = sl->pl;

	if (!sl->pl->ips) {
		return -1;
	}
	ips = sl->pl->ips;

	/* check rcf_ipsec{} sa_* set or NULL */
	if (ips->sa_ah) {
		satype |= SAT_AH;
	}
	if (ips->sa_esp) {
		satype |= SAT_ESP;
	}
	if (ips->sa_ipcomp) {
		satype |= SAT_IPCOMP;
	}
	if ( (satype&SAT_ESP) && !(satype&SAT_AH) && !(satype&SAT_IPCOMP) ) {
		rc->satype = RCT_SATYPE_ESP; 
	}
	else if ( !(satype&SAT_ESP) && (satype&SAT_AH) && !(satype&SAT_IPCOMP) ) {
		rc->satype = RCT_SATYPE_AH;
	}
	else if ( !(satype&SAT_ESP) && !(satype&SAT_AH) && (satype&SAT_IPCOMP) ) {
		rc->satype = RCT_SATYPE_IPCOMP;
	}
	else if ( (satype&SAT_ESP) && (satype&SAT_AH) && !(satype&SAT_IPCOMP) ) {
		rc->satype = RCT_SATYPE_AH_ESP;
	}
	else if ( !(satype&SAT_ESP) && (satype&SAT_AH) && (satype&SAT_IPCOMP) ) {
		rc->satype = RCT_SATYPE_AH_IPCOMP;
	}
	else if ( (satype&SAT_ESP) && !(satype&SAT_AH) && (satype&SAT_IPCOMP) ) {
		rc->satype = RCT_SATYPE_ESP_IPCOMP;
	}
	else if ( (satype&SAT_ESP) && (satype&SAT_AH) && (satype&SAT_IPCOMP) ) {
		rc->satype = RCT_SATYPE_AH_ESP_IPCOMP;
	}
	else {
		spmd_log(SPMD_L_ERR, "unknown SA type");
		return -1;
	}

	/* tunnel or transport */
	rc->samode = ips->ipsec_mode;

	/* set policy lifetime */
	rc->lft_hard_time = 0; /* at init time */

	rc->flags = 0;

	/* XXX do we need to check pl->action? */
	if (rc->samode == RCT_IPSM_TUNNEL) {
		if (pl->my_sa_addr) { /* always single entry */
			al = pl->my_sa_addr;
			if (al->type == RCT_ADDR_MACRO) { /* check type MACRO */
				rcs_getaddrlistbymacro(al->a.vstr, &ipal);
				rc->sa_src = rcs_sadup(ipal->a.ipaddr);
				rcs_free_addrlist(ipal);
				ipal = NULL;
			} else if (al->type == RCT_ADDR_INET) {
				rc->sa_src = rcs_sadup(al->a.ipaddr);
			} else if (al->type == RCT_ADDR_FQDN) {
				return -1; /* XXX */
			} else {
				return -1;
			}
		} else { return -1; } /* XXX */
		if (pl->peers_sa_addr) { /* always single entry */
			al = pl->peers_sa_addr;
			if (al->type == RCT_ADDR_MACRO) { /* check type MACRO */
				rcs_getaddrlistbymacro(al->a.vstr, &ipal);
				rc->sa_dst = rcs_sadup(ipal->a.ipaddr);
				rcs_free_addrlist(ipal);
				ipal = NULL;
			} else if (al->type == RCT_ADDR_INET) {
				rc->sa_dst = rcs_sadup(al->a.ipaddr);
			} else if (al->type == RCT_ADDR_FQDN) {
				return -1; /* XXX */
			} else {
				return -1;
			}
		} else { return -1; } /* XXX */
	}

	/* selector */
	rc->slid = (sl->sl_index<<SPMD_SLID_SHIFT_WIDTH); /* selector index */

	if (sl->src) { /* XXX do we need to care multiple entries? */
		al = sl->src;
		if (al->type == RCT_ADDR_MACRO) { /* check type MACRO */
			macro = vmem2str(al->a.vstr);
			if (!strncmp(macro, "IP_ANY", strlen(macro))) {
				goto err;
			}
			rcs_getaddrlistbymacro(al->a.vstr, &ipal);
			rc->sp_src = rcs_sadup(ipal->a.ipaddr);
			rcs_free_addrlist(ipal);
			ipal = NULL;
		} else if (al->type == RCT_ADDR_INET) {
			rc->sp_src = rcs_sadup(al->a.ipaddr);
		} else if (al->type == RCT_ADDR_FQDN) {
			goto err; /* XXX */
		} else {
			goto err;
		}
	} else {
		goto err;
	}
	rc->pref_src = sl->src->prefixlen;

	if (sl->dst) { /* XXX do we need to care multiple entries? */
		al = sl->dst;
		if (al->type == RCT_ADDR_MACRO) { /* check type MACRO */
			macro = vmem2str(al->a.vstr);
			if (!strncmp(macro, "IP_ANY", strlen(macro))) {
				goto err;
			}
			rcs_getaddrlistbymacro(al->a.vstr, &ipal);
			rc->sp_dst = rcs_sadup(ipal->a.ipaddr);
			rcs_free_addrlist(ipal);
			ipal = NULL;
		} else if (al->type == RCT_ADDR_INET) {
			rc->sp_dst = rcs_sadup(al->a.ipaddr);
		} else if (al->type == RCT_ADDR_FQDN) {
			goto err; /* XXX */
		} else {
			goto err;
		}
	} else {
		goto err;
	}
	rc->pref_dst = sl->dst->prefixlen;

	rc->ul_proto= sl->upper_layer_protocol;
	rc->dir = sl->direction;

	spmd_spd_update(rc);

	spmd_rcpfk_cont_sock_free(rc);
	return 0;
err:
	spmd_rcpfk_cont_sock_free(rc);
	return -1;
}

int
spmd_spd_update(struct rcpfk_msg *rc)
{
	int ret = 0;

#ifdef SPMD_DEBUG
	{
		struct sockaddr_storage src, dst;
		char shost[NI_MAXHOST];
		char dhost[NI_MAXHOST];
		char sserv[NI_MAXSERV];
		char dserv[NI_MAXSERV];
		int ret = 0;

		memset(&src, 0, sizeof(src));
		memset(&dst, 0, sizeof(dst));

		ret = getnameinfo(rc->sp_src, SPMD_SALEN(rc->sp_src), shost, sizeof(shost), sserv, sizeof(sserv), NI_NUMERICHOST|NI_NUMERICSERV);
		if (ret) {
			spmd_log(SPMD_L_ERR, "ERR: getnameinfo(src):%s", strerror(errno));
		}
		ret = getnameinfo(rc->sp_dst, SPMD_SALEN(rc->sp_dst), dhost, sizeof(dhost), dserv, sizeof(dserv), NI_NUMERICHOST|NI_NUMERICSERV);
		if (ret) {
			spmd_log(SPMD_L_ERR, "ERR: getnameinfo(dst):%s", strerror(errno));
		}
		spmd_log(SPMD_L_DEBUG, "[SP UPDATE] SRC=[%s]:%s DST=[%s]:%s", shost, sserv, dhost, dserv);
			
	}
#endif
	/* del & add SP */
#if 1
	ret = rcpfk_send_spdupdate(rc);
#else
	ret = rcpfk_send_spddelete(rc);
	if (ret < 0) {
		spmd_log(SPMD_L_WARNING, "can not delete SP");
	}

	ret = rcpfk_send_spdadd(rc);
	if (ret < 0) {
		spmd_log(SPMD_L_ERR, "can not add SP");
	} 
#endif

	return ret;
}

static void
spmd_rcpfk_cont_sock_free(struct rcpfk_msg *rc)
{
	if (!rc) 
		return;

	if (rc->sa_src)
		rc_free(rc->sa_src);
	if (rc->sa_dst)
		rc_free(rc->sa_dst);
	if (rc->sp_src)
		rc_free(rc->sp_src);
	if (rc->sp_dst)
		rc_free(rc->sp_dst);

	return;
}

/* call back */
static int 
spmd_pfkey_spdadd(struct rcpfk_msg *rc)
{
	DPRINTF("%s:called", __FUNCTION__);
	return 0;
}

static int 
spmd_pfkey_spddelete(struct rcpfk_msg *rc)
{
	DPRINTF("%s:called", __FUNCTION__);
	return 0;
}

/* 
 * This function convert only non-FQDN/non-MACRO sl to rc.
 * FQDN/MACRO valiables are set to NULL.
 */
int
sl_to_rc(struct rcf_selector *sl, struct rcpfk_msg *rc)
{
	struct rcf_policy *pl = NULL;
	struct rcf_ipsec *ips = NULL;
	struct rc_addrlist *al = NULL;
	//struct rc_addrlist *ipal = NULL;
	uint32_t satype = 0;

	if (!sl->pl) {
		return -1;
	}
	pl = sl->pl;

	if (!sl->pl->ips) {
		return -1;
	}
	ips = sl->pl->ips;

	/*** set rc->satype ***/
	if (ips->sa_ah) {
		satype |= SAT_AH;
	}
	if (ips->sa_esp) {
		satype |= SAT_ESP;
	}
	if (ips->sa_ipcomp) {
		satype |= SAT_IPCOMP;
	}
	if ( (satype&SAT_ESP) && !(satype&SAT_AH) && !(satype&SAT_IPCOMP) ) {
		rc->satype = RCT_SATYPE_ESP; 
	} else if ( !(satype&SAT_ESP) && (satype&SAT_AH) && !(satype&SAT_IPCOMP) ) {
		rc->satype = RCT_SATYPE_AH;
	} else if ( !(satype&SAT_ESP) && !(satype&SAT_AH) && (satype&SAT_IPCOMP) ) {
		rc->satype = RCT_SATYPE_IPCOMP;
	} else if ( (satype&SAT_ESP) && (satype&SAT_AH) && !(satype&SAT_IPCOMP) ) {
		rc->satype = RCT_SATYPE_AH_ESP;
	} else if ( !(satype&SAT_ESP) && (satype&SAT_AH) && (satype&SAT_IPCOMP) ) {
		rc->satype = RCT_SATYPE_AH_IPCOMP;
	} else if ( (satype&SAT_ESP) && !(satype&SAT_AH) && (satype&SAT_IPCOMP) ) {
		rc->satype = RCT_SATYPE_ESP_IPCOMP;
	} else if ( (satype&SAT_ESP) && (satype&SAT_AH) && (satype&SAT_IPCOMP) ) {
		rc->satype = RCT_SATYPE_AH_ESP_IPCOMP;
	} else {
		spmd_log(SPMD_L_ERR, "unknown SA type");
		return -1;
	}

	/*** set rc->samode (tunnel or transport) ***/
	rc->samode = ips->ipsec_mode;

	/*** set rc->lft_hardtime (policy lifetime) ***/
	rc->lft_hard_time = 0; /* at init time */

	/*** set rc->flags ***/
	rc->flags = 0;

	/*** set rc->sa_src, rc->sa_dst ***/
	if (rc->samode == RCT_IPSM_TUNNEL) {
		if (pl->my_sa_addr) { /* always single entry */
			al = pl->my_sa_addr;
			if (al->type == RCT_ADDR_MACRO) { /* check type MACRO */
				rc->sa_src = NULL; /* just ignore */
			} else {
				rc->sa_src = rcs_sadup(al->a.ipaddr);
			}
		} else { return -1; }
		if (pl->peers_sa_addr) { /* always single entry */
			al = pl->peers_sa_addr;
			if (rcs_is_addrmacro(al->a.vstr)) { /* check type MACRO */
				rc->sa_dst = rcs_sadup(al->a.ipaddr);
			} else {
				rc->sa_dst = NULL; /* just ignore */
			}
		} else { return -1; }
	}

	/*** set rc->slid (selector index) ***/
	rc->slid = (sl->sl_index<<SPMD_SLID_SHIFT_WIDTH); /* selector index */

	/*** set rc->sp_src, rc->pref_src, rc->sp_dst, rc->pref_dst ***/
	if (sl->src) { /* XXX do we need to care multiple entries? */
		al = sl->src;
		if (rcs_is_addrmacro(al->a.vstr)) { /* check type MACRO */
			rc->sp_src = rcs_sadup(al->a.ipaddr);
		} else {
			rc->sp_src = NULL;
		}
	} else {
		goto err;
	}
	rc->pref_src = sl->src->prefixlen;

	if (sl->dst) { /* XXX do we need to care multiple entries? */
		al = sl->dst;
		if (rcs_is_addrmacro(al->a.vstr)) { /* check type MACRO */
			rc->sp_dst = rcs_sadup(al->a.ipaddr);
		} else {
			rc->sp_dst = NULL;
		}
	} else {
		goto err;
	}
	rc->pref_dst = sl->dst->prefixlen;

	/*** set rc->ul_proto (upper layer protocol) ***/
	rc->ul_proto= sl->upper_layer_protocol;

	/*** set rc->dir (direction) ***/
	rc->dir = sl->direction;

	/* del & add SP */
	if (rcpfk_send_spddelete(rc) < 0) {
		spmd_log(SPMD_L_WARNING, "can not delete SP");
	}
	if (rcpfk_send_spdadd(rc) < 0) {
		spmd_log(SPMD_L_ERR, "can not add SP");
	} 

	spmd_rcpfk_cont_sock_free(rc);
	return 0;
err:
	spmd_rcpfk_cont_sock_free(rc);
	return -1;
}

/*--------------------------------------------------*/
static struct sp_queue *sp_queue_top = NULL;
/*
struct sp_queue {
	struct sp_queue *next;
	int sl_index;
	rc_type src_type;
	union {
		struct fqdn_list *src_fl;
		struct sockaddr *src_sa;
	} src;
	rc_type dst_type;
	union {
		struct fqdn_list *dst_fl;
		struct sockaddr *dst_sa;
	} dst;
};
*/

/* this have to be caled after add_fqdn_db() */
int
sp_queue_add(int sl_index, const struct rc_addrlist *src, const struct rc_addrlist *dst)
{
	struct sp_queue *spq = NULL, *new_spq=NULL, *spq_tmp = NULL;
	struct fqdn_list *fl = NULL;
	char *fqdn_str=NULL;
	size_t fqdn_strlen=0;

	DPRINTF("%s:called", __FUNCTION__);

	spq = sp_queue_search(sl_index);
	if (!spq) {
		new_spq = (struct sp_queue *)spmd_calloc(M_CACHE, sizeof(*new_spq));
		if (!new_spq) {
			goto err;
		}

		new_spq->sl_index = sl_index;

		new_spq->src_type = src->type;
		if (new_spq->src_type == RCT_ADDR_FQDN) {
			fqdn_str = (char *)vmem2str(src->a.vstr);
			fqdn_strlen = strlen(fqdn_str);
			fl = find_fqdn_db(fqdn_str, fqdn_strlen);
			if (!fl) goto err;
			new_spq->src.src_fl = fl;
		} else if (new_spq->src_type == RCT_ADDR_INET) {
			new_spq->src.src_sa = rcs_sadup(src->a.ipaddr);
		} else { /* XXX */
			goto err;
		}

		new_spq->dst_type = dst->type;
		if (new_spq->dst_type == RCT_ADDR_FQDN) {
			fqdn_str = (char *)vmem2str(dst->a.vstr);
			fqdn_strlen = strlen(fqdn_str);
			fl = find_fqdn_db(fqdn_str, fqdn_strlen);
			if (!fl) goto err;
			new_spq->dst.dst_fl = fl;
		} else if (new_spq->dst_type == RCT_ADDR_INET) {
			new_spq->dst.dst_sa = rcs_sadup(dst->a.ipaddr);
		} else { /* XXX */
			goto err;
		}

		if (!sp_queue_top) {
			sp_queue_top = new_spq;
		} else {
			spq_tmp=sp_queue_top;
			while (spq_tmp->next)
				spq_tmp = spq_tmp->next;
			spq_tmp->next = new_spq;
		}
	} else {
		return 0;
	}

	return 0;
err:
	if (new_spq) {
		if (new_spq->src_type == RCT_ADDR_INET) {
			spmd_free(M_CACHE, new_spq->src.src_sa);
		}
		if (new_spq->dst_type == RCT_ADDR_INET) {
			spmd_free(M_CACHE, new_spq->dst.dst_sa);
		}
		spmd_free(M_CACHE, new_spq);
	}
	return -1;
}

struct sp_queue *
sp_queue_search(int sl_index)
{
	struct sp_queue *spq = NULL;

	if (!sp_queue_top)
		return NULL;

	for (spq=sp_queue_top;spq;spq=spq->next) {
		if (spq->sl_index == sl_index)
			return spq;
	}

	return NULL;
}

int
fqdn_sp_update(void)
{
	struct sp_queue *sp = NULL;
	struct rcpfk_msg *rc = &pfkey_container;
	struct rcf_selector *sl = NULL;
	struct fqdn_list *fl = NULL;
	struct fqdn_addr_list *fal_src0 = NULL, *fal_src = NULL;
	struct fqdn_addr_list *fal_dst0 = NULL, *fal_dst = NULL;
	sa_family_t af;

	DPRINTF("%s:called", __FUNCTION__);

	if (!sp_queue_top)
		return 0;

	for (sp=sp_queue_top; sp; sp=sp->next) {
		if (sp->src_type == RCT_ADDR_FQDN) {
			fl = sp->src.src_fl;
			if (!fl)
				continue;
			fal_src0 = fl->fal;
			if (!fal_src0)
				continue;
		}
		if (sp->dst_type == RCT_ADDR_FQDN) {
			fl = sp->dst.dst_fl;
			if (!fl)
				continue;
			fal_dst0 = fl->fal;
			if (!fal_dst0)
				continue;
		}
		rcf_get_selector(sp->sl_index, &sl);
		sl_to_rc_wo_addr(sl, rc);
		/* fill rc sa */
		if ( (!fal_src) && (sp->src_type==RCT_ADDR_INET) ) {
			af = sp->src.src_sa->sa_family;
			rc->sp_src = sp->src.src_sa;
			for (fal_dst=fal_dst0; fal_dst; fal_dst=fal_dst->next) {
				if (af != fal_dst->sa->sa_family)
					continue;
				rc->sp_dst = fal_dst->sa;
				if (af == AF_INET) {
					rc->pref_dst = 32;
					((struct sockaddr_in *)rc->sp_dst)->sin_port = htons(sl->dst->port);
				} else if (af == AF_INET6) {
					rc->pref_dst = 128;
					((struct sockaddr_in6 *)rc->sp_dst)->sin6_port = htons(sl->dst->port);
				} else {
					/* error */
					continue;
				}
				spmd_spd_update(rc);
				rc->slid++;
			}
		}
		else { 
			for (fal_src=fal_src0; fal_src; fal_src=fal_src->next) {
				af = fal_src->sa->sa_family;
				rc->sp_src = fal_src->sa;
				//rc->pref_src = sl->src->prefixlen; /* can not determine */
				if (af == AF_INET) {
					rc->pref_src = 32;
					((struct sockaddr_in *)rc->sp_src)->sin_port = htons(sl->src->port);
				} else if (af == AF_INET6) {
					rc->pref_src = 128;
					((struct sockaddr_in6 *)rc->sp_src)->sin6_port = htons(sl->src->port);
				} else {
					/* error */
					continue;
				}
				if ( (!fal_dst) && (sp->dst_type==RCT_ADDR_INET) ) {
					if (af != sp->dst.dst_sa->sa_family) 
						continue;
					rc->sp_dst = sp->dst.dst_sa;
					spmd_spd_update(rc);
					rc->slid++;
				}
				else { 
					for (fal_dst=fal_dst0; fal_dst; fal_dst=fal_dst->next) {
						if (af != fal_dst->sa->sa_family)
							continue;
						rc->sp_dst = fal_dst->sa;
						if (af == AF_INET) {
							rc->pref_dst = 32;
							((struct sockaddr_in *)rc->sp_dst)->sin_port = htons(sl->dst->port);
						} else if (af == AF_INET6) {
							rc->pref_dst = 128;
							((struct sockaddr_in6 *)rc->sp_dst)->sin6_port = htons(sl->dst->port);
						} else {
							/* error */
							continue;
						}
						spmd_spd_update(rc);
						rc->slid++;
					}
				}
			} 
		}

		rcf_free_selector(sl);
	}

	return 0;
}

/* fill rc from sl except for sp_src/sp_dst and prefix */
int
sl_to_rc_wo_addr(struct rcf_selector *sl, struct rcpfk_msg *rc)
{
	struct rcf_policy *pl = NULL;
	struct rcf_ipsec *ips = NULL;
	struct rc_addrlist *al = NULL;
	uint32_t satype = 0;

	if (!sl->pl) {
		return -1;
	}
	pl = sl->pl;

	if (!sl->pl->ips) {
		return -1;
	}
	ips = sl->pl->ips;

	/*** set rc->satype ***/
	if (ips->sa_ah) {
		satype |= SAT_AH;
	}
	if (ips->sa_esp) {
		satype |= SAT_ESP;
	}
	if (ips->sa_ipcomp) {
		satype |= SAT_IPCOMP;
	}
	if ( (satype&SAT_ESP) && !(satype&SAT_AH) && !(satype&SAT_IPCOMP) ) {
		rc->satype = RCT_SATYPE_ESP; 
	} else if ( !(satype&SAT_ESP) && (satype&SAT_AH) && !(satype&SAT_IPCOMP) ) {
		rc->satype = RCT_SATYPE_AH;
	} else if ( !(satype&SAT_ESP) && !(satype&SAT_AH) && (satype&SAT_IPCOMP) ) {
		rc->satype = RCT_SATYPE_IPCOMP;
	} else if ( (satype&SAT_ESP) && (satype&SAT_AH) && !(satype&SAT_IPCOMP) ) {
		rc->satype = RCT_SATYPE_AH_ESP;
	} else if ( !(satype&SAT_ESP) && (satype&SAT_AH) && (satype&SAT_IPCOMP) ) {
		rc->satype = RCT_SATYPE_AH_IPCOMP;
	} else if ( (satype&SAT_ESP) && !(satype&SAT_AH) && (satype&SAT_IPCOMP) ) {
		rc->satype = RCT_SATYPE_ESP_IPCOMP;
	} else if ( (satype&SAT_ESP) && (satype&SAT_AH) && (satype&SAT_IPCOMP) ) {
		rc->satype = RCT_SATYPE_AH_ESP_IPCOMP;
	} else {
		spmd_log(SPMD_L_ERR, "unknown SA type");
		return -1;
	}

	/*** set rc->samode (tunnel or transport) ***/
	rc->samode = ips->ipsec_mode;

	/*** set rc->lft_hardtime (policy lifetime) ***/
	rc->lft_hard_time = 0; /* at init time */

	/*** set rc->flags ***/
	rc->flags = 0;

	/*** set rc->sa_src, rc->sa_dst ***/
	if (rc->samode == RCT_IPSM_TUNNEL) {
		if (pl->my_sa_addr) { /* always single entry */
			al = pl->my_sa_addr;
			if (al->type == RCT_ADDR_MACRO) { /* check type MACRO */
				rc->sa_src = NULL; /* just ignore */
			} else {
				rc->sa_src = rcs_sadup(al->a.ipaddr);
			}
		} else { return -1; }
		if (pl->peers_sa_addr) { /* always single entry */
			al = pl->peers_sa_addr;
			if (rcs_is_addrmacro(al->a.vstr)) { /* check type MACRO */
				rc->sa_dst = rcs_sadup(al->a.ipaddr);
			} else {
				rc->sa_dst = NULL; /* just ignore */
			}
		} else { return -1; }
	}

	/*** set rc->slid (selector index) ***/
	rc->slid = (sl->sl_index<<SPMD_SLID_SHIFT_WIDTH); /* selector index */

	/*** set rc->ul_proto (upper layer protocol) ***/
	rc->ul_proto= sl->upper_layer_protocol;

	/*** set rc->dir (direction) ***/
	rc->dir = sl->direction;

	return 0;
}
