/*
 * Copyright (C) 2004 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <netinet/tcp.h>
#include <signal.h>
#include <errno.h>
#include "spmd_includes.h"
#include "spmd_internal.h"

#define TYPE_NS_ADD 1
#define TYPE_NS_DEL 2
#define TYPE_NS_LST 4
#define TYPE_STAT   8
#define TYPE_RELOAD 16

static int
getline(int fd, char *buf, int len)
{
	char c;    /* read char */
	int n = 0; /* read bytes */

	while ( (read(fd, &c, 1)) == 1 ) {
		*buf++ = c;
		n++;
		if ( c == '\n' ) {
			if (*(buf-2) != '\r')
				goto err;
			*(buf-2) = '\0';
			n -= 2;
			return n;
		}

		if (n == len -1) {
			buf = '\0';
			return n;
		}
	}
err:
	/* error */
	buf = '\0';
	return -1;
}

static int
status(int s)
{
	int w;
	char rbuf[512];

	w = write(s, "STAT\r\n", strlen("STAT\r\n"));
	while ( getline(s, rbuf, sizeof(rbuf)) > 0) {
		if (rbuf[0] != '2')
			return -1;
		fprintf(stdout, "%s\n", rbuf+4);
		if (rbuf[3] == ' ')
			return 0;
	}

	return 0;
}

static int
ns(int s, char *addr, int flag)
{
	int w;
	char rbuf[512];
	char wbuf[512];
	char naddr[INET6_ADDRSTRLEN];
	int match=0;
	struct addrinfo hints, *res;
	int gai_err;

	if (addr) {
		/* normalization */
		memset(&hints, 0, sizeof(hints));
		hints.ai_flags = AI_NUMERICHOST;
		gai_err = getaddrinfo(addr, NULL, &hints, &res);
		if (gai_err < 0) {
			fprintf(stderr, "%s\n", gai_strerror(gai_err));
			return -1;
		}
		if ( getnameinfo(res->ai_addr, res->ai_addrlen, 
				naddr, sizeof(naddr), NULL, 0, NI_NUMERICHOST) < 0) {
			perror("getnameinfo");
			return -1;
		}
		freeaddrinfo(res);
	}


	if (flag == TYPE_NS_ADD) {
		w = write(s, "NS LIST\r\n", strlen("NS LIST\r\n"));
		while ( getline(s, rbuf, sizeof(rbuf)) > 0) {
			if (rbuf[0] != '2')
				return -1;
			if (!strncasecmp(rbuf+4, naddr, strlen(naddr))) { /* match */
				match=1;
			}
			if (rbuf[3] == ' ') /* last line */
				break;
		}

		if (match) {
			snprintf(wbuf, sizeof(wbuf), "NS CHANGE %s\r\n", naddr);
			w= write(s, wbuf, strlen(wbuf));
		} else {
			snprintf(wbuf, sizeof(wbuf), "NS ADD %s\r\n", naddr);
			w= write(s, wbuf, strlen(wbuf));
		}
		return 0;
	} else if (flag == TYPE_NS_DEL) {
		int lines=0;
		w = write(s, "NS LIST\r\n", strlen("NS LIST\r\n"));
		while ( getline(s, rbuf, sizeof(rbuf)) > 0) {
			if (rbuf[0] != '2')
				return -1;
			if (!strncasecmp(rbuf+4, naddr, strlen(naddr))) { /* match */
				match=1;
			}
			lines++;
			if (rbuf[3] == ' ')
				break;
		}

		if (match && lines >1) {
			snprintf(wbuf, sizeof(wbuf), "NS DELETE %s\r\n", naddr);
			w= write(s, wbuf, strlen(wbuf));
		}
		return 0;
	} else if (flag == TYPE_NS_LST) {
		write(s, "NS LIST\r\n", strlen("NS LIST\r\n"));
		while ( getline(s, rbuf, sizeof(rbuf)) > 0) {
			if (rbuf[0] != '2')
				return -1;
			fprintf(stdout, "%s\n", rbuf+4);
			if (rbuf[3] == ' ')
				return 0;
		}
	} else {
		return -1;
	}

	return -1;
}

static int
reload(void)
{
	pid_t pid=-1;

	pid = read_pid(SPMD_PID_FILE);
	if (pid<0) {
		fprintf(stderr, "can't read pid file\n");
		return -1;
	}
	if (kill(pid, SIGHUP)<0) {
		fprintf(stderr, "can't send SIGUP:%s\n",
			strerror(errno));
		return -1;
	}
	return 0;
}

static int
sock_open_sa(const struct sockaddr *sa)
{
	int rtn = 0;
	int on = 1;
	int s;

	s = socket(sa->sa_family, SOCK_STREAM, 0);
	if (s<0) {
		fprintf(stderr, "%s", strerror(errno));
		s = -1;
		goto fin; 
	} 

	rtn = setsockopt(s, IPPROTO_TCP, TCP_NODELAY, (char *) &on, sizeof (on));
	if (rtn < 0 && (sa->sa_family != AF_UNIX) ) {
		fprintf(stderr, "setsockopt(TCP_NODELAY) failed");
		close(s);
		s = -1;
		goto fin; 
	}

	if (connect(s, sa, SPMD_SALEN(sa))<0) {
		close(s);
		s = -1;
	} 

fin:
	return s;
}

static struct sockaddr *
build_sock_unix(const char *path)
{
	struct sockaddr_un *sun = NULL;

	sun = malloc(sizeof(struct sockaddr_un));
	if (!sun) {
		fprintf(stderr, "can not allocate memory");
		return NULL;
	}
	memset(sun, 0, sizeof(struct sockaddr_un));

	if (strlen(path) >= sizeof(sun->sun_path)) {
		fprintf(stderr, "path too long");
		free(sun);
		return NULL;
	}

	sun->sun_family = AF_UNIX;
	strcpy(sun->sun_path, path);
#ifdef HAVE_SA_LEN
	sun->sun_len = SUN_LEN(sun);
#endif
	return (struct sockaddr *)sun;
}

static int
sock_open_file(const struct sockaddr *sa)
{
	int s = -1;
	int on = 1;

	if (!sa) {
		goto fin;
	}

	s = socket(PF_UNIX, SOCK_STREAM, 0);
	if (s<0) {
		fprintf(stderr, "%s", strerror(errno));
		s = -1;
		goto fin; 
	} 

	setsockopt(s, IPPROTO_TCP, TCP_NODELAY, (char *) &on, sizeof (on));

	if (connect(s, sa, SUN_LEN((struct sockaddr_un *)sa))<0) {
		close(s);
		s = -1;
	} 

fin:
	return s;
}

static int
login()
{
	char rbuf[512];
	int r,w;
	int s = -1;
	struct rc_addrlist *rcl_top = NULL, *rcl;
	struct sockaddr *sa;

	rcf_get_spmd_interfaces(&rcl_top);
	for (rcl=rcl_top; rcl; rcl=rcl->next) {
		switch (rcl->type) {
		case RCT_ADDR_INET:
			s = sock_open_sa(rcl->a.ipaddr);
			if (s < 0) {
				continue;
			}
			goto connect_ok;
			break;
		case RCT_ADDR_FILE:
			sa = build_sock_unix(vmem2str(rcl->a.vstr));
			s = sock_open_file(sa);
			if (s < 0) {
				free(sa);
				continue;
			}
			goto connect_ok;
			break;
		case RCT_ADDR_FQDN:
			{
				char *fqdn = NULL;
				char portstr[16];
				struct addrinfo hints, *res0, *res;
				int gai_err;
				char host[NI_MAXHOST];

				fqdn = (char *)vmem2str(rcl->a.vstr);
				memset(portstr, 0, sizeof(portstr));
				if (rcl->port == 0) {
					snprintf(portstr, sizeof(portstr), "%d", SPMD_SHELL_PORT);
				} else {
					snprintf(portstr, sizeof(portstr), "%d", rcl->port);
				}

				memset(&hints, 0, sizeof(hints));
				hints.ai_family = PF_UNSPEC;
				hints.ai_socktype = SOCK_STREAM;
				gai_err = getaddrinfo(fqdn, portstr, &hints, &res0);
				if (gai_err < 0) {
					fprintf(stderr, "%s", gai_strerror(gai_err));
					continue;
				}
				for (res=res0; res; res=res->ai_next) {
					getnameinfo(res->ai_addr, res->ai_addrlen, host, sizeof(host), NULL, 0, NI_NUMERICHOST);
					if (res->ai_family == AF_INET) {
						if (!strncmp(host, "::1", strlen(host))) {
							continue;
						}
					} else if (res->ai_family == AF_INET) {
						if (!strncmp(host, "127.0.0.1", strlen(host))) {
							continue;
						}
					} else {
						continue;
					}
					s = sock_open_sa(res->ai_addr);
					if (s<0) {
						continue;
					} else {
						freeaddrinfo(res0);
						goto connect_ok;
					}
				}
				freeaddrinfo(res0);
				continue;
			}
			break;
		default:
			continue;
			break;
		}
	}
		
connect_ok:
	/* read banner */
	r = read(s, rbuf, sizeof(rbuf));
	if (r<0) {
		perror("LOGIN:read");
		close(s);
		exit(EXIT_FAILURE);
	}
	
	if (rbuf[0] != '2') {
		fprintf(stderr, "get error status\n");
		close(s);
		exit(EXIT_FAILURE);
	}

	w = write(s, "LOGIN racoon2\r\n", strlen("LOGIN racoon2\r\n"));
	r = read(s, rbuf, sizeof(rbuf));
	if (r<0) {
		perror("LOGIN:read");
		exit(EXIT_FAILURE);
	}
	if (rbuf[0] != '2') {
		fprintf(stderr, "login failure\n");
		exit(EXIT_FAILURE);
	}
	return s;
}

static int
quit(int s)
{
	char rbuf[512];
	int r,w;

	w = write(s, "QUIT\r\n", strlen("QUIT\r\n"));
	r = read(s, rbuf, sizeof(rbuf));
	if (r<0) {
		perror("QUIT:read");
		close(s);
		return -1;
	}
	close(s);
	return 0;
}

static void
print_help(void)
{
	fprintf(stderr, 
		"spmdctl [-f RACOON2_CONF_FILE] COMMANDS\n"
		"\tCOMMANDS:\n"
		"\t\tns {add|delete} address : add/delete nameserver\n"
		"\t\tns list                 : show nameservers\n"
		"\t\tstatus                  : show statistics\n"
		/* "\t\treload                  : reload\n" */
	       );
}

int
main(int argc, char **argv)
{
	int s;
	char addr[INET6_ADDRSTRLEN];
	int ret;
	int type=0;
	char config[PATH_MAX];
	char **cargv;
	int cargc;

	memset(config, 0, sizeof(config));
	//strncpy(config, RACOON2_CONFIG_FILE, sizeof(config)-1);
	strlcpy(config, RACOON2_CONFIG_FILE, sizeof(config));

	/* options */
	cargc = argc;
	cargv = argv;
	if (cargc > 1) {
		if (!strncasecmp(cargv[1], "-f", strlen(cargv[1]))) {
			if (cargc > 2) {
				//strncpy(config, cargv[2], sizeof(config)-1);
				strlcpy(config, cargv[2], sizeof(config));
				cargv += 2;
				cargc -= 2;;
			} else {
				print_help();
				exit(EXIT_FAILURE);
			}
		}
		if (!strncasecmp(cargv[1], "ns", strlen(cargv[1]))) {
			if (cargc==4 && !strncasecmp(cargv[2], "add", strlen(cargv[2]))) {
				//strncpy(addr, cargv[3], sizeof(addr)-1);
				strlcpy(addr, cargv[3], sizeof(addr));
				type=TYPE_NS_ADD;
			} else if (cargc==4 && !strncasecmp(cargv[2], "delete", strlen(cargv[2]))) {
				//strncpy(addr, cargv[3], sizeof(addr)-1);
				strlcpy(addr, cargv[3], sizeof(addr));
				type=TYPE_NS_DEL;
			} else if (cargc==3 && !strncasecmp(cargv[2], "list", strlen(cargv[2]))) {
				type=TYPE_NS_LST;
			} else {
				print_help();
				exit(EXIT_SUCCESS);
			}
		} else if (cargc==2 && !strncasecmp(cargv[1], "status", strlen(cargv[1]))) {
			type=TYPE_STAT;
		} else if (cargc==2 && !strncasecmp(cargv[1], "reload", strlen(cargv[1]))) {
			type=TYPE_RELOAD;
		} else {
			print_help();
			exit(EXIT_SUCCESS);
		}
	} else {
		print_help();
		exit(EXIT_SUCCESS);
	}

	if (geteuid() != 0) {
		fprintf(stderr, "you should run this program as root.\n");
		exit(EXIT_FAILURE);
	}

	/* init libracoon */
	plog_init(0, NULL, argv[0], 1, 1);      /* XXX arg vals are correct? */
	if (rbuf_init(8, 80, 4, 160, 4) == -1) {/* XXX arg vals are correct? */
		fprintf(stderr, "failed to initilize libracoon (rbuf_init())");
		exit(EXIT_FAILURE);
	}

	if (rcf_read(config, 0) < 0) {
		fprintf(stderr, "failed to parse config file:%s", config);
		exit(EXIT_FAILURE);
	}

	switch (type) {
		case TYPE_NS_ADD:
			s = login();
			ret = ns(s, addr, TYPE_NS_ADD);
			if (ret<0) 
				fprintf(stderr, "operation failed\n");
			quit(s);
			break;
		case TYPE_NS_DEL:
			s = login();
			ret = ns(s, addr, TYPE_NS_DEL);
			if (ret<0) 
				fprintf(stderr, "operation failed\n");
			quit(s);
			break;
		case TYPE_NS_LST:
			s = login();
			ret = ns(s, NULL, TYPE_NS_LST);
			if (ret<0) 
				fprintf(stderr, "operation failed\n");
			quit(s);
			break;
		case TYPE_STAT:
			s = login();
			ret = status(s);
			if (ret<0) 
				fprintf(stderr, "operation failed\n");
			quit(s);
			break;
		case TYPE_RELOAD:
			ret = reload();
			if (ret<0)
				fprintf(stderr, "operation failed\n");
			break;
		default: /* not reach */
			fprintf(stderr, "internal error\n");
			break;
	}

	rcf_clean();

	return 0;
}


